"""
This module contains utility functions for the calculations of the reweighting factors.

Maximilian Welsch 2018.
"""
import numpy as np
import pandas as pd

from typing import Dict

from analysistools.physics_vectors.fourvector import FourVector


def var_m2_dstst(df: pd.DataFrame) -> np.array:
    """
    Calculates the invariant mass of the D** meson.

    :param df:
    :return: Inv. Mass of D**
    """

    p_d_star_star = FourVector(df[["Dstarstar_E", "Dstarstar_px", "Dstarstar_py", "Dstarstar_pz"]].values)

    return p_d_star_star.M()


def var_m_b_meson(df: pd.DataFrame) -> np.array:
    """
    Calculates the invariant mass of the D** meson.

    :param df:
    :return: Inv. Mass of D**
    """

    p_d_star_star = FourVector(df[["B_E", "B_px", "B_py", "B_pz"]].values)

    return p_d_star_star.M()


def var_w(df: pd.DataFrame) -> np.array:
    """
    Calculates w, where
    w = (m_B^2 + m_{D^{**}}^2 - q`^2)/(2*m_B*m_{D^{**}}).
    :param df: DataFrame
    :return: w
    """
    p_b_meson = FourVector(df[["B_E", "B_px", "B_py", "B_pz"]].values)
    p_d_star_star = FourVector(df[["Dstarstar_E", "Dstarstar_px", "Dstarstar_py", "Dstarstar_pz"]].values)

    q = p_b_meson - p_d_star_star
    return (p_b_meson.M2() + p_d_star_star.M2() - q.M2()) / (2 * p_b_meson.M() * p_d_star_star.M())


def var_costhetal(df: pd.DataFrame) -> np.array:
    """
    Calculates cos(theta)_l, which is the angle between the lepton and the
    W boson in the W boson rest frame.
    :param df: DataFrame
    :return: cos(theta)_l
    """
    p_b_meson = FourVector(df[["B_E", "B_px", "B_py", "B_pz"]].values)
    p_l = FourVector(df[["l_E", "l_px", "l_py", "l_pz"]].values)
    p_nu = FourVector(df[["nu_E", "nu_px", "nu_py", "nu_pz"]].values)

    p_w = p_l + p_nu
    # boost W and lepton in the CMS of the B meson
    b_b_meson = -p_b_meson.BoostVector()
    p_l.Boost(b_b_meson)
    p_w.Boost(b_b_meson)
    # boost lepton in the CMS of the W boson
    b_w = -p_w.BoostVector()
    p_l.Boost(b_w)
    return np.cos(p_w.Angle(p_l.Vec))


def var_costhetanu(df: pd.DataFrame) -> np.array:
    """
    Calculates cos(theta)_nu, which is the angle between the D^{(*)}
    and the D^{**} in the D^{**} rest frame
    :param df: DataFrame
    :return: cos(theta)_nu
    """
    p_b_meson = FourVector(df[["B_E", "B_px", "B_py", "B_pz"]].values)
    p_d_star_star = FourVector(df[["Dstarstar_E", "Dstarstar_px", "Dstarstar_py", "Dstarstar_pz"]].values)
    p_d_meson = FourVector(df[["D_E", "D_px", "D_py", "D_pz"]].values)

    b_b_meson = -p_b_meson.BoostVector()
    p_d_star_star.Boost(b_b_meson)
    p_d_meson.Boost(b_b_meson)
    b_d_star_star = -p_d_star_star.BoostVector()
    p_d_meson.Boost(b_d_star_star)
    return np.cos(p_d_star_star.Angle(p_d_meson.Vec))


def var_chi_prime(df: pd.DataFrame) -> np.array:
    """
    Calculates chi', which is the angle bewtween
    the decay plane of the D^{**} and the decay plane of the W boson.
    :param df: DataFrame
    :return: chi'
    """
    pl = FourVector(df[["l_E", "l_px", "l_py", "l_pz"]].values)
    pnu = FourVector(df[["nu_E", "nu_px", "nu_py", "nu_pz"]].values)
    p_w = pl + pnu
    p_d_st_st = FourVector(df[["Dstarstar_E", "Dstarstar_px", "Dstarstar_py", "Dstarstar_pz"]].values)

    p_d = FourVector(df[["D_E", "D_px", "D_py", "D_pz"]].values)

    boost_w_cm = -p_w.BoostVector()
    boost_d_star_star_cm = -p_d_st_st.BoostVector()

    p_d.Boost(boost_d_star_star_cm)
    pl.Boost(boost_w_cm)

    costhetav = (p_d.X * p_d_st_st.X + p_d.Y * p_d_st_st.Y + p_d.Z * p_d_st_st.Z) / (p_d.T * p_d_st_st.T)

    costhetawl = (-pl.X * p_d_st_st.X - pl.Y * p_d_st_st.Y - pl.Z * p_d_st_st.Z) / (pl.T * p_d_st_st.T)

    # actual chi calculation
    ltrans = pl.Vec - p_w.Vec.Unit().Scale(costhetawl * pl.Vec.Perp())
    vtrans = p_d.Vec - p_d_st_st.Vec.Unit().Scale(costhetav * p_d.Vec.Perp())

    chi = np.arctan2(ltrans.Cross(vtrans) * p_w.Vec.Unit(), ltrans * vtrans)
    chi[chi < 0] += 2 * np.pi

    return chi


def calculate_reweighting_variables(df: pd.DataFrame) -> None:
    """
    Calculates variables needed for reweighting and creates a new data frame.
    :param df: Original data frame
    :return: DataFrame with reweighting variables.
    """
    new_df_vars_dict = {
        "w": var_w(df),
        "costhetal": var_costhetal(df),
        "costhetanu": var_costhetanu(df),
        "chi_prime": var_chi_prime(df),
        "mDss": var_m2_dstst(df),
        "mB": var_m_b_meson(df),
        "__weight__": np.ones((len(df.index))),
    }  # type: Dict[str, np.ndarray]

    assert all(col_name not in df.columns or col_name == "__weight__" for col_name in new_df_vars_dict.keys()), [
        cn for cn in new_df_vars_dict.keys() if cn in df.columns and cn != "__weight__"
    ]

    for col_name, values in new_df_vars_dict.items():
        assert len(values.shape) == 1, values.shape
        assert len(values) == len(df.index), (len(values), len(df.index))
        df.loc[:, col_name] = values
