"""
Validation of MVA Trainings

Felix Metzner 2020
"""

import os
import json
import pickle

import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from sklearn.metrics import roc_curve, roc_auc_score
from typing import Union, Optional, Tuple, List, Dict, Any, Type, NamedTuple

from rdstar.utilities import PathType
from rdstar.offline_analysis.rdstar_plots import rdstar_color_map
from rdstar.latex_labels.latex_observable_names import LatexObsNames
from rdstar.mva_methods.mva_methods import MVABaseClass, FeatureImportanceInfo

from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.plot_utilities import export, color_fader
from templatefitter.plotter.histogram_plots import SimpleHistogramPlot
from templatefitter.plotter.plot_style import KITColors, set_matplotlibrc_params, xlabel_pos, ylabel_pos

set_matplotlibrc_params()

__all__ = [
    "MVAValidation",
]


class WorkingPointSetup(NamedTuple):
    working_point: float

    @property
    def appendix(self) -> str:
        return "alt_working_point_" + str(self.working_point).replace(".", "p")

    @property
    def appendix_for_train_data(self) -> str:
        return f"{self.appendix}_on_train_data"


class MVAValidation:
    default_fig_size = (4, 4)  # type: Tuple[Union[int, float], Union[int, float]]
    default_fig_dpi = 200  # type: int

    default_correlation_fig_size = (7, 6)  # type: Tuple[float, float]
    default_correlation_fig_dpi = 200  # type: int
    default_correlation_aspect_ratio = "equal"  # type: str

    def __init__(
        self,
        mva_class: Type[MVABaseClass],
        path_train_df: PathType,
        path_test_df: PathType,
        path_var_list: PathType,
        path_bdt_weights: PathType,
        target_var: str,
        training_histogram_variables_dict: Dict[str, HistVariable],
        validation_variables: List[HistVariable],
        mva_output_name: str,
        output_dir_path: PathType,
        additional_options: Optional[Dict[str, Any]] = None,
    ) -> None:
        assert all(v.df_label == k for k, v in training_histogram_variables_dict.items()), [
            (k, v.df_label) for k, v in training_histogram_variables_dict.items() if v.df_label != k
        ]

        self._mva_class = mva_class  # type: Type[MVABaseClass]  # A MVA class which inherits MVABaseClass

        self._train_df_path = path_train_df
        self._test_df_path = path_test_df

        self._bdt_info_var_list_path = path_var_list
        self._bdt_info_weights_path = path_bdt_weights

        self._target_var = target_var
        self._hist_var_dict = training_histogram_variables_dict
        self._validation_variables = validation_variables
        self._mva_output_name = mva_output_name

        self._target_dir = output_dir_path

        if isinstance(additional_options, dict):
            assert all(isinstance(k, str) for k in additional_options.keys()), (
                k for k in additional_options.keys() if not isinstance(k, str)
            )
            assert all(v is not None for v in additional_options.values()), (
                k for k, v in additional_options.items() if v is None
            )
        else:
            assert additional_options is None, (type(additional_options).__name__, additional_options)
        self._options_dict = {} if additional_options is None else additional_options  # type: Dict[str, Any]

    @property
    def target_variable_column_name(self) -> str:
        return self._target_var

    @property
    def train_variables_column_names(self) -> List[str]:
        return list(self._hist_var_dict.keys())

    @property
    def train_variable_col_name_to_hist_var_dict(self) -> Dict[str, HistVariable]:
        return self._hist_var_dict

    @property
    def mva_output_column_name(self) -> str:
        return self._mva_output_name

    @property
    def figure_size(self) -> Tuple[Union[int, float], Union[int, float]]:
        return self.default_fig_size

    @property
    def figure_dpi(self) -> int:
        return self.default_fig_dpi

    @property
    def correlation_figure_size(self) -> Tuple[Union[int, float], Union[int, float]]:
        if self.has_optional_info(key="correlation_figure_size"):
            fig_size = self.get_optional_info(key="correlation_figure_size")
            assert isinstance(fig_size, tuple), fig_size
            assert len(fig_size) == 2, fig_size
            assert all(isinstance(v, (int, float)) for v in fig_size), fig_size
            tested_fig_size = (fig_size[0], fig_size[1])  # type: Tuple[Union[int, float], Union[int, float]]
            return tested_fig_size
        return self.default_correlation_fig_size

    @property
    def correlation_figure_dpi(self) -> int:
        if self.has_optional_info(key="correlation_figure_dpi"):
            fig_dpi = self.get_optional_info(key="correlation_figure_dpi")
            assert isinstance(fig_dpi, int), fig_dpi
            return fig_dpi
        return self.default_correlation_fig_dpi

    @property
    def correlation_aspect_ratio(self) -> Union[str, float]:
        _aspect_ratio = self.get_optional_info_with_default(
            key="correlation_aspect_ratio",
            default=self.default_correlation_aspect_ratio,
        )  # type: Any
        assert isinstance(_aspect_ratio, (float, str, int)), _aspect_ratio
        if isinstance(_aspect_ratio, int):
            return float(_aspect_ratio)
        elif isinstance(_aspect_ratio, str):
            assert _aspect_ratio in ("auto", "equal")
            return _aspect_ratio
        else:
            return _aspect_ratio

    @property
    def target_dir(self) -> PathType:
        return self._target_dir

    @property
    def optional_options(self) -> Dict[str, Any]:
        return self._options_dict

    def has_optional_info(self, key: str) -> bool:
        return key in self.optional_options.keys()

    def get_optional_info(self, key: str) -> Optional[Any]:
        if self.has_optional_info(key=key):
            return self.optional_options[key]
        return None

    def get_optional_info_with_default(self, key: str, default: Any) -> Any:
        return self.optional_options.get(key, default)

    @property
    def alternative_working_point_setups(self) -> Tuple[WorkingPointSetup, ...]:
        if not self.has_optional_info(key="alternative_working_points"):
            return tuple()
        alt_wps = self.get_optional_info(key="alternative_working_points")
        assert isinstance(alt_wps, (tuple, list)), alt_wps
        assert all(isinstance(v, float) for v in alt_wps), alt_wps
        return tuple([WorkingPointSetup(working_point=v) for v in list(set(alt_wps))])

    def load_train_and_test_dfs(self, dry_run: bool) -> Tuple[Union[pd.DataFrame, object], Union[pd.DataFrame, object]]:
        if dry_run:
            return pd.DataFrame(), pd.DataFrame()
        else:
            return pd.read_hdf(self._train_df_path), pd.read_hdf(self._test_df_path)

    @staticmethod
    def load_cs_vars(path: PathType) -> List[str]:
        with open(path, "rb") as f:
            cs_vars = pickle.load(f)
        return cs_vars

    def get_sig_bkg_masks(self, df: pd.DataFrame) -> Tuple[pd.Series, pd.Series]:
        is_sig = df[self.target_variable_column_name] == 1
        assert isinstance(is_sig, pd.Series), type(is_sig)
        return is_sig, ~is_sig

    def plot_training_vars(
        self,
        df: pd.DataFrame,
        title: Optional[str] = None,
        base_name: Optional[str] = None,
        dry_run: bool = False,
    ) -> List[Tuple[str, PathType]]:
        if not dry_run:
            signal_mask, background_mask = self.get_sig_bkg_masks(df=df)
            sig_df = df.loc[signal_mask]
            bkg_df = df.loc[background_mask]
        else:
            sig_df, bkg_df = pd.DataFrame(), pd.DataFrame()

        output_infos = []  # type: List[Tuple[str, PathType]]

        for var in self.train_variables_column_names:
            hist_var = self.train_variable_col_name_to_hist_var_dict[var]  # type: HistVariable

            name = f"{base_name}_validation_plot_" if base_name else ""
            file_name = f"{name}training_variable_distribution_{hist_var.df_label}"
            file_path = os.path.join(self.target_dir, f"{file_name}.pdf")
            output_infos.append((file_name, file_path))
            if dry_run:
                continue

            shp = SimpleHistogramPlot(hist_var)
            shp.add_component(label="Background", data=bkg_df, hist_type="stepfilled", color=KITColors.kit_red)
            shp.add_component(label="Signal", data=sig_df, hist_type="stepfilled", color=KITColors.kit_blue, alpha=0.8)

            fig, ax = plt.subplots(nrows=1, ncols=1, figsize=self.figure_size, dpi=self.figure_dpi)
            shp.plot_on(ax, normed=True, y_axis_scale=1.1)

            if title is not None:
                ax.set_title(title)

            export(fig=fig, filename=file_name, target_dir=self.target_dir, close_figure=True)

        return output_infos

    def plot_mva_output(
        self,
        mva_name: str,
        train_df: pd.DataFrame,
        test_df: pd.DataFrame,
        appendix: Optional[str] = None,
        title: Optional[str] = None,
        base_name: Optional[str] = None,
        dry_run: bool = False,
    ) -> Tuple[str, PathType]:
        name = f"{base_name}_validation_plot_" if base_name else ""
        file_name = f"{name}mva_output{'_' + appendix if appendix is not None else ''}"
        file_path = os.path.join(self.target_dir, f"{file_name}.pdf")
        if dry_run:
            return file_name, file_path

        train_signal_mask, train_background_mask = self.get_sig_bkg_masks(df=train_df)
        train_sig_df = train_df.loc[train_signal_mask]
        train_bkg_df = train_df.loc[train_background_mask]
        test_signal_mask, test_background_mask = self.get_sig_bkg_masks(df=test_df)
        test_sig_df = test_df.loc[test_signal_mask]
        test_bkg_df = test_df.loc[test_background_mask]

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=self.figure_size, dpi=self.figure_dpi)

        bins = np.linspace(0, 1.0, 26)
        bin_mid_values = (bins[:-1] + bins[1:]) / 2

        ax.hist(
            train_sig_df[self.mva_output_column_name],
            bins=bins,
            density=True,
            color=KITColors.kit_blue,
            alpha=0.8,
            histtype="stepfilled",
            lw=0.5,
            edgecolor=KITColors.kit_black,
            label="Signal (Train)",
        )
        ax.hist(
            train_bkg_df[self.mva_output_column_name],
            bins=bins,
            density=True,
            color=KITColors.kit_red,
            alpha=0.8,
            histtype="stepfilled",
            lw=0.5,
            edgecolor=KITColors.kit_black,
            label="Background (Train)",
        )

        test_sig_bc, _ = np.histogram(test_sig_df[self.mva_output_column_name], bins=bins, density=True)
        test_bkg_bc, _ = np.histogram(test_bkg_df[self.mva_output_column_name], bins=bins, density=True)

        ax.plot(
            bin_mid_values,
            test_sig_bc,
            ls="",
            marker=".",
            color=KITColors.kit_blue,
            markeredgewidth=0.5,
            markeredgecolor=KITColors.kit_black,
            label="Signal (Test)",
        )
        ax.plot(
            bin_mid_values,
            test_bkg_bc,
            ls="",
            marker=".",
            color=KITColors.kit_red,
            markeredgewidth=0.5,
            markeredgecolor=KITColors.kit_black,
            label="Background (Test)",
        )

        if title is not None:
            ax.set_title(title)

        ax.set_xlabel(f"{mva_name} Output", **xlabel_pos)
        ax.set_ylabel("Normalized in arb. units", **ylabel_pos)
        ax.legend(loc="best")

        export(fig=fig, filename=file_name, target_dir=self.target_dir, close_figure=True)
        return file_name, file_path

    def plot_roc_curve(
        self,
        df: pd.DataFrame,
        working_point: float,
        mva_output_latex_name: Optional[str] = None,
        appendix: Optional[str] = None,
        title: Optional[str] = None,
        base_name: Optional[str] = None,
        dry_run: bool = False,
    ) -> Tuple[str, PathType, float]:
        name = f"{base_name}_validation_plot_" if base_name else ""
        file_name = f"{name}roc_curve{'_' + appendix if appendix is not None else ''}"
        file_path = os.path.join(self.target_dir, f"{file_name}.pdf")
        if dry_run:
            return file_name, file_path, -1.0

        if mva_output_latex_name is None:
            mva_output_latex_label = LatexObsNames.mva_default_output
        else:
            mva_output_latex_label = mva_output_latex_name

        y_true = df[self.target_variable_column_name]
        y_predict = df[self.mva_output_column_name]

        selected = y_predict > working_point

        sig_eff = np.sum(selected & (y_true == 1)) / np.sum(y_true == 1)
        bkg_rej = np.sum(selected & (y_true == 0)) / np.sum(y_true == 0)

        fp_rate, tp_rate, thr = roc_curve(y_true=y_true, y_score=y_predict)
        roc_auc = roc_auc_score(y_true=y_true, y_score=y_predict)

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=self.figure_size, dpi=self.figure_dpi)

        ax.plot(1 - fp_rate, tp_rate, lw=1.5, color=KITColors().kit_blue, label="ROC Curve")
        ax.plot(
            1 - bkg_rej,
            sig_eff,
            ls="",
            marker="o",
            color=KITColors().kit_red,
            label=rf"Working Point, {mva_output_latex_label}$ > {working_point}$",
        )

        ax.axhline(sig_eff, 0, 1 - bkg_rej, ls="--", color=KITColors().kit_red)
        ax.axvline(1 - bkg_rej, 0, sig_eff, ls="--", color=KITColors().kit_red)

        info_text = (
            f"ROC Integral = {roc_auc:.3f}\n"
            f"Signal Efficiency = {100 * sig_eff:.1f}%\n"
            f"Background Rejection = {100 * (1 - bkg_rej):.1f}%"
        )

        ax.text(x=0.05, y=0.5, s=info_text)

        if title is not None:
            ax.set_title(title)

        ax.set_xlabel("Background Suppression", xlabel_pos)
        ax.set_ylabel("Signal Efficiency", ylabel_pos)
        plt.legend()

        ax.set_xlim([0.0, 1.05])
        ax.set_ylim([0.0, 1.05])

        export(fig=fig, filename=file_name, target_dir=self.target_dir, close_figure=True)

        return file_name, file_path, roc_auc

    def plot_feature_importance(
        self,
        importance_info: FeatureImportanceInfo,
        base_name: Optional[str] = None,
        plot_best_n_features: Optional[int] = None,
        dry_run: bool = False,
    ) -> Tuple[str, PathType]:
        name = f"{base_name}_validation_plot_" if base_name else ""
        file_name = f"{name}_{importance_info.feature_importance_type}_feature_importance_ranking"
        if plot_best_n_features is not None:
            file_name += f"best_{plot_best_n_features}_features"
        output_path = os.path.join(self.target_dir, f"{file_name}.pdf")

        if dry_run:
            return file_name, output_path

        variables = np.array(list(importance_info.feature_importance.keys()))
        importance_values = np.array(list(importance_info.feature_importance.values()))

        hist_variable_dict = self.train_variable_col_name_to_hist_var_dict
        variable_names = np.array([hist_variable_dict[var].variable_name for var in variables])

        sorted_indices = np.argsort(importance_values)[::-1]
        n_bars = plot_best_n_features if plot_best_n_features else len(variables)
        bar_colors = [
            color_fader(
                color_1=KITColors.kit_green,
                color_2=KITColors.grey,
                mix=i / (n_bars - 1),
            )
            for i in range(n_bars)
        ]

        with sns.axes_style("white"):
            fig, ax = plt.subplots(nrows=1, ncols=1, figsize=self.figure_size, dpi=self.figure_dpi)
            sns.despine()
            ax.set_xlabel("Feature Importance in arb. units")
            sns.barplot(
                x=importance_values[sorted_indices[:plot_best_n_features]],
                y=variable_names[sorted_indices[:plot_best_n_features]],
                # palette=sns.cubehelix_palette(n_colors=n_bars, rot=-.4),
                palette=sns.color_palette(bar_colors),
                orient="h",
                ax=ax,
            )
            ax.set_title(f"{importance_info.feature_importance_type.capitalize()} Feature Importance")
            export(fig=fig, filename=file_name, target_dir=self.target_dir, close_figure=True)

        return file_name, output_path

    @staticmethod
    def save_feature_importance_dict(
        feature_importance_dict_info: Dict[str, Any],
        output_path: PathType,
    ) -> None:
        dir_path, base_name = os.path.split(output_path)
        tmp_output_path = os.path.join(dir_path, f"temp_{base_name}")
        with open(tmp_output_path, "w") as out_file:
            json.dump(feature_importance_dict_info, out_file, indent=2)
        os.rename(src=tmp_output_path, dst=output_path)

    def evaluate_feature_importance(
        self,
        train_df: pd.DataFrame,
        test_df: pd.DataFrame,
        base_name: str,
        mva_classifier: Optional[MVABaseClass] = None,
        dry_run: bool = False,
    ) -> List[Tuple[str, PathType]]:
        output_info_list = []  # type: List[Tuple[str, PathType]]
        if dry_run:
            feature_importance_getter_function = self._mva_class.get_empty_feature_importance_infos  # type: ignore
            feature_getter_kwargs = {}
        else:
            assert isinstance(mva_classifier, MVABaseClass), type(mva_classifier).__name__
            feature_importance_getter_function = mva_classifier.get_feature_importance  # type: ignore
            feature_getter_kwargs = {"train_data": train_df, "test_data": test_df}

        importance_info_dict = {}  # type: Dict[str, Dict[str, Union[int, float]]]

        for importance_info in feature_importance_getter_function(**feature_getter_kwargs):  # type: ignore
            assert importance_info.feature_importance_type not in importance_info_dict.keys(), importance_info
            importance_info_dict.update({importance_info.feature_importance_type: importance_info.feature_importance})
            x_best_output_key, x_best_output_path = self.plot_feature_importance(
                importance_info=importance_info,
                base_name=base_name,
                plot_best_n_features=10,
                dry_run=dry_run,
            )
            output_info_list.append((x_best_output_key, x_best_output_path))

            output_key, output_path = self.plot_feature_importance(
                importance_info=importance_info,
                base_name=base_name,
                plot_best_n_features=None,
                dry_run=dry_run,
            )
            output_info_list.append((output_key, output_path))

        importance_dict_name = f"{base_name}_feature_importance_ranking_info"
        importance_dict_path = os.path.join(self.target_dir, f"{importance_dict_name}.json")
        if not dry_run:
            self.save_feature_importance_dict(
                feature_importance_dict_info=importance_info_dict, output_path=importance_dict_path
            )
        output_info_list.append((importance_dict_name, importance_dict_path))

        return output_info_list

    def plot_variable_correlation(
        self,
        df: pd.DataFrame,
        base_name: Optional[str] = None,
        dry_run: bool = False,
    ) -> Tuple[str, PathType]:
        name = f"{base_name}_validation_plot_" if base_name else ""
        file_name = f"{name}correlation"
        file_path = os.path.join(self.target_dir, f"{file_name}.pdf")

        if dry_run:
            return file_name, file_path

        variable_set = [self.train_variable_col_name_to_hist_var_dict[var] for var in self.train_variables_column_names]
        variable_set.extend(self._validation_variables)

        var_columns = [var.df_label for var in variable_set]
        corr_df = df.loc[:, var_columns]
        corr_matrix = np.corrcoef(corr_df.T)

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=self.correlation_figure_size, dpi=self.correlation_figure_dpi)
        heatmap = ax.imshow(
            corr_matrix,
            cmap=rdstar_color_map,
            vmin=-1.0,
            vmax=1.0,
            aspect=self.correlation_aspect_ratio,
        )
        plt.colorbar(heatmap)

        plt.xticks(
            np.arange(0.5, len(corr_df.columns) + 0.5),
            labels=[var.variable_name for var in variable_set],
            rotation=45,
            horizontalalignment="right",
            verticalalignment="top",
        )
        plt.yticks(
            np.arange(0.5, len(corr_df.columns) + 0.5),
            labels=[var.variable_name for var in variable_set],
            rotation=0,
            verticalalignment="bottom",
        )

        export(fig=fig, filename=file_name, target_dir=self.target_dir, close_figure=True)

        return file_name, file_path

    def plot_validation_variables(
        self,
        df: pd.DataFrame,
        working_point: Optional[float] = None,
        mva_output_latex_name: Optional[str] = None,
        base_name: Optional[str] = None,
        dry_run: bool = False,
    ) -> List[Tuple[str, PathType]]:
        if not dry_run:
            signal_mask, background_mask = self.get_sig_bkg_masks(df=df)
            sig_df = df.loc[signal_mask]
            bkg_df = df.loc[background_mask]
        else:
            sig_df, bkg_df = pd.DataFrame(), pd.DataFrame()

        if mva_output_latex_name is None:
            mva_output_latex_label = LatexObsNames.mva_default_output
        else:
            mva_output_latex_label = mva_output_latex_name

        output_infos = []  # type: List[Tuple[str, PathType]]

        for var in self._validation_variables:
            name = f"{base_name}_validation_plot_" if base_name else ""
            if working_point:
                file_name = f"{name}validation_variable_distribution_{var.df_label}_mva_output_cut"
            else:
                file_name = f"{name}validation_variable_distribution_{var.df_label}"
            file_path = os.path.join(self.target_dir, f"{file_name}.pdf")
            output_infos.append((file_name, file_path))
            if dry_run:
                continue

            if working_point:
                sig_df.query(f"{self._mva_output_name} > {working_point}", inplace=True)
                bkg_df.query(f"{self._mva_output_name} > {working_point}", inplace=True)

            shp = SimpleHistogramPlot(var)
            shp.add_component(label="Background", data=bkg_df, hist_type="stepfilled", color=KITColors.kit_red)
            shp.add_component(label="Signal", data=sig_df, hist_type="stepfilled", color=KITColors.kit_blue, alpha=0.8)

            fig, ax = plt.subplots(nrows=1, ncols=1, figsize=self.figure_size, dpi=self.figure_dpi)
            shp.plot_on(ax, normed=True, y_axis_scale=1.1)

            if working_point is not None:
                ax.set_title(rf"Working Point, {mva_output_latex_label}$ > {working_point}$", loc="right")

            export(fig=fig, filename=file_name, target_dir=self.target_dir, close_figure=True)

        return output_infos

    def run_validation(
        self,
        base_name: str,
        working_point: float,
        mva_output_latex_name: str,
        dry_run: bool = False,
    ) -> Tuple[List[Tuple[str, PathType]], float, float]:
        output_info_list = []  # type: List[Tuple[str, PathType]]
        if not dry_run:
            os.makedirs(self.target_dir, exist_ok=True)

        train_df, test_df = self.load_train_and_test_dfs(dry_run=dry_run)  # type: pd.DataFrame, pd.DataFrame

        classifier = None  # type: Optional[MVABaseClass]
        if not dry_run:
            train_var_list = self.load_cs_vars(path=self._bdt_info_var_list_path)  # type: List[str]
            assert all(v in self.train_variables_column_names for v in train_var_list), [
                v for v in train_var_list if v not in self.train_variables_column_names
            ]

            _classifier = self._mva_class.load_classifier(
                weight_file_path=self._bdt_info_weights_path,
                var_file_path=self._bdt_info_var_list_path,
                target_var=self._target_var,
            )  # type: MVABaseClass

            for df in [train_df, test_df]:
                _classifier.apply(data=df, output_column=self.mva_output_column_name)
            classifier = _classifier

        # Input variables
        var_output_info_list = self.plot_training_vars(
            df=train_df,
            title=None,
            base_name=base_name,
            dry_run=dry_run,
        )
        output_info_list.extend(var_output_info_list)

        # ROC curves
        test_roc_key, test_roc_path, test_roc_auc = self.plot_roc_curve(
            df=test_df,
            working_point=working_point,
            mva_output_latex_name=mva_output_latex_name,
            appendix=None,
            title=None,
            base_name=base_name,
            dry_run=dry_run,
        )
        output_info_list.append((test_roc_key, test_roc_path))
        train_roc_key, train_roc_path, train_roc_auc = self.plot_roc_curve(
            df=train_df,
            working_point=working_point,
            mva_output_latex_name=mva_output_latex_name,
            appendix="on_train_data",
            title=None,
            base_name=base_name,
            dry_run=dry_run,
        )
        output_info_list.append((train_roc_key, train_roc_path))

        for alt_wp_setup in self.alternative_working_point_setups:
            alt_test_roc_key, alt_test_roc_path, _ = self.plot_roc_curve(
                df=test_df,
                working_point=alt_wp_setup.working_point,
                mva_output_latex_name=mva_output_latex_name,
                appendix=alt_wp_setup.appendix,
                title=None,
                base_name=base_name,
                dry_run=dry_run,
            )
            output_info_list.append((alt_test_roc_key, alt_test_roc_path))
            alt_train_roc_key, alt_train_roc_path, _ = self.plot_roc_curve(
                df=train_df,
                working_point=alt_wp_setup.working_point,
                mva_output_latex_name=mva_output_latex_name,
                appendix=alt_wp_setup.appendix_for_train_data,
                title=None,
                base_name=base_name,
                dry_run=dry_run,
            )
            output_info_list.append((alt_train_roc_key, alt_train_roc_path))

        # Feature Importance
        importance_output_info_list = self.evaluate_feature_importance(
            base_name=base_name,
            train_df=train_df,
            test_df=test_df,
            mva_classifier=classifier,
            dry_run=dry_run,
        )
        output_info_list.extend(importance_output_info_list)

        # Output distribution
        mva_output_key, mva_output_path = self.plot_mva_output(
            mva_name=mva_output_latex_name,
            train_df=train_df,
            test_df=test_df,
            appendix=None,
            title=None,
            base_name=base_name,
            dry_run=dry_run,
        )
        output_info_list.append((mva_output_key, mva_output_path))

        # Correlation plot
        correlation_key, correlation_path = self.plot_variable_correlation(
            df=train_df,
            base_name=base_name,
            dry_run=dry_run,
        )
        output_info_list.append((correlation_key, correlation_path))

        # Validation variables plots
        validation_outputs = self.plot_validation_variables(
            df=test_df,
            base_name=base_name,
            dry_run=dry_run,
        )
        output_info_list.extend(validation_outputs)

        # Validation variables plots with cut on MVA output
        validation_outputs = self.plot_validation_variables(
            df=test_df,
            working_point=working_point,
            base_name=base_name,
            dry_run=dry_run,
        )
        output_info_list.extend(validation_outputs)

        return output_info_list, test_roc_auc, train_roc_auc
