from typing import Dict, List, Optional
from uncertainties import UFloat

from rdstar.branching_fractions import b_meson_bf_calculation as bf_calc

from rdstar.offline_analysis.dss_fitting.setup import dss_bf_fit_cases
from rdstar.offline_analysis.dss_fitting.utilities import dss_bf_fitting_utils

__all__ = [
    "BFFittingSetup",
    "BFFittingSetupCollection",
]


class BFFittingSetup:
    def __init__(
        self,
        name: str,
        fit_case_name: str,
        yield_parameters: List[str],
        norm_parameter: str,
        norm_bf: UFloat,
        charged_mixed_factor: Optional[float] = 1.0,
        nominal_bfs: Optional[Dict[str, UFloat]] = None,
    ) -> None:

        self._name = name
        self._fit_case_name = fit_case_name
        self._yield_parameters = yield_parameters
        self._norm_parameter = norm_parameter
        self._norm_bf = norm_bf
        self._charged_mixed_factor = charged_mixed_factor
        self._nominal_bfs = nominal_bfs

    @property
    def name(self) -> str:
        return self._name

    @property
    def fit_case_name(self) -> str:
        return self._fit_case_name

    @property
    def fit_case(self) -> dss_bf_fit_cases.FitCase:
        return dss_bf_fit_cases.BFFitCaseDict.get_fit_case(fit_case_name=self.fit_case_name)

    @property
    def yield_parameters(self) -> List[str]:
        return self._yield_parameters

    @property
    def norm_parameter(self) -> str:
        return self._norm_parameter

    @property
    def norm_bf(self) -> float:
        return self._norm_bf.nominal_value

    @property
    def norm_bf_std_dev(self) -> float:
        return self._norm_bf.std_dev

    @property
    def charged_mixed_factor(self) -> float:
        assert self._charged_mixed_factor is not None
        return self._charged_mixed_factor

    @property
    def nominal_bfs(self) -> Dict[str, UFloat]:
        assert self._nominal_bfs is not None
        return self._nominal_bfs


# Asimov Setups
asimov_charged_B_charged_pi_bf_fitting_setup = BFFittingSetup(
    name="asimov_charged_B_charged_pi_bf_fitting_setup",
    fit_case_name="asimov_charged_B_charged_pi_fit_case",
    yield_parameters=["ChargedBDzeroStarPiP", "ChargedBDonePrimePiP", "ChargedBDonePiP", "ChargedBDtwoStarPiP"],
    norm_parameter="NormBchargedDs",
    norm_bf=bf_calc.bf_bp_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

asimov_charged_B_neutral_pi_bf_fitting_setup = BFFittingSetup(
    name="asimov_charged_B_neutral_pi_bf_fitting_setup",
    fit_case_name="asimov_charged_B_neutral_pi_fit_case",
    yield_parameters=["ChargedBDzeroStarPiZ", "ChargedBDonePrimePiZ", "ChargedBDonePiZ", "ChargedBDtwoStarPiZ"],
    norm_parameter="NormBchargedDs",
    norm_bf=bf_calc.bf_bp_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

asimov_neutral_B_charged_pi_bf_fitting_setup = BFFittingSetup(
    name="asimov_neutral_B_charged_pi_bf_fitting_setup",
    fit_case_name="asimov_neutral_B_charged_pi_fit_case",
    yield_parameters=["NeutralBDzeroStarPiP", "NeutralBDonePrimePiP", "NeutralBDonePiP", "NeutralBDtwoStarPiP"],
    norm_parameter="NormBzeroDs",
    norm_bf=bf_calc.bf_b0_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

asimov_neutral_B_neutral_pi_bf_fitting_setup = BFFittingSetup(
    name="asimov_neutral_B_neutral_pi_bf_fitting_setup",
    fit_case_name="asimov_neutral_B_neutral_pi_fit_case",
    yield_parameters=["NeutralBDzeroStarPiZ", "NeutralBDonePrimePiZ", "NeutralBDonePiZ", "NeutralBDtwoStarPiZ"],
    norm_parameter="NormBzeroDs",
    norm_bf=bf_calc.bf_b0_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

# Decoupled Asimov Setups
asimov_decoupled_charged_B_charged_pi_bf_fitting_setup_via_DS_DSZ = BFFittingSetup(
    name="asimov_decoupled_charged_B_charged_pi_bf_fitting_setup_via_DS_DSZ",
    fit_case_name="asimov_decoupled_charged_B_charged_pi_fit_case",
    yield_parameters=["ChargedBDzeroStarPiP", "ChargedBDonePrimePiP", "ChargedBDonePiP", "ChargedBDtwoStarPiP"],
    norm_parameter="NormBchargedDs_DSZ",
    norm_bf=bf_calc.bf_bp_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

asimov_decoupled_charged_B_neutral_pi_bf_fitting_setup_via_DS_DSZ = BFFittingSetup(
    name="asimov_decoupled_charged_B_neutral_pi_bf_fitting_setup_via_DS_DSZ",
    fit_case_name="asimov_decoupled_charged_B_neutral_pi_fit_case",
    yield_parameters=["ChargedBDzeroStarPiZ", "ChargedBDonePrimePiZ", "ChargedBDonePiZ", "ChargedBDtwoStarPiZ"],
    norm_parameter="NormBchargedDs_DSZ",
    norm_bf=bf_calc.bf_bp_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

asimov_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_DS_DSM = BFFittingSetup(
    name="asimov_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_DS_DSM",
    fit_case_name="asimov_decoupled_neutral_B_charged_pi_fit_case",
    yield_parameters=["NeutralBDzeroStarPiP", "NeutralBDonePrimePiP", "NeutralBDonePiP", "NeutralBDtwoStarPiP"],
    norm_parameter="NormBzeroDs_DSM",
    norm_bf=bf_calc.bf_b0_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

asimov_decoupled_neutral_B_neutral_pi_bf_fitting_setup_via_DS_DSM = BFFittingSetup(
    name="asimov_decoupled_neutral_B_neutral_pi_bf_fitting_setup_via_DS_DSM",
    fit_case_name="asimov_decoupled_neutral_B_neutral_pi_fit_case",
    yield_parameters=["NeutralBDzeroStarPiZ", "NeutralBDonePrimePiZ", "NeutralBDonePiZ", "NeutralBDtwoStarPiZ"],
    norm_parameter="NormBzeroDs_DSM",
    norm_bf=bf_calc.bf_b0_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

asimov_decoupled_charged_B_charged_pi_bf_fitting_setup_via_DS_DZ = BFFittingSetup(
    name="asimov_decoupled_charged_B_charged_pi_bf_fitting_setup_via_DS_DZ",
    fit_case_name="asimov_decoupled_charged_B_charged_pi_fit_case",
    yield_parameters=["ChargedBDzeroStarPiP", "ChargedBDonePrimePiP", "ChargedBDonePiP", "ChargedBDtwoStarPiP"],
    norm_parameter="NormBchargedDs_DZ",
    norm_bf=bf_calc.bf_bp_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

asimov_decoupled_charged_B_neutral_pi_bf_fitting_setup_via_DS_DZ = BFFittingSetup(
    name="asimov_decoupled_charged_B_neutral_pi_bf_fitting_setup_via_DS_DZ",
    fit_case_name="asimov_decoupled_charged_B_neutral_pi_fit_case",
    yield_parameters=["ChargedBDzeroStarPiZ", "ChargedBDonePrimePiZ", "ChargedBDonePiZ", "ChargedBDtwoStarPiZ"],
    norm_parameter="NormBchargedDs_DZ",
    norm_bf=bf_calc.bf_bp_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

asimov_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_DS_DM = BFFittingSetup(
    name="asimov_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_DS_DM",
    fit_case_name="asimov_decoupled_neutral_B_charged_pi_fit_case",
    yield_parameters=["NeutralBDzeroStarPiP", "NeutralBDonePrimePiP", "NeutralBDonePiP", "NeutralBDtwoStarPiP"],
    norm_parameter="NormBzeroDs_DM",
    norm_bf=bf_calc.bf_b0_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

asimov_decoupled_neutral_B_neutral_pi_bf_fitting_setup_via_DS_DM = BFFittingSetup(
    name="asimov_decoupled_neutral_B_neutral_pi_bf_fitting_setup_via_DS_DM",
    fit_case_name="asimov_decoupled_neutral_B_neutral_pi_fit_case",
    yield_parameters=["NeutralBDzeroStarPiZ", "NeutralBDonePrimePiZ", "NeutralBDonePiZ", "NeutralBDtwoStarPiZ"],
    norm_parameter="NormBzeroDs_DM",
    norm_bf=bf_calc.bf_b0_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

asimov_decoupled_charged_B_charged_pi_bf_fitting_setup_via_D_DZ = BFFittingSetup(
    name="asimov_decoupled_charged_B_charged_pi_bf_fitting_setup_via_D_DZ",
    fit_case_name="asimov_decoupled_charged_B_charged_pi_fit_case",
    yield_parameters=["ChargedBDzeroStarPiP", "ChargedBDonePrimePiP", "ChargedBDonePiP", "ChargedBDtwoStarPiP"],
    norm_parameter="NormBchargedD_DZ",
    norm_bf=bf_calc.bf_bp_to_d_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

asimov_decoupled_charged_B_neutral_pi_bf_fitting_setup_via_D_DZ = BFFittingSetup(
    name="asimov_decoupled_charged_B_neutral_pi_bf_fitting_setup_via_D_DZ",
    fit_case_name="asimov_decoupled_charged_B_neutral_pi_fit_case",
    yield_parameters=["ChargedBDzeroStarPiZ", "ChargedBDonePrimePiZ", "ChargedBDonePiZ", "ChargedBDtwoStarPiZ"],
    norm_parameter="NormBchargedD_DZ",
    norm_bf=bf_calc.bf_bp_to_d_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

asimov_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_D_DM = BFFittingSetup(
    name="asimov_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_D_DM",
    fit_case_name="asimov_decoupled_neutral_B_charged_pi_fit_case",
    yield_parameters=["NeutralBDzeroStarPiP", "NeutralBDonePrimePiP", "NeutralBDonePiP", "NeutralBDtwoStarPiP"],
    norm_parameter="NormBzeroD_DM",
    norm_bf=bf_calc.bf_b0_to_d_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

asimov_decoupled_neutral_B_neutral_pi_bf_fitting_setup_via_D_DM = BFFittingSetup(
    name="asimov_decoupled_neutral_B_neutral_pi_bf_fitting_setup_via_D_DM",
    fit_case_name="asimov_decoupled_neutral_B_neutral_pi_fit_case",
    yield_parameters=["NeutralBDzeroStarPiZ", "NeutralBDonePrimePiZ", "NeutralBDonePiZ", "NeutralBDtwoStarPiZ"],
    norm_parameter="NormBzeroD_DM",
    norm_bf=bf_calc.bf_b0_to_d_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

# Asimov Setups to fit Normalisation Branching Fractions
asimov_charged_B_charged_pi_D_bf_fitting_setup = BFFittingSetup(
    name="asimov_charged_B_charged_pi_D_bf_fitting_setup",
    fit_case_name="asimov_charged_B_charged_pi_fit_case",
    yield_parameters=["NormBchargedD"],
    norm_parameter="NormBchargedDs",
    norm_bf=bf_calc.bf_bp_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

asimov_charged_B_neutral_pi_D_bf_fitting_setup = BFFittingSetup(
    name="asimov_charged_B_neutral_pi_D_bf_fitting_setup",
    fit_case_name="asimov_charged_B_neutral_pi_fit_case",
    yield_parameters=["NormBchargedD"],
    norm_parameter="NormBchargedDs",
    norm_bf=bf_calc.bf_bp_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

asimov_neutral_B_charged_pi_D_bf_fitting_setup = BFFittingSetup(
    name="asimov_neutral_B_charged_pi_D_bf_fitting_setup",
    fit_case_name="asimov_neutral_B_charged_pi_fit_case",
    yield_parameters=["NormBzeroD"],
    norm_parameter="NormBzeroDs",
    norm_bf=bf_calc.bf_b0_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

asimov_neutral_B_neutral_pi_D_bf_fitting_setup = BFFittingSetup(
    name="asimov_neutral_B_neutral_pi_D_bf_fitting_setup",
    fit_case_name="asimov_neutral_B_neutral_pi_fit_case",
    yield_parameters=["NormBzeroD"],
    norm_parameter="NormBzeroDs",
    norm_bf=bf_calc.bf_b0_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

# Alternative Asimov Setups to fit Normalisation Branching Fractions
asimov_alternative_charged_B_charged_pi_D_bf_fitting_setup = BFFittingSetup(
    name="asimov_alternative_charged_B_charged_pi_D_bf_fitting_setup",
    fit_case_name="asimov_alternative_charged_B_charged_pi_fit_case",
    yield_parameters=["NormBchargedD"],
    norm_parameter="NormBchargedDs",
    norm_bf=bf_calc.bf_bp_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

# Mbc Sideband Setups
# todo: No neutral pion cases here since the signal modes can't be fitted in the Mbc Sideband in these modes
mbc_charged_B_charged_pi_bf_fitting_setup = BFFittingSetup(
    name="mbc_charged_B_charged_pi_bf_fitting_setup",
    fit_case_name="mbc_charged_B_charged_pi_fit_case",
    yield_parameters=["ChargedBDzeroStarPiP", "ChargedBDonePrimePiP", "ChargedBDonePiP", "ChargedBDtwoStarPiP"],
    norm_parameter="NormBchargedDs",
    norm_bf=bf_calc.bf_bp_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

mbc_neutral_B_charged_pi_bf_fitting_setup = BFFittingSetup(
    name="mbc_neutral_B_charged_pi_bf_fitting_setup",
    fit_case_name="mbc_neutral_B_charged_pi_fit_case",
    yield_parameters=["NeutralBDzeroStarPiP", "NeutralBDonePrimePiP", "NeutralBDonePiP", "NeutralBDtwoStarPiP"],
    norm_parameter="NormBzeroDs",
    norm_bf=bf_calc.bf_b0_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

# Decoupled Mbc Sideband Setups
# todo: No neutral pion cases here since the signal modes can't be fitted in the Mbc Sideband in these modes
mbc_decoupled_charged_B_charged_pi_bf_fitting_setup_via_DS_DSZ = BFFittingSetup(
    name="mbc_decoupled_charged_B_charged_pi_bf_fitting_setup_via_DS_DSZ",
    fit_case_name="mbc_decoupled_charged_B_charged_pi_fit_case",
    yield_parameters=["ChargedBDzeroStarPiP", "ChargedBDonePrimePiP", "ChargedBDonePiP", "ChargedBDtwoStarPiP"],
    norm_parameter="NormBchargedDs_DSZ",
    norm_bf=bf_calc.bf_bp_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

mbc_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_DS_DSM = BFFittingSetup(
    name="mbc_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_DS_DSM",
    fit_case_name="mbc_decoupled_neutral_B_charged_pi_fit_case",
    yield_parameters=["NeutralBDzeroStarPiP", "NeutralBDonePrimePiP", "NeutralBDonePiP", "NeutralBDtwoStarPiP"],
    norm_parameter="NormBzeroDs_DSM",
    norm_bf=bf_calc.bf_b0_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

mbc_decoupled_charged_B_charged_pi_bf_fitting_setup_via_DS_DZ = BFFittingSetup(
    name="mbc_decoupled_charged_B_charged_pi_bf_fitting_setup_via_DS_DZ",
    fit_case_name="mbc_decoupled_charged_B_charged_pi_fit_case",
    yield_parameters=["ChargedBDzeroStarPiP", "ChargedBDonePrimePiP", "ChargedBDonePiP", "ChargedBDtwoStarPiP"],
    norm_parameter="NormBchargedDs_DZ",
    norm_bf=bf_calc.bf_bp_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

mbc_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_DS_DM = BFFittingSetup(
    name="mbc_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_DS_DM",
    fit_case_name="mbc_decoupled_neutral_B_charged_pi_fit_case",
    yield_parameters=["NeutralBDzeroStarPiP", "NeutralBDonePrimePiP", "NeutralBDonePiP", "NeutralBDtwoStarPiP"],
    norm_parameter="NormBzeroDs_DM",
    norm_bf=bf_calc.bf_b0_to_d_st_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

mbc_decoupled_charged_B_charged_pi_bf_fitting_setup_via_D_DZ = BFFittingSetup(
    name="mbc_decoupled_charged_B_charged_pi_bf_fitting_setup_via_D_DZ",
    fit_case_name="mbc_decoupled_charged_B_charged_pi_fit_case",
    yield_parameters=["ChargedBDzeroStarPiP", "ChargedBDonePrimePiP", "ChargedBDonePiP", "ChargedBDtwoStarPiP"],
    norm_parameter="NormBchargedD_DZ",
    norm_bf=bf_calc.bf_bp_to_d_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

mbc_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_D_DM = BFFittingSetup(
    name="mbc_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_D_DM",
    fit_case_name="mbc_decoupled_neutral_B_charged_pi_fit_case",
    yield_parameters=["NeutralBDzeroStarPiP", "NeutralBDonePrimePiP", "NeutralBDonePiP", "NeutralBDtwoStarPiP"],
    norm_parameter="NormBzeroD_DM",
    norm_bf=bf_calc.bf_b0_to_d_l_nu_avg,
    nominal_bfs=dss_bf_fitting_utils.sig_dss_nominal_bfs,
)

# Mbc Sideband Setups to fit Normalisation Branching Fractions
mbc_charged_B_charged_pi_D_bf_fitting_setup = BFFittingSetup(
    name="mbc_charged_B_charged_pi_D_bf_fitting_setup",
    fit_case_name="mbc_charged_B_charged_pi_fit_case",
    yield_parameters=["NormBchargedD"],
    norm_parameter="NormBchargedDs",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBchargedDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

mbc_charged_B_neutral_pi_D_bf_fitting_setup = BFFittingSetup(
    name="mbc_charged_B_neutral_pi_D_bf_fitting_setup",
    fit_case_name="mbc_charged_B_neutral_pi_fit_case",
    yield_parameters=["NormBchargedD"],
    norm_parameter="NormBchargedDs",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBchargedDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

mbc_neutral_B_charged_pi_D_bf_fitting_setup = BFFittingSetup(
    name="mbc_neutral_B_charged_pi_D_bf_fitting_setup",
    fit_case_name="mbc_neutral_B_charged_pi_fit_case",
    yield_parameters=["NormBzeroD"],
    norm_parameter="NormBzeroDs",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBzeroDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

mbc_neutral_B_neutral_pi_D_bf_fitting_setup = BFFittingSetup(
    name="mbc_neutral_B_neutral_pi_D_bf_fitting_setup",
    fit_case_name="mbc_neutral_B_neutral_pi_fit_case",
    yield_parameters=["NormBzeroD"],
    norm_parameter="NormBzeroDs",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBzeroDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

# Decoupled Mbc Sideband Setups to fit the Normalisation Branching Fractions
mbc_decoupled_charged_B_charged_pi_D_bf_fitting_setup_via_DS_DSZ = BFFittingSetup(
    name="mbc_decoupled_charged_B_charged_pi_D_bf_fitting_setup_via_DS_DSZ",
    fit_case_name="mbc_decoupled_charged_B_charged_pi_fit_case",
    yield_parameters=["NormBchargedD_DSZ"],
    norm_parameter="NormBchargedDs_DSZ",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBchargedDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

mbc_decoupled_neutral_B_charged_pi_D_bf_fitting_setup_via_DS_DSM = BFFittingSetup(
    name="mbc_decoupled_neutral_B_charged_pi_D_bf_fitting_setup_via_DS_DSM",
    fit_case_name="mbc_decoupled_neutral_B_charged_pi_fit_case",
    yield_parameters=["NormBzeroD_DSM"],
    norm_parameter="NormBzeroDs_DSM",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBzeroDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

mbc_decoupled_charged_B_charged_pi_D_bf_fitting_setup_via_DS_DZ = BFFittingSetup(
    name="mbc_decoupled_charged_B_charged_pi_D_bf_fitting_setup_via_DS_DZ",
    fit_case_name="mbc_decoupled_charged_B_charged_pi_fit_case",
    yield_parameters=["NormBchargedD_DZ"],
    norm_parameter="NormBchargedDs_DZ",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBchargedDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

mbc_decoupled_neutral_B_charged_pi_D_bf_fitting_setup_via_DS_DM = BFFittingSetup(
    name="mbc_decoupled_neutral_B_charged_pi_D_bf_fitting_setup_via_DS_DM",
    fit_case_name="mbc_decoupled_neutral_B_charged_pi_fit_case",
    yield_parameters=["NormBzeroD_DM"],
    norm_parameter="NormBzeroDs_DM",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBzeroDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

mbc_decoupled_charged_B_charged_pi_D_bf_fitting_setup_via_D_DZ = BFFittingSetup(
    name="mbc_decoupled_charged_B_charged_pi_D_bf_fitting_setup_via_D_DZ",
    fit_case_name="mbc_decoupled_charged_B_charged_pi_fit_case",
    yield_parameters=["ChargedBDzeroStarPiP", "ChargedBDonePrimePiP", "ChargedBDonePiP", "ChargedBDtwoStarPiP"],
    norm_parameter="NormBchargedD_DZ",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBzeroDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

mbc_decoupled_neutral_B_charged_pi_D_bf_fitting_setup_via_D_DM = BFFittingSetup(
    name="mbc_decoupled_neutral_B_charged_pi_D_bf_fitting_setup_via_D_DM",
    fit_case_name="mbc_decoupled_neutral_B_charged_pi_fit_case",
    yield_parameters=["NeutralBDzeroStarPiP", "NeutralBDonePrimePiP", "NeutralBDonePiP", "NeutralBDtwoStarPiP"],
    norm_parameter="NormBzeroD_DM",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBzeroDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

mbc_decoupled_charged_B_neutral_pi_D_bf_fitting_setup_via_DS_DSZ = BFFittingSetup(
    name="mbc_decoupled_charged_B_neutral_pi_D_bf_fitting_setup_via_DS_DSZ",
    fit_case_name="mbc_decoupled_charged_B_neutral_pi_fit_case",
    yield_parameters=["NormBchargedD_DSZ"],
    norm_parameter="NormBchargedDs_DSZ",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBchargedDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

mbc_decoupled_neutral_B_neutral_pi_D_bf_fitting_setup_via_DS_DSM = BFFittingSetup(
    name="mbc_decoupled_neutral_B_neutral_pi_D_bf_fitting_setup_via_DS_DSM",
    fit_case_name="mbc_decoupled_neutral_B_neutral_pi_fit_case",
    yield_parameters=["NormBzeroD_DSM"],
    norm_parameter="NormBzeroDs_DSM",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBzeroDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

mbc_decoupled_charged_B_neutral_pi_D_bf_fitting_setup_via_DS_DZ = BFFittingSetup(
    name="mbc_decoupled_charged_B_neutral_pi_D_bf_fitting_setup_via_DS_DZ",
    fit_case_name="mbc_decoupled_charged_B_neutral_pi_fit_case",
    yield_parameters=["NormBchargedD_DZ"],
    norm_parameter="NormBchargedDs_DZ",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBchargedDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

mbc_decoupled_neutral_B_neutral_pi_D_bf_fitting_setup_via_DS_DM = BFFittingSetup(
    name="mbc_decoupled_neutral_B_neutral_pi_D_bf_fitting_setup_via_DS_DM",
    fit_case_name="mbc_decoupled_neutral_B_neutral_pi_fit_case",
    yield_parameters=["NormBzeroD_DM"],
    norm_parameter="NormBzeroDs_DM",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBzeroDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

mbc_decoupled_charged_B_neutral_pi_D_bf_fitting_setup_via_D_DZ = BFFittingSetup(
    name="mbc_decoupled_charged_B_neutral_pi_D_bf_fitting_setup_via_D_DZ",
    fit_case_name="mbc_decoupled_charged_B_neutral_pi_fit_case",
    yield_parameters=["ChargedBDzeroStarPiZ", "ChargedBDonePrimePiZ", "ChargedBDonePiZ", "ChargedBDtwoStarPiZ"],
    norm_parameter="NormBchargedD_DZ",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBzeroDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

mbc_decoupled_neutral_B_neutral_pi_D_bf_fitting_setup_via_D_DM = BFFittingSetup(
    name="mbc_decoupled_neutral_B_neutral_pi_D_bf_fitting_setup_via_D_DM",
    fit_case_name="mbc_decoupled_neutral_B_neutral_pi_fit_case",
    yield_parameters=["NeutralBDzeroStarPiZ", "NeutralBDonePrimePiZ", "NeutralBDonePiZ", "NeutralBDtwoStarPiZ"],
    norm_parameter="NormBzeroD_DM",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBzeroDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

# Alternative Mbc Sideband Setups to fit Normalisation Branching Fractions
mbc_alternative_charged_B_charged_pi_D_bf_fitting_setup = BFFittingSetup(
    name="mbc_alternative_charged_B_charged_pi_D_bf_fitting_setup",
    fit_case_name="mbc_alternative_charged_B_charged_pi_fit_case",
    yield_parameters=["NormBchargedD"],
    norm_parameter="NormBchargedDs",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBchargedDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

mbc_alternative_decoupled_charged_B_charged_pi_D_bf_fitting_setup = BFFittingSetup(
    name="mbc_alternative_decoupled_charged_B_charged_pi_D_bf_fitting_setup",
    fit_case_name="mbc_alternative_decoupled_charged_B_charged_pi_fit_case",
    yield_parameters=["NormBchargedD_DZ"],
    norm_parameter="NormBchargedDs_DSZ",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBchargedDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)


# Asimov Normalization Fits
asimov_normalization_coupled_charged_B_bf_fitting_setup = BFFittingSetup(
    name="asimov_normalization_coupled_charged_B_bf_fitting_setup",
    fit_case_name=dss_bf_fit_cases.BFFitCaseDict.asimov_normalization_fit_coupled_charged_B_fit_case.name,
    yield_parameters=["NormBchargedD"],
    norm_parameter="NormBchargedDs",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBchargedDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

asimov_normalization_coupled_neutral_B_bf_fitting_setup = BFFittingSetup(
    name="asimov_normalization_coupled_neutral_B_bf_fitting_setup",
    fit_case_name=dss_bf_fit_cases.BFFitCaseDict.asimov_normalization_fit_coupled_neutral_B_fit_case.name,
    yield_parameters=["NormBzeroD"],
    norm_parameter="NormBzeroDs",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBzeroDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

asimov_normalization_decoupled_charged_B_bf_fitting_setup = BFFittingSetup(
    name="asimov_normalization_decoupled_charged_B_bf_fitting_setup",
    fit_case_name=dss_bf_fit_cases.BFFitCaseDict.asimov_normalization_fit_decoupled_charged_B_fit_case.name,
    yield_parameters=["NormBchargedD_DZ"],
    norm_parameter="NormBchargedDs_DSZ",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBchargedDs_DSZ"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

asimov_normalization_decoupled_neutral_B_bf_fitting_setup = BFFittingSetup(
    name="asimov_normalization_decoupled_neutral_B_bf_fitting_setup",
    fit_case_name=dss_bf_fit_cases.BFFitCaseDict.asimov_normalization_fit_decoupled_neutral_B_fit_case.name,
    yield_parameters=["NormBzeroD_DM"],
    norm_parameter="NormBzeroDs_DSM",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBzeroDs_DSM"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)


# Normalization Fits on Data
normalization_coupled_charged_B_bf_fitting_setup = BFFittingSetup(
    name="normalization_coupled_charged_B_bf_fitting_setup",
    fit_case_name=dss_bf_fit_cases.BFFitCaseDict.normalization_fit_coupled_charged_B_fit_case.name,
    yield_parameters=["NormBchargedD"],
    norm_parameter="NormBchargedDs",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBchargedDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

normalization_coupled_neutral_B_bf_fitting_setup = BFFittingSetup(
    name="normalization_coupled_neutral_B_bf_fitting_setup",
    fit_case_name=dss_bf_fit_cases.BFFitCaseDict.normalization_fit_coupled_neutral_B_fit_case.name,
    yield_parameters=["NormBzeroD"],
    norm_parameter="NormBzeroDs",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBzeroDs"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

normalization_decoupled_charged_B_bf_fitting_setup = BFFittingSetup(
    name="normalization_decoupled_charged_B_bf_fitting_setup",
    fit_case_name=dss_bf_fit_cases.BFFitCaseDict.normalization_fit_decoupled_charged_B_fit_case.name,
    yield_parameters=["NormBchargedD_DZ"],
    norm_parameter="NormBchargedDs_DSZ",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBchargedDs_DSZ"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)

normalization_decoupled_neutral_B_bf_fitting_setup = BFFittingSetup(
    name="normalization_decoupled_neutral_B_bf_fitting_setup",
    fit_case_name=dss_bf_fit_cases.BFFitCaseDict.normalization_fit_decoupled_neutral_B_fit_case.name,
    yield_parameters=["NormBzeroD_DM"],
    norm_parameter="NormBzeroDs_DSM",
    norm_bf=dss_bf_fitting_utils.current_pdg_bfs["NormBzeroDs_DSM"],
    nominal_bfs=dss_bf_fitting_utils.current_pdg_bfs,
)


class BFFittingSetupCollection:
    available_setups = {
        # Asimov Setups
        asimov_charged_B_charged_pi_bf_fitting_setup.name: asimov_charged_B_charged_pi_bf_fitting_setup,
        asimov_charged_B_neutral_pi_bf_fitting_setup.name: asimov_charged_B_neutral_pi_bf_fitting_setup,
        asimov_neutral_B_charged_pi_bf_fitting_setup.name: asimov_neutral_B_charged_pi_bf_fitting_setup,
        asimov_neutral_B_neutral_pi_bf_fitting_setup.name: asimov_neutral_B_neutral_pi_bf_fitting_setup,
        #
        # Decoupled Asimov Setups
        asimov_decoupled_charged_B_charged_pi_bf_fitting_setup_via_DS_DSZ.name: asimov_decoupled_charged_B_charged_pi_bf_fitting_setup_via_DS_DSZ,
        asimov_decoupled_charged_B_neutral_pi_bf_fitting_setup_via_DS_DSZ.name: asimov_decoupled_charged_B_neutral_pi_bf_fitting_setup_via_DS_DSZ,
        asimov_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_DS_DSM.name: asimov_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_DS_DSM,
        asimov_decoupled_neutral_B_neutral_pi_bf_fitting_setup_via_DS_DSM.name: asimov_decoupled_neutral_B_neutral_pi_bf_fitting_setup_via_DS_DSM,
        asimov_decoupled_charged_B_charged_pi_bf_fitting_setup_via_DS_DZ.name: asimov_decoupled_charged_B_charged_pi_bf_fitting_setup_via_DS_DZ,
        asimov_decoupled_charged_B_neutral_pi_bf_fitting_setup_via_DS_DZ.name: asimov_decoupled_charged_B_neutral_pi_bf_fitting_setup_via_DS_DZ,
        asimov_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_DS_DM.name: asimov_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_DS_DM,
        asimov_decoupled_neutral_B_neutral_pi_bf_fitting_setup_via_DS_DM.name: asimov_decoupled_neutral_B_neutral_pi_bf_fitting_setup_via_DS_DM,
        asimov_decoupled_charged_B_charged_pi_bf_fitting_setup_via_D_DZ.name: asimov_decoupled_charged_B_charged_pi_bf_fitting_setup_via_D_DZ,
        asimov_decoupled_charged_B_neutral_pi_bf_fitting_setup_via_D_DZ.name: asimov_decoupled_charged_B_neutral_pi_bf_fitting_setup_via_D_DZ,
        asimov_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_D_DM.name: asimov_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_D_DM,
        asimov_decoupled_neutral_B_neutral_pi_bf_fitting_setup_via_D_DM.name: asimov_decoupled_neutral_B_neutral_pi_bf_fitting_setup_via_D_DM,
        #
        # Asimov Setups to fit Normalisation Branching Fractions
        asimov_charged_B_charged_pi_D_bf_fitting_setup.name: asimov_charged_B_charged_pi_D_bf_fitting_setup,
        asimov_charged_B_neutral_pi_D_bf_fitting_setup.name: asimov_charged_B_neutral_pi_D_bf_fitting_setup,
        asimov_neutral_B_charged_pi_D_bf_fitting_setup.name: asimov_neutral_B_charged_pi_D_bf_fitting_setup,
        asimov_neutral_B_neutral_pi_D_bf_fitting_setup.name: asimov_neutral_B_neutral_pi_D_bf_fitting_setup,
        #
        # Alternative Asimov Setups to fit Normalisation Branching Fractions
        asimov_alternative_charged_B_charged_pi_D_bf_fitting_setup.name: asimov_alternative_charged_B_charged_pi_D_bf_fitting_setup,
        #
        # # Mbc Sideband Setups
        # mbc_charged_B_charged_pi_bf_fitting_setup.name: mbc_charged_B_charged_pi_bf_fitting_setup,
        # mbc_neutral_B_charged_pi_bf_fitting_setup.name: mbc_neutral_B_charged_pi_bf_fitting_setup,
        # mbc_decoupled_charged_B_charged_pi_bf_fitting_setup_via_DS_DSZ.name: mbc_decoupled_charged_B_charged_pi_bf_fitting_setup_via_DS_DSZ,
        # mbc_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_DS_DSM.name: mbc_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_DS_DSM,
        # mbc_decoupled_charged_B_charged_pi_bf_fitting_setup_via_DS_DZ.name: mbc_decoupled_charged_B_charged_pi_bf_fitting_setup_via_DS_DZ,
        # mbc_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_DS_DM.name: mbc_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_DS_DM,
        # mbc_decoupled_charged_B_charged_pi_bf_fitting_setup_via_D_DZ.name: mbc_decoupled_charged_B_charged_pi_bf_fitting_setup_via_D_DZ,
        # mbc_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_D_DM.name: mbc_decoupled_neutral_B_charged_pi_bf_fitting_setup_via_D_DM,
        # #
        # # Mbc Sideband Setups to fit Normalisation Branching Fractions
        # mbc_charged_B_neutral_pi_D_bf_fitting_setup.name: mbc_charged_B_neutral_pi_D_bf_fitting_setup,
        # mbc_charged_B_charged_pi_D_bf_fitting_setup.name: mbc_charged_B_charged_pi_D_bf_fitting_setup,
        # mbc_neutral_B_neutral_pi_D_bf_fitting_setup.name: mbc_neutral_B_neutral_pi_D_bf_fitting_setup,
        # mbc_neutral_B_charged_pi_D_bf_fitting_setup.name: mbc_neutral_B_charged_pi_D_bf_fitting_setup,
        # #
        # # Decoupled Mbc Sideband Setups to fit the Normalisation Branching Fractions
        # mbc_decoupled_charged_B_charged_pi_D_bf_fitting_setup_via_DS_DSZ.name: mbc_decoupled_charged_B_charged_pi_D_bf_fitting_setup_via_DS_DSZ,
        # mbc_decoupled_neutral_B_charged_pi_D_bf_fitting_setup_via_DS_DSM.name: mbc_decoupled_neutral_B_charged_pi_D_bf_fitting_setup_via_DS_DSM,
        # mbc_decoupled_charged_B_charged_pi_D_bf_fitting_setup_via_DS_DZ.name: mbc_decoupled_charged_B_charged_pi_D_bf_fitting_setup_via_DS_DZ,
        # mbc_decoupled_neutral_B_charged_pi_D_bf_fitting_setup_via_DS_DM.name: mbc_decoupled_neutral_B_charged_pi_D_bf_fitting_setup_via_DS_DM,
        # mbc_decoupled_charged_B_neutral_pi_D_bf_fitting_setup_via_DS_DSZ.name: mbc_decoupled_charged_B_neutral_pi_D_bf_fitting_setup_via_DS_DSZ,
        # mbc_decoupled_neutral_B_neutral_pi_D_bf_fitting_setup_via_DS_DSM.name: mbc_decoupled_neutral_B_neutral_pi_D_bf_fitting_setup_via_DS_DSM,
        # mbc_decoupled_charged_B_neutral_pi_D_bf_fitting_setup_via_DS_DZ.name: mbc_decoupled_charged_B_neutral_pi_D_bf_fitting_setup_via_DS_DZ,
        # mbc_decoupled_neutral_B_neutral_pi_D_bf_fitting_setup_via_DS_DM.name: mbc_decoupled_neutral_B_neutral_pi_D_bf_fitting_setup_via_DS_DM,
        # #
        # # Alternative Mbc Sideband Setups to fit Normalisation Branching Fractions
        # mbc_alternative_charged_B_charged_pi_D_bf_fitting_setup.name: mbc_alternative_charged_B_charged_pi_D_bf_fitting_setup,
        # mbc_alternative_decoupled_charged_B_charged_pi_D_bf_fitting_setup.name: mbc_alternative_decoupled_charged_B_charged_pi_D_bf_fitting_setup,
        #
        # Asimov Normalization Fit
        asimov_normalization_coupled_charged_B_bf_fitting_setup.name: asimov_normalization_coupled_charged_B_bf_fitting_setup,
        asimov_normalization_coupled_neutral_B_bf_fitting_setup.name: asimov_normalization_coupled_neutral_B_bf_fitting_setup,
        asimov_normalization_decoupled_charged_B_bf_fitting_setup.name: asimov_normalization_decoupled_charged_B_bf_fitting_setup,
        asimov_normalization_decoupled_neutral_B_bf_fitting_setup.name: asimov_normalization_decoupled_neutral_B_bf_fitting_setup,
        #
        # Normalization Fit on Data
        normalization_coupled_charged_B_bf_fitting_setup.name: normalization_coupled_charged_B_bf_fitting_setup,
        normalization_coupled_neutral_B_bf_fitting_setup.name: normalization_coupled_neutral_B_bf_fitting_setup,
        normalization_decoupled_charged_B_bf_fitting_setup.name: normalization_decoupled_charged_B_bf_fitting_setup,
        normalization_decoupled_neutral_B_bf_fitting_setup.name: normalization_decoupled_neutral_B_bf_fitting_setup,
    }  # type: Dict[str, BFFittingSetup]
