from typing import Dict, Tuple, List

from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames

from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction import MissingMassResolutionCorrector
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction_setups import (
    ResolutionCorrectionSetupManager,
)

from rdstar.offline_analysis.fitting.fit_case import FitCase, FitVariable

from rdstar.offline_analysis.dss_fitting.utilities import dss_bf_fitting_utils
from rdstar.offline_analysis.dss_fitting.utilities import utilities as fitting_utils
from rdstar.offline_analysis.dss_fitting.utilities import dss_data_type_fractions as dtf


__all__ = [
    "BFFitCaseDict",
]


do_toys: bool = True
mva_variable_name: str = "dss_sec_sel_mva_output"


delta_M_for_bf_fitting_hvar = HistVariable(
    df_label=Cols.diff_dstar_d_mass,
    n_bins=60,
    scope=(0.2, 0.9),
    var_name=LatexObsNames.d_st_st_delta_m,
    unit=LatexUnits.gev,
)

m2Recoil_for_bf_fitting_var = HistVariable(
    df_label=MissingMassResolutionCorrector.get_single_stage_corrected_col_name(
        correction_setup_name=ResolutionCorrectionSetupManager.setup_name_to_use,
        use_laplace_smeared_version=False,
    ),
    n_bins=10,
    scope=(-1.2, 0.4),
    var_name=LatexObsNames.fitting_mm2,
    unit=LatexUnits.gev_sq,
)

pl_star_for_bf_fitting_hvar = HistVariable(
    df_label="pStarL",
    n_bins=10,
    scope=(0.1, 2.2),
    var_name=LatexObsNames.fitting_p_l_star,
    unit=LatexUnits.gev,
)


twoD_fit_variable_dict: Dict[str, Tuple[FitVariable, ...]] = {
    "DZPiZ": (
        FitVariable(variable=delta_M_for_bf_fitting_hvar, alternative_binning=30, alternative_scope=(0.1, 0.9)),
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=3, alternative_scope=(-0.4, 0.4)),
    ),
    "DMPiZ": (
        FitVariable(variable=delta_M_for_bf_fitting_hvar, alternative_binning=30, alternative_scope=(0.1, 0.9)),
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=3, alternative_scope=(-0.4, 0.4)),
    ),
    "DSZPiZ": (
        FitVariable(variable=delta_M_for_bf_fitting_hvar, alternative_binning=30, alternative_scope=(0.2, 0.9)),
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=3, alternative_scope=(-0.4, 0.4)),
    ),
    "DSMPiZ": (
        FitVariable(variable=delta_M_for_bf_fitting_hvar, alternative_binning=30, alternative_scope=(0.2, 0.9)),
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=2, alternative_scope=(-0.4, 0.4)),
    ),
    "DZPiM": (
        FitVariable(variable=delta_M_for_bf_fitting_hvar, alternative_binning=40, alternative_scope=(0.15, 0.9)),
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=3, alternative_scope=(-0.4, 0.4)),
    ),
    "DPPiM": (
        FitVariable(variable=delta_M_for_bf_fitting_hvar, alternative_binning=40, alternative_scope=(0.1, 0.9)),
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=3, alternative_scope=(-0.4, 0.4)),
    ),
    "DSZPiM": (
        FitVariable(variable=delta_M_for_bf_fitting_hvar, alternative_binning=40, alternative_scope=(0.2, 0.9)),
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=3, alternative_scope=(-0.4, 0.4)),
    ),
    "DSPPiM": (
        FitVariable(variable=delta_M_for_bf_fitting_hvar, alternative_binning=40, alternative_scope=(0.2, 0.9)),
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=3, alternative_scope=(-0.4, 0.4)),
    ),
    "DZ": (
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=20, alternative_scope=(-1.0, 2.0)),
        FitVariable(variable=pl_star_for_bf_fitting_hvar, alternative_binning=5, alternative_scope=(0.4, 2.0)),
    ),
    "DM": (
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=20, alternative_scope=(-1.0, 2.0)),
        FitVariable(variable=pl_star_for_bf_fitting_hvar, alternative_binning=5, alternative_scope=(0.4, 2.0)),
    ),
    "DSZ": (
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=20, alternative_scope=(-1.0, 2.0)),
        FitVariable(variable=pl_star_for_bf_fitting_hvar, alternative_binning=5, alternative_scope=(0.4, 2.0)),
    ),
    "DSM": (
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=20, alternative_scope=(-1.0, 2.0)),
        FitVariable(variable=pl_star_for_bf_fitting_hvar, alternative_binning=5, alternative_scope=(0.4, 2.0)),
    ),
}

mbc_fit_variable_dict: Dict[str, Tuple[FitVariable, ...]] = {
    "DZPiZ": (
        FitVariable(variable=delta_M_for_bf_fitting_hvar, alternative_binning=15, alternative_scope=(0.1, 0.9)),
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=3, alternative_scope=(-0.4, 0.4)),
    ),
    "DMPiZ": (
        FitVariable(variable=delta_M_for_bf_fitting_hvar, alternative_binning=15, alternative_scope=(0.1, 0.9)),
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=3, alternative_scope=(-0.4, 0.4)),
    ),
    "DSZPiZ": (
        FitVariable(variable=delta_M_for_bf_fitting_hvar, alternative_binning=15, alternative_scope=(0.2, 0.9)),
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=3, alternative_scope=(-0.4, 0.4)),
    ),
    "DSMPiZ": (
        FitVariable(variable=delta_M_for_bf_fitting_hvar, alternative_binning=15, alternative_scope=(0.2, 0.9)),
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=2, alternative_scope=(-0.4, 0.4)),
    ),
    "DZPiM": (
        FitVariable(variable=delta_M_for_bf_fitting_hvar, alternative_binning=20, alternative_scope=(0.15, 0.9)),
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=4, alternative_scope=(-0.67, 0.4)),
    ),
    "DPPiM": (
        FitVariable(variable=delta_M_for_bf_fitting_hvar, alternative_binning=20, alternative_scope=(0.1, 0.9)),
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=3, alternative_scope=(-0.4, 0.4)),
    ),
    "DSZPiM": (
        FitVariable(variable=delta_M_for_bf_fitting_hvar, alternative_binning=20, alternative_scope=(0.2, 0.9)),
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=3, alternative_scope=(-0.4, 0.4)),
    ),
    "DSPPiM": (
        FitVariable(variable=delta_M_for_bf_fitting_hvar, alternative_binning=20, alternative_scope=(0.2, 0.9)),
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=3, alternative_scope=(-0.4, 0.4)),
    ),
    "DZ": (
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=20, alternative_scope=(-1.0, 2.0)),
        FitVariable(variable=pl_star_for_bf_fitting_hvar, alternative_binning=5, alternative_scope=(0.4, 2.0)),
    ),
    "DM": (
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=20, alternative_scope=(-1.0, 2.0)),
        FitVariable(variable=pl_star_for_bf_fitting_hvar, alternative_binning=5, alternative_scope=(0.4, 2.0)),
    ),
    "DSZ": (
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=20, alternative_scope=(-1.0, 2.0)),
        FitVariable(variable=pl_star_for_bf_fitting_hvar, alternative_binning=5, alternative_scope=(0.4, 2.0)),
    ),
    "DSM": (
        FitVariable(variable=m2Recoil_for_bf_fitting_var, alternative_binning=20, alternative_scope=(-1.0, 2.0)),
        FitVariable(variable=pl_star_for_bf_fitting_hvar, alternative_binning=2, alternative_scope=(0.4, 2.0)),
    ),
}


class BFFitCaseDict:

    # Asimov Fit cases
    asimov_charged_B_charged_pi_fit_case = FitCase(
        name="asimov_charged_B_charged_pi_fit_case",
        base_dir_name="Fitting_Asimov_DSS",
        sample_key="",
        output_tag="AsimovFit_DSS_FixedNuisanceParams",
        plot_info_text="Asimov Data",
        fix_nuisance_parameters=True,
        asimov_data=True,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: twoD_fit_variable_dict[rm] for rm in fitting_utils.dss_charged_B_charged_pi_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_charged_B_charged_pi_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.charged_B_charged_pi_template_setup,
        sig_id_to_data_type_mapping=dtf.charged_B_charged_pi_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.charged_B_charged_pi_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.charged_B_charged_pi_sigID_customizing_dict,
            "do_significance": False,
            "do_toys": do_toys,
        },
    )

    asimov_charged_B_neutral_pi_fit_case = FitCase(
        name="asimov_charged_B_neutral_pi_fit_case",
        base_dir_name="Fitting_Asimov_DSS",
        sample_key="",
        output_tag="AsimovFit_DSS_FixedNuisanceParams",
        plot_info_text="Asimov Data",
        fix_nuisance_parameters=True,
        asimov_data=True,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: twoD_fit_variable_dict[rm] for rm in fitting_utils.dss_charged_B_neutral_pi_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_charged_B_neutral_pi_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.charged_B_neutral_pi_template_setup,
        sig_id_to_data_type_mapping=dtf.charged_B_neutral_pi_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.charged_B_neutral_pi_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.charged_B_neutral_pi_sigID_customizing_dict,
            "do_significance": False,
            "do_toys": do_toys,
        },
    )

    asimov_neutral_B_charged_pi_fit_case = FitCase(
        name="asimov_neutral_B_charged_pi_fit_case",
        base_dir_name="Fitting_Asimov_DSS",
        sample_key="",
        output_tag="AsimovFit_DSS_FixedNuisanceParams",
        plot_info_text="Asimov Data",
        fix_nuisance_parameters=True,
        asimov_data=True,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: twoD_fit_variable_dict[rm] for rm in fitting_utils.dss_neutral_B_charged_pi_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_neutral_B_charged_pi_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.neutral_B_charged_pi_template_setup,
        sig_id_to_data_type_mapping=dtf.neutral_B_charged_pi_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.neutral_B_charged_pi_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.neutral_B_charged_pi_sigID_customizing_dict,
            "do_significance": False,
            "do_toys": do_toys,
        },
    )

    asimov_neutral_B_neutral_pi_fit_case = FitCase(
        name="asimov_neutral_B_neutral_pi_fit_case",
        base_dir_name="Fitting_Asimov_DSS",
        sample_key="",
        output_tag="AsimovFit_DSS_FixedNuisanceParams",
        plot_info_text="Asimov Data",
        fix_nuisance_parameters=True,
        asimov_data=True,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: twoD_fit_variable_dict[rm] for rm in fitting_utils.dss_neutral_B_neutral_pi_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_neutral_B_neutral_pi_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.neutral_B_neutral_pi_template_setup,
        sig_id_to_data_type_mapping=dtf.neutral_B_neutral_pi_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.neutral_B_neutral_pi_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.neutral_B_neutral_pi_sigID_customizing_dict,
            "do_significance": False,
            "do_toys": do_toys,
        },
    )

    # Decoupled Fit Cases
    asimov_decoupled_charged_B_charged_pi_fit_case = FitCase(
        name="asimov_decoupled_charged_B_charged_pi_fit_case",
        base_dir_name="Fitting_Asimov_DSS",
        sample_key="",
        output_tag="AsimovFit_DSS_FixedNuisanceParams",
        plot_info_text="Asimov Data",
        fix_nuisance_parameters=True,
        asimov_data=True,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: twoD_fit_variable_dict[rm] for rm in fitting_utils.dss_charged_B_charged_pi_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_charged_B_charged_pi_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.decoupled_charged_B_charged_pi_template_setup,
        sig_id_to_data_type_mapping=dtf.decoupled_charged_B_charged_pi_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.decoupled_charged_B_charged_pi_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.decoupled_charged_B_charged_pi_sigID_customizing_dict,
            "do_significance": False,
            "do_toys": do_toys,
        },
    )

    asimov_decoupled_charged_B_neutral_pi_fit_case = FitCase(
        name="asimov_decoupled_charged_B_neutral_pi_fit_case",
        base_dir_name="Fitting_Asimov_DSS",
        sample_key="",
        output_tag="AsimovFit_DSS_FixedNuisanceParams",
        plot_info_text="Asimov Data",
        fix_nuisance_parameters=True,
        asimov_data=True,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: twoD_fit_variable_dict[rm] for rm in fitting_utils.dss_charged_B_neutral_pi_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_charged_B_neutral_pi_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.decoupled_charged_B_neutral_pi_template_setup,
        sig_id_to_data_type_mapping=dtf.decoupled_charged_B_neutral_pi_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.decoupled_charged_B_neutral_pi_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.decoupled_charged_B_neutral_pi_sigID_customizing_dict,
            "do_significance": False,
            "do_toys": do_toys,
        },
    )

    asimov_decoupled_neutral_B_charged_pi_fit_case = FitCase(
        name="asimov_decoupled_neutral_B_charged_pi_fit_case",
        base_dir_name="Fitting_Asimov_DSS",
        sample_key="",
        output_tag="AsimovFit_DSS_FixedNuisanceParams",
        plot_info_text="Asimov Data",
        fix_nuisance_parameters=True,
        asimov_data=True,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: twoD_fit_variable_dict[rm] for rm in fitting_utils.dss_neutral_B_charged_pi_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_neutral_B_charged_pi_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.decoupled_neutral_B_charged_pi_template_setup,
        sig_id_to_data_type_mapping=dtf.decoupled_neutral_B_charged_pi_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.decoupled_neutral_B_charged_pi_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.decoupled_neutral_B_charged_pi_sigID_customizing_dict,
            "do_significance": False,
            "do_toys": do_toys,
        },
    )

    asimov_decoupled_neutral_B_neutral_pi_fit_case = FitCase(
        name="asimov_decoupled_neutral_B_neutral_pi_fit_case",
        base_dir_name="Fitting_Asimov_DSS",
        sample_key="",
        output_tag="AsimovFit_DSS_FixedNuisanceParams",
        plot_info_text="Asimov Data",
        fix_nuisance_parameters=True,
        asimov_data=True,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: twoD_fit_variable_dict[rm] for rm in fitting_utils.dss_neutral_B_neutral_pi_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_neutral_B_neutral_pi_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.decoupled_neutral_B_neutral_pi_template_setup,
        sig_id_to_data_type_mapping=dtf.decoupled_neutral_B_neutral_pi_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.decoupled_neutral_B_neutral_pi_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.decoupled_neutral_B_neutral_pi_sigID_customizing_dict,
            "do_significance": False,
            "do_toys": do_toys,
        },
    )

    # Alternative Asimov Setups
    asimov_alternative_charged_B_charged_pi_fit_case = FitCase(
        name="asimov_alternative_charged_B_charged_pi_fit_case",
        base_dir_name="Fitting_Asimov_DSS",
        sample_key="",
        output_tag="AsimovFit_DSS_FixedNuisanceParams",
        plot_info_text="Asimov Data",
        fix_nuisance_parameters=True,
        asimov_data=True,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: twoD_fit_variable_dict[rm] for rm in fitting_utils.dss_charged_B_charged_pi_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_charged_B_charged_pi_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.alternative_charged_B_charged_pi_template_setup,
        sig_id_to_data_type_mapping=dtf.charged_B_charged_pi_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.alternative_charged_B_charged_pi_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.alternative_charged_B_charged_pi_sigID_customizing_dict,
            "do_significance": False,
            "do_toys": do_toys,
        },
    )

    # Mbc Sideband Fit Cases
    mbc_charged_B_charged_pi_fit_case = FitCase(
        name="mbc_charged_B_charged_pi_fit_case",
        base_dir_name="Fitting_Mbc_DSS",
        sample_key="data_prepared",
        output_tag="Mbc_FixedNuisanceParams",
        plot_info_text=r"$\mathrm{M}_{\mathrm{bc}}$ Sideband Data",
        fix_nuisance_parameters=True,
        asimov_data=False,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: mbc_fit_variable_dict[rm] for rm in fitting_utils.dss_charged_B_charged_pi_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_charged_B_charged_pi_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.charged_B_charged_pi_template_setup,
        sig_id_to_data_type_mapping=dtf.charged_B_charged_pi_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.charged_B_charged_pi_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.charged_B_charged_pi_sigID_customizing_dict,
            "sideband": "mbc_sideband",
        },
    )

    mbc_charged_B_neutral_pi_fit_case = FitCase(
        name="mbc_charged_B_neutral_pi_fit_case",
        base_dir_name="Fitting_Mbc_DSS",
        sample_key="data_prepared",
        output_tag="Mbc_FixedNuisanceParams",
        plot_info_text=r"$\mathrm{M}_{\mathrm{bc}}$ Sideband Data",
        fix_nuisance_parameters=True,
        asimov_data=False,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: mbc_fit_variable_dict[rm] for rm in fitting_utils.mbc_dss_charged_B_neutral_pi_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.mbc_dss_charged_B_neutral_pi_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.charged_B_neutral_pi_template_setup,
        sig_id_to_data_type_mapping=dtf.charged_B_neutral_pi_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.charged_B_neutral_pi_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.charged_B_neutral_pi_sigID_customizing_dict,
            "sideband": "mbc_sideband",
        },
    )

    mbc_neutral_B_charged_pi_fit_case = FitCase(
        name="mbc_neutral_B_charged_pi_fit_case",
        base_dir_name="Fitting_Mbc_DSS",
        sample_key="data_prepared",
        output_tag="Mbc_FixedNuisanceParams",
        plot_info_text=r"$\mathrm{M}_{\mathrm{bc}}$ Sideband Data",
        fix_nuisance_parameters=True,
        asimov_data=False,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: mbc_fit_variable_dict[rm] for rm in fitting_utils.dss_neutral_B_charged_pi_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_neutral_B_charged_pi_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.neutral_B_charged_pi_template_setup,
        sig_id_to_data_type_mapping=dtf.neutral_B_charged_pi_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.neutral_B_charged_pi_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.neutral_B_charged_pi_sigID_customizing_dict,
            "sideband": "mbc_sideband",
        },
    )

    mbc_neutral_B_neutral_pi_fit_case = FitCase(
        name="mbc_neutral_B_neutral_pi_fit_case",
        base_dir_name="Fitting_Mbc_DSS",
        sample_key="data_prepared",
        output_tag="Mbc_FixedNuisanceParams",
        plot_info_text=r"$\mathrm{M}_{\mathrm{bc}}$ Sideband Data",
        fix_nuisance_parameters=True,
        asimov_data=False,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: mbc_fit_variable_dict[rm] for rm in fitting_utils.mbc_dss_neutral_B_neutral_pi_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.mbc_dss_neutral_B_neutral_pi_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.neutral_B_neutral_pi_template_setup,
        sig_id_to_data_type_mapping=dtf.neutral_B_neutral_pi_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.neutral_B_neutral_pi_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.neutral_B_neutral_pi_sigID_customizing_dict,
            "sideband": "mbc_sideband",
        },
    )

    # Decoupled Fit Cases
    mbc_decoupled_charged_B_charged_pi_fit_case = FitCase(
        name="mbc_decoupled_charged_B_charged_pi_fit_case",
        base_dir_name="Fitting_Mbc_DSS",
        sample_key="data_prepared",
        output_tag="Mbc_FixedNuisanceParams",
        plot_info_text=r"$\mathrm{M}_{\mathrm{bc}}$ Sideband Data",
        fix_nuisance_parameters=True,
        asimov_data=False,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: mbc_fit_variable_dict[rm] for rm in fitting_utils.dss_charged_B_charged_pi_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_charged_B_charged_pi_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.decoupled_charged_B_charged_pi_template_setup,
        sig_id_to_data_type_mapping=dtf.decoupled_charged_B_charged_pi_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.decoupled_charged_B_charged_pi_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.decoupled_charged_B_charged_pi_sigID_customizing_dict,
            "sideband": "mbc_sideband",
        },
    )

    mbc_decoupled_charged_B_neutral_pi_fit_case = FitCase(
        name="mbc_decoupled_charged_B_neutral_pi_fit_case",
        base_dir_name="Fitting_Mbc_DSS",
        sample_key="data_prepared",
        output_tag="Mbc_FixedNuisanceParams",
        plot_info_text=r"$\mathrm{M}_{\mathrm{bc}}$ Sideband Data",
        fix_nuisance_parameters=True,
        asimov_data=False,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: mbc_fit_variable_dict[rm] for rm in fitting_utils.mbc_dss_charged_B_neutral_pi_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.mbc_dss_charged_B_neutral_pi_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.decoupled_charged_B_neutral_pi_template_setup,
        sig_id_to_data_type_mapping=dtf.decoupled_charged_B_neutral_pi_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.decoupled_charged_B_neutral_pi_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.decoupled_charged_B_neutral_pi_sigID_customizing_dict,
            "sideband": "mbc_sideband",
        },
    )

    mbc_decoupled_neutral_B_charged_pi_fit_case = FitCase(
        name="mbc_decoupled_neutral_B_charged_pi_fit_case",
        base_dir_name="Fitting_Mbc_DSS",
        sample_key="data_prepared",
        output_tag="Mbc_FixedNuisanceParams",
        plot_info_text=r"$\mathrm{M}_{\mathrm{bc}}$ Sideband Data",
        fix_nuisance_parameters=True,
        asimov_data=False,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: mbc_fit_variable_dict[rm] for rm in fitting_utils.dss_neutral_B_charged_pi_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_neutral_B_charged_pi_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.decoupled_neutral_B_charged_pi_template_setup,
        sig_id_to_data_type_mapping=dtf.decoupled_neutral_B_charged_pi_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.decoupled_neutral_B_charged_pi_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.decoupled_neutral_B_charged_pi_sigID_customizing_dict,
            "sideband": "mbc_sideband",
        },
    )

    mbc_decoupled_neutral_B_neutral_pi_fit_case = FitCase(
        name="mbc_decoupled_neutral_B_neutral_pi_fit_case",
        base_dir_name="Fitting_Mbc_DSS",
        sample_key="data_prepared",
        output_tag="Mbc_FixedNuisanceParams",
        plot_info_text=r"$\mathrm{M}_{\mathrm{bc}}$ Sideband Data",
        fix_nuisance_parameters=True,
        asimov_data=False,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: mbc_fit_variable_dict[rm] for rm in fitting_utils.mbc_dss_neutral_B_neutral_pi_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.mbc_dss_neutral_B_neutral_pi_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.decoupled_neutral_B_neutral_pi_template_setup,
        sig_id_to_data_type_mapping=dtf.decoupled_neutral_B_neutral_pi_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.decoupled_neutral_B_neutral_pi_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.decoupled_neutral_B_neutral_pi_sigID_customizing_dict,
            "sideband": "mbc_sideband",
        },
    )

    # Alternative Mbc Fit Cases
    mbc_alternative_charged_B_charged_pi_fit_case = FitCase(
        name="mbc_alternative_charged_B_charged_pi_fit_case",
        base_dir_name="Fitting_Mbc_DSS",
        sample_key="data_prepared",
        output_tag="Mbc_FixedNuisanceParams",
        plot_info_text=r"$\mathrm{M}_{\mathrm{bc}}$ Sideband Data",
        fix_nuisance_parameters=True,
        asimov_data=False,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: mbc_fit_variable_dict[rm] for rm in fitting_utils.dss_charged_B_charged_pi_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_charged_B_charged_pi_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.alternative_charged_B_charged_pi_template_setup,
        sig_id_to_data_type_mapping=dtf.charged_B_charged_pi_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.alternative_charged_B_charged_pi_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.alternative_charged_B_charged_pi_sigID_customizing_dict,
            "sideband": "mbc_sideband",
        },
    )

    mbc_alternative_decoupled_charged_B_charged_pi_fit_case = FitCase(
        name="mbc_alternative_decoupled_charged_B_charged_pi_fit_case",
        base_dir_name="Fitting_Mbc_DSS",
        sample_key="data_prepared",
        output_tag="Mbc_FixedNuisanceParams",
        plot_info_text=r"$\mathrm{M}_{\mathrm{bc}}$ Sideband Data",
        fix_nuisance_parameters=True,
        asimov_data=False,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: mbc_fit_variable_dict[rm] for rm in fitting_utils.dss_charged_B_charged_pi_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_charged_B_charged_pi_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.alternative_decoupled_charged_B_charged_pi_template_setup,
        sig_id_to_data_type_mapping=dtf.decoupled_charged_B_charged_pi_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.alternative_decoupled_charged_B_charged_pi_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.alternative_decoupled_charged_B_charged_pi_sigID_customizing_dict,
            "sideband": "mbc_sideband",
        },
    )

    # Asimov Normalization Fits
    asimov_normalization_fit_decoupled_charged_B_fit_case = FitCase(
        name="asimov_normalization_fit_decoupled_charged_B_fit_case",
        base_dir_name="Fitting_DSS_Normalization_Asimov",
        sample_key="",
        output_tag="NormalizationAsimov_FixedNuisanceParams",
        plot_info_text="Asimov Data",
        fix_nuisance_parameters=True,
        asimov_data=True,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: mbc_fit_variable_dict[rm] for rm in fitting_utils.normalization_fit_dss_charged_B_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.normalization_fit_dss_charged_B_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.decoupled_charged_B_normalization_fit_template_setup,
        sig_id_to_data_type_mapping=dtf.normalization_fit_decoupled_charged_B_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.decoupled_charged_B_normalization_fit_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.decoupled_charged_B_normalization_fit_sigID_customizing_dict,
        },
    )

    asimov_normalization_fit_decoupled_neutral_B_fit_case = FitCase(
        name="asimov_normalization_fit_decoupled_neutral_B_fit_case",
        base_dir_name="Fitting_DSS_Normalization_Asimov",
        sample_key="",
        output_tag="NormalizationAsimov_FixedNuisanceParams",
        plot_info_text="Asimov Data",
        fix_nuisance_parameters=True,
        asimov_data=True,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: mbc_fit_variable_dict[rm] for rm in fitting_utils.normalization_fit_dss_neutral_B_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.normalization_fit_dss_neutral_B_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.decoupled_neutral_B_normalization_fit_template_setup,
        sig_id_to_data_type_mapping=dtf.normalization_fit_decoupled_neutral_B_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.decoupled_neutral_B_normalization_fit_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.decoupled_neutral_B_normalization_fit_sigID_customizing_dict,
        },
    )

    asimov_normalization_fit_coupled_charged_B_fit_case = FitCase(
        name="asimov_normalization_fit_coupled_charged_B_fit_case",
        base_dir_name="Fitting_DSS_Normalization_Asimov",
        sample_key="",
        output_tag="NormalizationAsimov_FixedNuisanceParams",
        plot_info_text="Asimov Data",
        fix_nuisance_parameters=True,
        asimov_data=True,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: mbc_fit_variable_dict[rm] for rm in fitting_utils.normalization_fit_dss_charged_B_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.normalization_fit_dss_charged_B_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.coupled_charged_B_normalization_fit_template_setup,
        sig_id_to_data_type_mapping=dtf.normalization_fit_coupled_charged_B_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.coupled_charged_B_normalization_fit_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.coupled_charged_B_normalization_fit_sigID_customizing_dict,
        },
    )

    asimov_normalization_fit_coupled_neutral_B_fit_case = FitCase(
        name="asimov_normalization_fit_coupled_neutral_B_fit_case",
        base_dir_name="Fitting_DSS_Normalization_Asimov",
        sample_key="",
        output_tag="NormalizationAsimov_FixedNuisanceParams",
        plot_info_text="Asimov Data",
        fix_nuisance_parameters=True,
        asimov_data=True,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: mbc_fit_variable_dict[rm] for rm in fitting_utils.normalization_fit_dss_neutral_B_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.normalization_fit_dss_neutral_B_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.coupled_neutral_B_normalization_fit_template_setup,
        sig_id_to_data_type_mapping=dtf.normalization_fit_coupled_neutral_B_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.coupled_neutral_B_normalization_fit_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.coupled_neutral_B_normalization_fit_sigID_customizing_dict,
        },
    )

    # Normalization Fits on Data
    normalization_fit_decoupled_charged_B_fit_case = FitCase(
        name="normalization_fit_decoupled_charged_B_fit_case",
        base_dir_name="Fitting_DSS_Normalization_on_Data",
        sample_key="data_prepared",
        output_tag="NormalizationOnData_FixedNuisanceParams",
        plot_info_text="Data",
        fix_nuisance_parameters=True,
        asimov_data=False,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: mbc_fit_variable_dict[rm] for rm in fitting_utils.normalization_fit_dss_charged_B_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.normalization_fit_dss_charged_B_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.decoupled_charged_B_normalization_fit_template_setup,
        sig_id_to_data_type_mapping=dtf.normalization_fit_decoupled_charged_B_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.decoupled_charged_B_normalization_fit_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.decoupled_charged_B_normalization_fit_sigID_customizing_dict,
        },
    )

    normalization_fit_decoupled_neutral_B_fit_case = FitCase(
        name="normalization_fit_decoupled_neutral_B_fit_case",
        base_dir_name="Fitting_DSS_Normalization_on_Data",
        sample_key="data_prepared",
        output_tag="NormalizationOnData_FixedNuisanceParams",
        plot_info_text="Data",
        fix_nuisance_parameters=True,
        asimov_data=False,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: mbc_fit_variable_dict[rm] for rm in fitting_utils.normalization_fit_dss_neutral_B_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.normalization_fit_dss_neutral_B_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.decoupled_neutral_B_normalization_fit_template_setup,
        sig_id_to_data_type_mapping=dtf.normalization_fit_decoupled_neutral_B_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.decoupled_neutral_B_normalization_fit_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.decoupled_neutral_B_normalization_fit_sigID_customizing_dict,
        },
    )

    normalization_fit_coupled_charged_B_fit_case = FitCase(
        name="normalization_fit_coupled_charged_B_fit_case",
        base_dir_name="Fitting_DSS_Normalization_on_Data",
        sample_key="data_prepared",
        output_tag="NormalizationOnData_FixedNuisanceParams",
        plot_info_text="Data",
        fix_nuisance_parameters=True,
        asimov_data=False,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: mbc_fit_variable_dict[rm] for rm in fitting_utils.normalization_fit_dss_charged_B_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.normalization_fit_dss_charged_B_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.coupled_charged_B_normalization_fit_template_setup,
        sig_id_to_data_type_mapping=dtf.normalization_fit_coupled_charged_B_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.coupled_charged_B_normalization_fit_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.coupled_charged_B_normalization_fit_sigID_customizing_dict,
        },
    )

    normalization_fit_coupled_neutral_B_fit_case = FitCase(
        name="normalization_fit_coupled_neutral_B_fit_case",
        base_dir_name="Fitting_DSS_Normalization_on_Data",
        sample_key="data_prepared",
        output_tag="NormalizationOnData_FixedNuisanceParams",
        plot_info_text="Data",
        fix_nuisance_parameters=True,
        asimov_data=False,
        round_asimov_data=False,
        fit_variables_by_mode_dict={
            rm: mbc_fit_variable_dict[rm] for rm in fitting_utils.normalization_fit_dss_neutral_B_modes.keys()
        },
        weight_column_name=fitting_utils.final_weight_column_for_asimov,
        systematics_cols=tuple([]),
        reco_mode_infos=[(k, v) for k, v in fitting_utils.normalization_fit_dss_neutral_B_modes.items()],
        reco_mode_labels_dict=fitting_utils.dss_labels,
        template_setup=dss_bf_fitting_utils.coupled_neutral_B_normalization_fit_template_setup,
        sig_id_to_data_type_mapping=dtf.normalization_fit_coupled_neutral_B_sig_dss_data_type_mapping,
        plot_colors=dss_bf_fitting_utils.dss_colors,
        plot_labels=dss_bf_fitting_utils.dss_labels,
        sig_id_column_name=dss_bf_fitting_utils.dss_sig_id_column,
        sig_queries=dss_bf_fitting_utils.coupled_neutral_B_normalization_fit_sig_queries,
        additional_info={
            "mva_variable_name": mva_variable_name,
            "sig_id_customizing_dict": dss_bf_fitting_utils.coupled_neutral_B_normalization_fit_sigID_customizing_dict,
        },
    )

    @staticmethod
    def get_attributes_mapping() -> Dict[str, FitCase]:
        return {attr_name: val for attr_name, val in BFFitCaseDict.__dict__.items()}

    @staticmethod
    def get_fit_case(fit_case_name) -> FitCase:
        fit_case = BFFitCaseDict.get_attributes_mapping()[fit_case_name]
        assert fit_case.name == fit_case_name, (fit_case.name, fit_case_name)
        return BFFitCaseDict.get_attributes_mapping()[fit_case_name]

    @staticmethod
    def get_toy_fit_case_names() -> List[str]:
        all_fit_cases = BFFitCaseDict.get_attributes_mapping()
        return [
            fit_case_name
            for fit_case_name, fit_case in all_fit_cases.items()
            if isinstance(fit_case, FitCase)
            and fit_case.get_additional_setting_with_default(setting_key="do_toys", default_setting=False)
        ]


if __name__ == "__main__":
    print(BFFitCaseDict.get_attributes_mapping())
    print(BFFitCaseDict.get_fit_case(fit_case_name="decoupled_neutral_B_neutral_pi_fit_case").name)
