"""
Translates the decay mode ids defined by the FEI human readable labels.

Based on Kilian's Class UniqueDecayModeIDSetter; See also
    - https://stash.desy.de/users/klieret/repos/tagcal/browse/tagcal/loading/udm_id_setter.py
    - https://stash.desy.de/users/klieret/repos/tagcal/browse/tagcal/loading/test_udm_id_setter.py

Felix Metzner 2021
"""

import fei
import numpy as np
import pandas as pd

from typing import Optional, Dict, Tuple, List, Set, Any


__all__ = [
    "FEIDecayModeIDInterpreter",
]


class FEIDecayModeIDInterpreter:
    charged_b_meson_key: str = "B+"
    neutral_b_meson_key: str = "B0"
    b_meson_keys: Tuple[str, str] = (charged_b_meson_key, neutral_b_meson_key)

    def __init__(
        self,
        fei_settings_dict: Optional[Dict[str, Any]] = None,
    ) -> None:
        self._fei_settings_dict: Optional[Dict[str, Any]] = fei_settings_dict

        self.fei_decay_mode_id_dict: Dict[str, Dict[int, str]] = self.get_fei_decay_mode_id_dict(
            fei_settings_dict=self._fei_settings_dict
        )

        self.modified_fei_decay_mode_id_dict: Dict[int, str] = self.get_modified_fei_decay_mode_id_dict(
            fei_settings_dict=self._fei_settings_dict
        )

        self.modified_decay_id_to_latex_label: Dict[int, str] = self.get_modified_fei_decay_mode_to_latex_label_dict(
            fei_settings_dict=self._fei_settings_dict
        )

    @staticmethod
    def get_modified_fei_decay_mode_ids(
        df: pd.DataFrame,
        b_meson_pdg_column_name: str,
        fei_decay_mode_id_column_name: str,
    ) -> np.ndarray:
        neutral_pdg_codes = [511, -511]
        charged_pdg_codes = [521, -521]

        assert b_meson_pdg_column_name in df.columns, b_meson_pdg_column_name
        assert fei_decay_mode_id_column_name in df.columns, fei_decay_mode_id_column_name

        original_fei_mapping: Dict[str, Dict[int, str]] = FEIDecayModeIDInterpreter.get_fei_decay_mode_id_dict()
        o_vals: Tuple[int, ...] = tuple([k for v in original_fei_mapping.values() for k in v.keys()])
        n_vals: Tuple[int, ...] = tuple(FEIDecayModeIDInterpreter.get_modified_fei_decay_mode_id_dict().keys())

        assert np.all(np.isin(element=df[fei_decay_mode_id_column_name].unique(), test_elements=list(o_vals))), [
            v for v in df[fei_decay_mode_id_column_name].unique() if v not in o_vals
        ]

        is_neutral_b_tag = df[b_meson_pdg_column_name].isin(neutral_pdg_codes)
        is_charged_b_tag = df[b_meson_pdg_column_name].isin(charged_pdg_codes)

        assert sum(is_neutral_b_tag | is_charged_b_tag) == len(np.unique(df.index))
        assert sum(is_neutral_b_tag & is_charged_b_tag) == 0

        new_values = np.select(
            condlist=[is_neutral_b_tag.values, is_charged_b_tag.values],
            choicelist=[
                -1 * (df[fei_decay_mode_id_column_name].values + 1),
                +1 * df[fei_decay_mode_id_column_name].values,
            ],
            default=np.nan,
        )

        assert np.all(np.isin(element=np.unique(new_values), test_elements=list(n_vals))), [
            v for v in np.unique(new_values) if v not in n_vals
        ]

        return new_values

    @staticmethod
    def decay_mode_id_transformation(mode_id: int, neutral_b: bool) -> int:
        factor = -1 if neutral_b else 1
        return factor * (mode_id + int(neutral_b))

    @staticmethod
    def is_charged_b_mode(modified_decay_id: int) -> bool:
        return modified_decay_id >= 0

    @staticmethod
    def fei_id_flavor_to_modified_id(fei_id: int, b_pdg_code: int) -> int:
        if abs(b_pdg_code) not in (511, 521):
            return -999
        return FEIDecayModeIDInterpreter.decay_mode_id_transformation(mode_id=fei_id, neutral_b=abs(b_pdg_code) == 511)

    @staticmethod
    def modified_id_to_fei_id_flavor(modified_decay_id: int) -> Tuple[int, int]:
        if int(modified_decay_id) == -999:
            return -999, -999
        return abs(modified_decay_id) - int(modified_decay_id < 0), 511 if modified_decay_id < 0 else 521

    @staticmethod
    def get_fei_decay_mode_id_dict(fei_settings_dict: Optional[Dict[str, Any]] = None) -> Dict[str, Dict[int, str]]:
        particles: List[fei.Particle] = fei.get_default_channels(
            convertedFromBelle=True,
            hadronic=True,
            semileptonic=False,
            specific=False,
        )
        if fei_settings_dict is not None:
            particles = fei.get_default_channels(**fei_settings_dict)

        dct = {}
        for p in particles:
            if p.name not in FEIDecayModeIDInterpreter.b_meson_keys:
                continue
            dct[p.name] = {c.decayModeID: c.decayString for c in p.channels}
        return dct

    @staticmethod
    def get_modified_fei_decay_mode_id_dict(fei_settings_dict: Optional[Dict[str, Any]] = None) -> Dict[int, str]:
        decay_mode_id_dict: Dict[str, Dict[int, str]] = FEIDecayModeIDInterpreter.get_fei_decay_mode_id_dict(
            fei_settings_dict=fei_settings_dict
        )

        new_mapping: Dict[int, str] = {}

        for mode_id, decay_str in decay_mode_id_dict[FEIDecayModeIDInterpreter.charged_b_meson_key].items():
            new_mode_id = FEIDecayModeIDInterpreter.decay_mode_id_transformation(mode_id=mode_id, neutral_b=False)
            new_mapping[new_mode_id] = decay_str
        for mode_id, decay_str in decay_mode_id_dict[FEIDecayModeIDInterpreter.neutral_b_meson_key].items():
            new_mode_id = FEIDecayModeIDInterpreter.decay_mode_id_transformation(mode_id=mode_id, neutral_b=True)
            new_mapping[new_mode_id] = decay_str

        old_len: int = sum([len(v) for v in decay_mode_id_dict.values()])
        assert len(new_mapping) == old_len, (len(new_mapping), old_len)
        old_values: Set[str] = set([v for vs in decay_mode_id_dict.values() for v in vs.values()])
        new_values: Set[str] = set(new_mapping.values())
        assert old_values == new_values, old_values.symmetric_difference(new_values)
        return new_mapping

    @staticmethod
    def get_modified_fei_decay_mode_to_latex_label_dict(
        fei_settings_dict: Optional[Dict[str, Any]] = None,
        include_b_to: bool = False,
    ) -> Dict[int, str]:
        id_map = FEIDecayModeIDInterpreter.get_modified_fei_decay_mode_id_dict(fei_settings_dict=fei_settings_dict)
        return {
            decay_id: FEIDecayModeIDInterpreter.fei_decay_string_to_latex(decay_str=decay_str, include_b_to=include_b_to)
            for decay_id, decay_str in id_map.items()
        }

    @staticmethod
    def fei_decay_string_to_latex(decay_str: str, include_b_to: bool = False) -> str:
        assert "==>" in decay_str, decay_str

        first_part = r""
        if include_b_to:
            first_part = (
                decay_str.split(sep="==>")[0]
                .rsplit(sep="_", maxsplit=1)[0]
                .strip()
                .replace("anti-", r"\bar ")
                .replace("0", "^{0}")
                .replace("+", "^{+}")
                .replace("-", "^{-}")
                .replace("B", r"\mathrm{B}")
                .replace(":generic", "")
                .replace(":semileptonic", r"_{\mathrm{SL}}")
            ) + r" \rightarrow "

        latex_str = (
            decay_str.split(sep="==>")[1]
            .strip()
            .replace("anti-", r"\bar ")
            .replace("0", r"^{0}")
            .replace("+", r"^{+}")
            .replace("-", r"^{-}")
            .replace("*^0", r"^{*0}")
            .replace("*^{+}", r"^{*+}")
            .replace("*^{-}", r"^{*-}")
            .replace(":generic", "")
            .replace(":semileptonic", r"_{\mathrm{SL}}")
            .replace(" mu", r" \mu")
            .replace(" pi", r" \pi")
            .replace("psi", r"\psi")
            .replace("D", r"\mathrm{D}")
            .replace("J", r"\mathrm{J}")
            .replace("K", r"\mathrm{K}")
            .replace("_s", r"_\mathrm{s}")
            .replace("_S", r"_\mathrm{S}")
        )
        return r"$" + first_part + latex_str + r"$"

    @staticmethod
    def get_backup_fei_reco_mode_latex_label_mapping() -> Dict[int, str]:
        return {
            -26: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{J}/\\psi \\mathrm{K}_\\mathrm{S}^{0} \\pi^{+} \\pi^{-}$",
            -25: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{J}/\\psi \\mathrm{K}^{+} \\pi^{-}$",
            -24: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{J}/\\psi \\mathrm{K}_\\mathrm{S}^{0}$",
            -23: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}_\\mathrm{s}^{*+} \\mathrm{D}^{*-}$",
            -22: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}_\\mathrm{s}^{+} \\mathrm{D}^{*-}$",
            -21: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}_\\mathrm{s}^{*+} \\mathrm{D}^{-}$",
            -20: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}^{*-} \\pi^{+} \\pi^{+} \\pi^{-} \\pi^{0}$",
            -19: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}^{*-} \\pi^{+} \\pi^{+} \\pi^{-}$",
            -18: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}^{*-} \\pi^{+} \\pi^{0} \\pi^{0}$",
            -17: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}^{*-} \\pi^{+} \\pi^{0}$",
            -16: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}^{*-} \\pi^{+}$",
            -15: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}_\\mathrm{s}^{+} \\mathrm{D}^{-}$",
            -14: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}^{*-} \\mathrm{D}^{*+} \\mathrm{K}_\\mathrm{S}^{0}$",
            -13: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}^{-} \\mathrm{D}^{*+} \\mathrm{K}_\\mathrm{S}^{0}$",
            -12: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}^{*-} \\mathrm{D}^{+} \\mathrm{K}_\\mathrm{S}^{0}$",
            -11: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}^{-} \\mathrm{D}^{+} \\mathrm{K}_\\mathrm{S}^{0}$",
            -10: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}^{*-} \\mathrm{D}*^{0} \\mathrm{K}^{+}$",
            -9: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}^{*-} \\mathrm{D}^{0} \\mathrm{K}^{+}$",
            -8: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}^{-} \\mathrm{D}*^{0} \\mathrm{K}^{+}$",
            -7: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}^{-} \\mathrm{D}^{0} \\mathrm{K}^{+}$",
            -6: "$\\mathrm{B}^{0} \\rightarrow \\bar \\mathrm{D}^{0} \\pi^{+} \\pi^{-}$",
            -5: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}^{-} \\pi^{+} \\pi^{+} \\pi^{-} \\pi^{0}$",
            -4: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}^{-} \\pi^{+} \\pi^{+} \\pi^{-}$",
            -3: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}^{-} \\pi^{+} \\pi^{0} \\pi^{0}$",
            -2: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}^{-} \\pi^{+} \\pi^{0}$",
            -1: "$\\mathrm{B}^{0} \\rightarrow \\mathrm{D}^{-} \\pi^{+}$",
            0: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}^{0} \\pi^{+}$",
            1: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}^{0} \\pi^{+} \\pi^{0}$",
            2: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}^{0} \\pi^{+} \\pi^{0} \\pi^{0}$",
            3: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}^{0} \\pi^{+} \\pi^{+} \\pi^{-}$",
            4: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}^{0} \\pi^{+} \\pi^{+} \\pi^{-} \\pi^{0}$",
            5: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}^{0} \\mathrm{D}^{+}$",
            6: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}^{0} \\mathrm{D}^{+} \\mathrm{K}_\\mathrm{S}^{0}$",
            7: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}*^{0} \\mathrm{D}^{+} \\mathrm{K}_\\mathrm{S}^{0}$",
            8: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}^{0} \\mathrm{D}^{*+} \\mathrm{K}_\\mathrm{S}^{0}$",
            9: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}*^{0} \\mathrm{D}^{*+} \\mathrm{K}_\\mathrm{S}^{0}$",
            10: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}^{0} \\mathrm{D}^{0} \\mathrm{K}^{+}$",
            11: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}*^{0} \\mathrm{D}^{0} \\mathrm{K}^{+}$",
            12: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}^{0} \\mathrm{D}*^{0} \\mathrm{K}^{+}$",
            13: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}*^{0} \\mathrm{D}*^{0} \\mathrm{K}^{+}$",
            14: "$\\mathrm{B}^{+} \\rightarrow \\mathrm{D}_\\mathrm{s}^{+} \\bar \\mathrm{D}^{0}$",
            15: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}*^{0} \\pi^{+}$",
            16: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}*^{0} \\pi^{+} \\pi^{0}$",
            17: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}*^{0} \\pi^{+} \\pi^{0} \\pi^{0}$",
            18: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}*^{0} \\pi^{+} \\pi^{+} \\pi^{-}$",
            19: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}*^{0} \\pi^{+} \\pi^{+} \\pi^{-} \\pi^{0}$",
            20: "$\\mathrm{B}^{+} \\rightarrow \\mathrm{D}_\\mathrm{s}^{*+} \\bar \\mathrm{D}^{0}$",
            21: "$\\mathrm{B}^{+} \\rightarrow \\mathrm{D}_\\mathrm{s}^{+} \\bar \\mathrm{D}*^{0}$",
            22: "$\\mathrm{B}^{+} \\rightarrow \\bar \\mathrm{D}^{0} \\mathrm{K}^{+}$",
            23: "$\\mathrm{B}^{+} \\rightarrow \\mathrm{D}^{-} \\pi^{+} \\pi^{+}$",
            24: "$\\mathrm{B}^{+} \\rightarrow \\mathrm{D}^{-} \\pi^{+} \\pi^{+} \\pi^{0}$",
            25: "$\\mathrm{B}^{+} \\rightarrow \\mathrm{J}/\\psi \\mathrm{K}^{+}$",
            26: "$\\mathrm{B}^{+} \\rightarrow \\mathrm{J}/\\psi \\mathrm{K}^{+} \\pi^{+} \\pi^{-}$'",
            27: "$\\mathrm{B}^{+} \\rightarrow \\mathrm{J}/\\psi \\mathrm{K}^{+} \\pi^{0}$'",
            28: "$\\mathrm{B}^{+} \\rightarrow \\mathrm{J}/\\psi \\mathrm{K}_\\mathrm{S}^{0} \\pi^{+}$",
        }


def check_fei_decay_mode_interpreter():
    fei_mapping: Dict[str, Dict[int, str]] = FEIDecayModeIDInterpreter.get_fei_decay_mode_id_dict()
    modified_fei_mapping: Dict[int, str] = FEIDecayModeIDInterpreter.get_modified_fei_decay_mode_id_dict()
    id_to_latex_mapping: Dict[int, str] = FEIDecayModeIDInterpreter.get_modified_fei_decay_mode_to_latex_label_dict()
    id_to_latex_w_b_mapping: Dict[int, str] = FEIDecayModeIDInterpreter.get_modified_fei_decay_mode_to_latex_label_dict(
        include_b_to=True
    )

    original_length = sum([len(d) for d in fei_mapping.values()])
    assert original_length == len(modified_fei_mapping), (original_length, len(modified_fei_mapping))
    assert original_length == len(id_to_latex_mapping), (original_length, len(id_to_latex_mapping))
    assert original_length == len(id_to_latex_mapping), (original_length, len(id_to_latex_w_b_mapping))

    print(fei_mapping)
    print(modified_fei_mapping)
    print(id_to_latex_mapping)
    print(id_to_latex_w_b_mapping)

    for d_id, latex_str in id_to_latex_mapping.items():
        print(f"\t-> {d_id}: {latex_str}")

    for d_id, latex_str in id_to_latex_w_b_mapping.items():
        print(f"\t-> {d_id}: {latex_str}")


if __name__ == "__main__":
    check_fei_decay_mode_interpreter()
