"""
Definition of fit result container.

Felix Metzner 2023
"""

import numpy as np

from typing import Tuple
from iminuit import Minuit
from dataclasses import dataclass
from iminuit.util import FMin, Params
from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fitter_rdstar import RDStarFitter


__all__ = [
    "RDStarFitResultContainer",
]


logger = log_to_default(name=__name__)


@dataclass(frozen=True)
class RDStarFitResultContainer:
    param_values: np.ndarray
    errors: np.ndarray
    cov_mat: np.ndarray
    corr_mat: np.ndarray
    fmin: FMin
    param_names: Tuple[str, ...]
    initial_params: Params
    event_yields: np.ndarray
    initial_event_yields: np.ndarray

    @property
    def initial_param_values(self) -> np.ndarray:
        return np.array([p.value for p in self.initial_params])

    @property
    def initial_param_errors(self) -> np.ndarray:
        return np.array([p.error for p in self.initial_params])

    @staticmethod
    def from_minimizer_and_fitter(
        m: Minuit,
        f: RDStarFitter,
    ) -> "RDStarFitResultContainer":
        m_initial_params: np.ndarray = np.array([p.value for p in m.init_params])
        f_initial_params: np.ndarray = f.initial_parameters
        assert np.all(m_initial_params == f_initial_params), [ips for ips in zip(m_initial_params, f_initial_params)]

        return RDStarFitResultContainer(
            param_values=np.array(m.values),
            errors=np.sqrt(np.array(m.covariance).diagonal()),
            cov_mat=np.array(m.covariance),
            corr_mat=np.array(m.covariance.correlation()),
            fmin=m.fmin,
            param_names=m.parameters,
            initial_params=m.init_params,
            event_yields=f.get_yields(params=np.array(m.values)),
            initial_event_yields=f.get_yields(params=f.initial_parameters),
        )
