"""
Fit setup info for dedicated tagging efficiency calibration fit.

Felix Metzner 2022
"""

from typing import Tuple
from uncertainties import ufloat

from rdstar.utilities.logging import log_to_default

from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses

from rdstar.offline_analysis.reco_modes.reco_mode_infos import TagRecoModeInfos
from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import (
    FitSetupTriplet,
    base_asimov_data_setup_info,
    tag_cal_per_tag_mode_fit_obs_setup,
    base_tag_cal_weight_info,
    tag_cal_bbbar_bkg_sig_ids,
    tag_cal_dstst_l_bkg_sig_ids,
    fei_tag_b_reco_based_fit_component_id_column_name,
    add_fei_tag_b_reco_based_fit_component_id,
    get_sig_ids_with_tag_b_reco_info,
)
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import (
    FitSetupInfoContainer,
    ComponentInfo,
    TagEfficiencyCorrectionInfo,
    RecoChannelInfo,
)


__all__ = [
    "tag_cal_fit_setup",
]


logger = log_to_default(name=__name__)


# Definition of components:


sig_d_sig_ids: Tuple[int, ...] = (
    OldMCMatchingIDsQueryInfo.Dl.id,
    OldMCMatchingIDsQueryInfo.Dtau.id,
)
sig_ds_sig_ids: Tuple[int, ...] = (
    OldMCMatchingIDsQueryInfo.DSl.id,
    OldMCMatchingIDsQueryInfo.DStau.id,
)
bkg_sig_ids: Tuple[int, ...] = (
    tag_cal_bbbar_bkg_sig_ids + tag_cal_dstst_l_bkg_sig_ids + (OldMCMatchingIDsQueryInfo.CBKG.id,)
)


# D Signal Components:

component_b_to_d_1cpi_d_sig: ComponentInfo = ComponentInfo(
    name="d_sig_for_b_tag_d_1cpi",
    comp_index=1,
    latex_label="D Sig in " + LatexDecayProcesses.b_to_d_1cpi_tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=sig_d_sig_ids, fei_reco_type="d_1cpi"),
)

component_b_to_d_mcpi_d_sig: ComponentInfo = ComponentInfo(
    name="d_sig_for_b_tag_d_mcpi",
    comp_index=2,
    latex_label="D Sig in " + LatexDecayProcesses.b_to_d_mcpi_tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=sig_d_sig_ids, fei_reco_type="d_mcpi"),
)

component_b_to_ocd_pipi_d_sig: ComponentInfo = ComponentInfo(
    name="d_sig_for_b_tag_ocd_pipi",
    comp_index=3,
    latex_label="D Sig in " + LatexDecayProcesses.b_to_ocd_pipi_tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=sig_d_sig_ids, fei_reco_type="ocd_pipi"),
)

component_b_to_ds_1cpi_d_sig: ComponentInfo = ComponentInfo(
    name="d_sig_for_b_tag_ds_1cpi",
    comp_index=4,
    latex_label="D Sig in " + LatexDecayProcesses.b_to_ds_1cpi_tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=sig_d_sig_ids, fei_reco_type="ds_1cpi"),
)

component_b_to_ds_mcpi_d_sig: ComponentInfo = ComponentInfo(
    name="d_sig_for_b_tag_ds_mcpi",
    comp_index=5,
    latex_label="D Sig in " + LatexDecayProcesses.b_to_ds_mcpi_tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=sig_d_sig_ids, fei_reco_type="ds_mcpi"),
)

component_b_to_j_psi_d_sig: ComponentInfo = ComponentInfo(
    name="d_sig_for_b_tag_j_psi",
    comp_index=6,
    latex_label="D Sig in " + LatexDecayProcesses.b_to_j_psi_tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=sig_d_sig_ids, fei_reco_type="j_psi"),
)

component_b_to_all_dd_d_sig: ComponentInfo = ComponentInfo(
    name="d_sig_for_b_tag_dd_and_dsd",
    comp_index=7,
    latex_label="D Sig in " + LatexDecayProcesses.b_to__other__tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=sig_d_sig_ids, fei_reco_type="dd_and_dsd"),
)

# D* Signal Components:

component_b_to_d_1cpi_ds_sig: ComponentInfo = ComponentInfo(
    name="ds_sig_for_b_tag_d_1cpi",
    comp_index=11,
    latex_label="D$^*$ Sig in " + LatexDecayProcesses.b_to_d_1cpi_tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=sig_ds_sig_ids, fei_reco_type="d_1cpi"),
)

component_b_to_d_mcpi_ds_sig: ComponentInfo = ComponentInfo(
    name="ds_sig_for_b_tag_d_mcpi",
    comp_index=12,
    latex_label="D$^*$ Sig in " + LatexDecayProcesses.b_to_d_mcpi_tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=sig_ds_sig_ids, fei_reco_type="d_mcpi"),
)

component_b_to_ocd_pipi_ds_sig: ComponentInfo = ComponentInfo(
    name="ds_sig_for_b_tag_ocd_pipi",
    comp_index=13,
    latex_label="D$^*$ Sig in " + LatexDecayProcesses.b_to_ocd_pipi_tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=sig_ds_sig_ids, fei_reco_type="ocd_pipi"),
)

component_b_to_ds_1cpi_ds_sig: ComponentInfo = ComponentInfo(
    name="ds_sig_for_b_tag_ds_1cpi",
    comp_index=14,
    latex_label="D$^*$ Sig in " + LatexDecayProcesses.b_to_ds_1cpi_tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=sig_ds_sig_ids, fei_reco_type="ds_1cpi"),
)

component_b_to_ds_mcpi_ds_sig: ComponentInfo = ComponentInfo(
    name="ds_sig_for_b_tag_ds_mcpi",
    comp_index=15,
    latex_label="D$^*$ Sig in " + LatexDecayProcesses.b_to_ds_mcpi_tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=sig_ds_sig_ids, fei_reco_type="ds_mcpi"),
)

component_b_to_j_psi_ds_sig: ComponentInfo = ComponentInfo(
    name="ds_sig_for_b_tag_j_psi",
    comp_index=16,
    latex_label="D$^*$ Sig in " + LatexDecayProcesses.b_to_j_psi_tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=sig_ds_sig_ids, fei_reco_type="j_psi"),
)

component_b_to_all_dd_ds_sig: ComponentInfo = ComponentInfo(
    name="ds_sig_for_b_tag_dd_and_dsd",
    comp_index=17,
    latex_label="D$^*$ Sig in " + LatexDecayProcesses.b_to__other__tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=sig_ds_sig_ids, fei_reco_type="dd_and_dsd"),
)


# Background Components:

component_b_to_d_1cpi_bkg: ComponentInfo = ComponentInfo(
    name="bkg_for_b_tag_d_1cpi",
    comp_index=21,
    latex_label="Bkg in " + LatexDecayProcesses.b_to_d_1cpi_tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=bkg_sig_ids, fei_reco_type="d_1cpi"),
)

component_b_to_d_mcpi_bkg: ComponentInfo = ComponentInfo(
    name="bkg_for_b_tag_d_mcpi",
    comp_index=22,
    latex_label="Bkg in " + LatexDecayProcesses.b_to_d_mcpi_tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=bkg_sig_ids, fei_reco_type="d_mcpi"),
)

component_b_to_ocd_pipi_bkg: ComponentInfo = ComponentInfo(
    name="bkg_for_b_tag_ocd_pipi",
    comp_index=23,
    latex_label="Bkg in " + LatexDecayProcesses.b_to_ocd_pipi_tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=bkg_sig_ids, fei_reco_type="ocd_pipi"),
)

component_b_to_ds_1cpi_bkg: ComponentInfo = ComponentInfo(
    name="bkg_for_b_tag_ds_1cpi",
    comp_index=24,
    latex_label="Bkg in " + LatexDecayProcesses.b_to_ds_1cpi_tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=bkg_sig_ids, fei_reco_type="ds_1cpi"),
)

component_b_to_ds_mcpi_bkg: ComponentInfo = ComponentInfo(
    name="bkg_for_b_tag_ds_mcpi",
    comp_index=25,
    latex_label="Bkg in " + LatexDecayProcesses.b_to_ds_mcpi_tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=bkg_sig_ids, fei_reco_type="ds_mcpi"),
)

component_b_to_j_psi_bkg: ComponentInfo = ComponentInfo(
    name="bkg_for_b_tag_j_psi",
    comp_index=26,
    latex_label="Bkg in " + LatexDecayProcesses.b_to_j_psi_tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=bkg_sig_ids, fei_reco_type="j_psi"),
)

component_b_to_all_dd_bkg: ComponentInfo = ComponentInfo(
    name="bkg_for_b_tag_dd_and_dsd",
    comp_index=27,
    latex_label="Bkg in " + LatexDecayProcesses.b_to__other__tag_modes,
    sig_ids=get_sig_ids_with_tag_b_reco_info(sig_ids=bkg_sig_ids, fei_reco_type="dd_and_dsd"),
)


tag_cal_fit_comps: Tuple[ComponentInfo, ...] = (
    component_b_to_d_1cpi_d_sig,
    component_b_to_d_mcpi_d_sig,
    component_b_to_ocd_pipi_d_sig,
    component_b_to_ds_1cpi_d_sig,
    component_b_to_ds_mcpi_d_sig,
    component_b_to_j_psi_d_sig,
    component_b_to_all_dd_d_sig,
    #
    component_b_to_d_1cpi_ds_sig,
    component_b_to_d_mcpi_ds_sig,
    component_b_to_ocd_pipi_ds_sig,
    component_b_to_ds_1cpi_ds_sig,
    component_b_to_ds_mcpi_ds_sig,
    component_b_to_j_psi_ds_sig,
    component_b_to_all_dd_ds_sig,
    #
    component_b_to_d_1cpi_bkg,
    component_b_to_d_mcpi_bkg,
    component_b_to_ocd_pipi_bkg,
    component_b_to_ds_1cpi_bkg,
    component_b_to_ds_mcpi_bkg,
    component_b_to_j_psi_bkg,
    component_b_to_all_dd_bkg,
)


tag_cal_fit_chs: Tuple[RecoChannelInfo, ...] = tuple(
    [
        RecoChannelInfo(
            name="B_to_D_1cpi",
            ch_index=0,
            reco_mode_info=TagRecoModeInfos.b_to_d_1cpi_tag_modes,
            components=tag_cal_fit_comps,
            observable_infos=tag_cal_per_tag_mode_fit_obs_setup.tag_cal_ptm_obs_info_d_1cpi,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
        RecoChannelInfo(
            name="B_to_D_mcpi",
            ch_index=1,
            reco_mode_info=TagRecoModeInfos.b_to_d_mcpi_tag_modes,
            components=tag_cal_fit_comps,
            observable_infos=tag_cal_per_tag_mode_fit_obs_setup.tag_cal_ptm_obs_info_d_mcpi,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
        RecoChannelInfo(
            name="B_to_ocD_pipi",
            ch_index=2,
            reco_mode_info=TagRecoModeInfos.b_to_ocd_pipi_tag_modes,
            components=tag_cal_fit_comps,
            observable_infos=tag_cal_per_tag_mode_fit_obs_setup.tag_cal_ptm_obs_info_ocd_pipi,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
        RecoChannelInfo(
            name="B_to_Ds_1cpi",
            ch_index=3,
            reco_mode_info=TagRecoModeInfos.b_to_ds_1cpi_tag_modes,
            components=tag_cal_fit_comps,
            observable_infos=tag_cal_per_tag_mode_fit_obs_setup.tag_cal_ptm_obs_info_ds_1cpi,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
        RecoChannelInfo(
            name="B_to_Ds_mcpi",
            ch_index=4,
            reco_mode_info=TagRecoModeInfos.b_to_ds_mcpi_tag_modes,
            components=tag_cal_fit_comps,
            observable_infos=tag_cal_per_tag_mode_fit_obs_setup.tag_cal_ptm_obs_info_ds_mcpi,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
        RecoChannelInfo(
            name="B_to_JPsi",
            ch_index=5,
            reco_mode_info=TagRecoModeInfos.b_to_j_psi_tag_modes,
            components=tag_cal_fit_comps,
            observable_infos=tag_cal_per_tag_mode_fit_obs_setup.tag_cal_ptm_obs_info_j_psi,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
        RecoChannelInfo(
            name="B_to_DD_and_DsD",
            ch_index=6,
            reco_mode_info=TagRecoModeInfos.b_to__all_dd__tag_modes,
            components=tag_cal_fit_comps,
            observable_infos=tag_cal_per_tag_mode_fit_obs_setup.tag_cal_ptm_obs_info_all_dd,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
    ]
)


tag_cal_fit_tag_eff_corr_info: TagEfficiencyCorrectionInfo = TagEfficiencyCorrectionInfo(
    items=[],
    key_order=(),
)


_tag_cal_asimov_fit_setup = FitSetupInfoContainer(
    components=tag_cal_fit_comps,
    reco_channels=tag_cal_fit_chs,
    observable_infos=tag_cal_per_tag_mode_fit_obs_setup.tag_cal_ptm_obs_infos,
    weight_info=base_tag_cal_weight_info,
    data_setup_info=base_asimov_data_setup_info,
    remove_empty_bins=True,
    use_alternative_gap=True,
    tag_efficiency_correction_info=tag_cal_fit_tag_eff_corr_info,
    reco_mode_column_name=TagRecoModeInfos.fei_reco_mode_col_name,
    component_id_column_name=fei_tag_b_reco_based_fit_component_id_column_name,
    component_id_function=add_fei_tag_b_reco_based_fit_component_id,
)


tag_cal_fit_setup: FitSetupTriplet = FitSetupTriplet(asimov=_tag_cal_asimov_fit_setup)
