"""
Provides template setups for fit setups

Felix Metzner 2021
"""

from typing import Union, Optional, Dict, Tuple, NamedTuple

from templatefitter.plotter.plot_style import KITColors

from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses

from rdstar.offline_analysis.mc_matching.sig_id_costomizer_tools import SigIDAndGenIDCombiner
from rdstar.offline_analysis.mc_matching.mc_matching_utilities import MCSigIDNameToQueryMappingType
from rdstar.offline_analysis.mc_matching.mc_matching_id_collections import MCMatchingCollections, MCMatchingCollection

from rdstar.offline_analysis.fitting.fit_case import TemplateSetup
from rdstar.offline_analysis.fitting.special_fit_template_setups.decoupled_bkg_template_setup import (
    SimpleSignalSplitBkgTemplateSetupInfo,
)
from rdstar.offline_analysis.fitting.special_fit_template_setups.decoupled_d_star_modes_template_setup import (
    Q2SidebandSplitBkgTemplateSetupInfo,
    Q2SidebandSplitBkgTemplateSetupInfoUsingDF,
)


__all__ = [
    "RDStarFitSetupContainers",
]


OriginalBaseMCMatchingCollection: MCMatchingCollection = MCMatchingCollections.original__base__collection


class FitSetupContainer(NamedTuple):
    template_setup: Union[TemplateSetup, Dict[str, TemplateSetup]]
    plot_color_dict: Dict[str, str]
    plot_labels_dict: Dict[str, str]
    alternative_template_colors: Optional[Dict[str, str]] = None
    custom_sig_queries: Optional[MCSigIDNameToQueryMappingType] = None
    sig_id_customizer: Optional[SigIDAndGenIDCombiner] = None


class RDStarFitSetupContainers:
    __sig_id_matching = OriginalBaseMCMatchingCollection

    # Default Setup
    default_template_setup: Tuple[str, ...] = OriginalBaseMCMatchingCollection.plot_order
    default_plot_colors: Dict[str, str] = OriginalBaseMCMatchingCollection.get_group_name_to_color_mapping()
    default_plot_labels: Dict[str, str] = OriginalBaseMCMatchingCollection.get_group_name_to_latex_label_mapping()

    default_alt_temp_colors: Dict[str, str] = {
        KITColors.grey: KITColors.kit_purple,
        KITColors.light_grey: KITColors.kit_yellow,
    }

    default_setup = FitSetupContainer(
        template_setup=TemplateSetup(template_setup=default_template_setup),
        plot_color_dict=default_plot_colors,
        plot_labels_dict=default_plot_labels,
        alternative_template_colors=default_alt_temp_colors,
    )

    #
    # Simple Signal Setup
    simple_signal_template_setup: Dict[str, Tuple[str, ...]] = {
        "continuum": ("CBKG",),
        "bbBkg": ("GBKG", "HBKG", "PiLNu"),
        "DSSl": ("DSSlbkg", "DSSl"),
        "DSl": ("DSl",),
        "Dl": ("Dl",),
        "DStau": ("DStau",),
        "Dtau": ("Dtau",),
    }
    simple_signal_plot_colors: Dict[str, str] = {
        "Dtau": KITColors.kit_green,
        "DStau": KITColors.kit_maygreen,
        "Dl": KITColors.kit_blue,
        "DSl": KITColors.kit_cyan,
        "DSSl": KITColors.kit_red,
        "bbBkg": KITColors.grey,
        "continuum": KITColors.light_grey,
    }
    simple_signal_plot_labels: Dict[str, str] = {
        "Dtau": LatexDecayProcesses.to_d_tau_nu,
        "DStau": LatexDecayProcesses.to_d_st_tau_nu,
        "Dl": LatexDecayProcesses.to_d_l_nu,
        "DSl": LatexDecayProcesses.to_d_st_l_nu,
        "DSSl": LatexDecayProcesses.to_dstst_l_nu,
        "bbBkg": LatexDecayProcesses.to_b_bbar_bkg,
        "continuum": LatexDecayProcesses.to_continuum_txt,
    }
    simple_alt_temp_colors: Dict[str, str] = {
        KITColors.grey: KITColors.kit_orange,
        KITColors.light_grey: KITColors.kit_purple,
    }

    simple_signal_setup = FitSetupContainer(
        template_setup=TemplateSetup(template_setup=simple_signal_template_setup),
        plot_color_dict=simple_signal_plot_colors,
        plot_labels_dict=simple_signal_plot_labels,
        alternative_template_colors=simple_alt_temp_colors,
    )

    #
    # Sideband Setup
    sideband_template_setup: Dict[str, Tuple[str, ...]] = {
        "continuum": ("CBKG",),
        "bbBkg": ("GBKG", "HBKG", "PiLNu", "DSSlbkg", "DSSl", "DStau", "Dtau"),
        "DSl": ("DSl",),
        "Dl": ("Dl",),
    }
    sideband_plot_colors: Dict[str, str] = {
        "Dl": KITColors.kit_blue,
        "DSl": KITColors.kit_cyan,
        "bbBkg": KITColors.kit_red,
        "continuum": KITColors.grey,
    }
    sideband_plot_labels: Dict[str, str] = {
        "Dl": LatexDecayProcesses.to_d_l_nu,
        "DSl": LatexDecayProcesses.to_d_st_l_nu,
        "bbBkg": LatexDecayProcesses.to_b_bbar_bkg,
        "continuum": LatexDecayProcesses.to_continuum_txt,
    }
    sideband_alt_temp_colors: Dict[str, str] = {
        KITColors.grey: KITColors.kit_orange,
    }

    sideband_setup = FitSetupContainer(
        template_setup=TemplateSetup(template_setup=sideband_template_setup),
        plot_color_dict=sideband_plot_colors,
        plot_labels_dict=sideband_plot_labels,
        alternative_template_colors=sideband_alt_temp_colors,
    )

    #
    # Complex Signal Setup
    __dss_bins: int = 8
    complex_signal_template_setup: Dict[str, Tuple[str, ...]] = {
        "continuum": ("CBKG",),
        "bbBkg": ("GBKG", "HBKG", "PiLNu"),
        "DSSl": ("DSSl",),
        "DSSlbkg": ("DSSlbkg",),
        "DSl": ("DSl",),
        "Dl": ("Dl",),
        "DStau": ("DStau",),
        "Dtau": ("Dtau",),
    }
    complex_signal_plot_colors: Dict[str, str] = {
        "Dtau": KITColors.kit_green,
        "DStau": KITColors.kit_maygreen,
        "Dl": KITColors.kit_blue,
        "DSl": KITColors.kit_cyan,
        "DSSl": KITColors.kit_red,
        "DSSlbkg": KITColors.kit_orange,
        "bbBkg": KITColors.grey,
        "continuum": KITColors.light_grey,
    }
    complex_signal_plot_labels: Dict[str, str] = {
        "Dtau": LatexDecayProcesses.to_d_tau_nu,
        "DStau": LatexDecayProcesses.to_d_st_tau_nu,
        "Dl": LatexDecayProcesses.to_d_l_nu,
        "DSl": LatexDecayProcesses.to_d_st_l_nu,
        "DSSl": LatexDecayProcesses.to_dstst_l_nu_with_dstst_to_d_wwost_pi_zero,
        "DSSlbkg": LatexDecayProcesses.to_dstst_l_nu_with_dstst_to_d_wwost_pi_pm,
        "bbBkg": LatexDecayProcesses.to_b_bbar_bkg,
        "continuum": LatexDecayProcesses.to_continuum_txt,
    }
    complex_alt_temp_colors: Dict[str, str] = {
        KITColors.grey: KITColors.kit_brown,
        KITColors.light_grey: KITColors.kit_purple,
    }

    complex_signal_setup = FitSetupContainer(
        template_setup=TemplateSetup(template_setup=complex_signal_template_setup),
        plot_color_dict=complex_signal_plot_colors,
        plot_labels_dict=complex_signal_plot_labels,
        alternative_template_colors=complex_alt_temp_colors,
    )

    #
    # Simple Signal Setup with Decoupled BKGs
    simple_signal_split_bkg_alt_temp_colors: Dict[str, str] = {
        KITColors.grey: KITColors.kit_orange,
        KITColors.light_grey: KITColors.kit_purple,
    }

    simple_signal_split_bkg_setup = FitSetupContainer(
        template_setup=SimpleSignalSplitBkgTemplateSetupInfo.template_setup,
        plot_color_dict=SimpleSignalSplitBkgTemplateSetupInfo.plot_color_dict,
        plot_labels_dict=SimpleSignalSplitBkgTemplateSetupInfo.plot_labels_dict,
        alternative_template_colors=simple_signal_split_bkg_alt_temp_colors,
        custom_sig_queries=SimpleSignalSplitBkgTemplateSetupInfo.custom_sig_queries,
        sig_id_customizer=SimpleSignalSplitBkgTemplateSetupInfo.sig_id_customizer,
    )

    #
    # q2-Sideband Setup with Decoupled D* modes (regular and down-feed)
    q2_sideband_split_d_star_alt_temp_colors: Dict[str, str] = {
        KITColors.grey: KITColors.kit_orange,
        KITColors.light_grey: KITColors.kit_purple,
    }

    q2_sideband_split_d_star_setup = FitSetupContainer(
        template_setup=Q2SidebandSplitBkgTemplateSetupInfo.template_setup,
        plot_color_dict=Q2SidebandSplitBkgTemplateSetupInfo.plot_color_dict,
        plot_labels_dict=Q2SidebandSplitBkgTemplateSetupInfo.plot_labels_dict,
        alternative_template_colors=q2_sideband_split_d_star_alt_temp_colors,
        custom_sig_queries=Q2SidebandSplitBkgTemplateSetupInfo.custom_sig_queries,
        sig_id_customizer=Q2SidebandSplitBkgTemplateSetupInfo.sig_id_customizer,
    )

    # q2-Sideband Setup with Decoupled D* modes using Down-Feed component for normalization
    q2_sideband_split_d_star_inverted_alt_temp_colors: Dict[str, str] = {
        KITColors.grey: KITColors.kit_orange,
        KITColors.light_grey: KITColors.kit_purple,
    }

    q2_sideband_split_d_star_inverted_setup = FitSetupContainer(
        template_setup=Q2SidebandSplitBkgTemplateSetupInfoUsingDF.template_setup,
        plot_color_dict=Q2SidebandSplitBkgTemplateSetupInfoUsingDF.plot_color_dict,
        plot_labels_dict=Q2SidebandSplitBkgTemplateSetupInfoUsingDF.plot_labels_dict,
        alternative_template_colors=q2_sideband_split_d_star_inverted_alt_temp_colors,
        custom_sig_queries=Q2SidebandSplitBkgTemplateSetupInfoUsingDF.custom_sig_queries,
        sig_id_customizer=Q2SidebandSplitBkgTemplateSetupInfoUsingDF.sig_id_customizer,
    )

    #
    # Mbc Sideband Setup
    mbc_sideband_template_setup: Dict[str, Tuple[str, ...]] = {
        "continuum": ("CBKG",),
        "bbBkg": ("GBKG", "HBKG", "PiLNu", "DSSlbkg", "DSSl", "DStau", "Dtau"),
        "DSl": ("DSl",),
        "Dl": ("Dl",),
    }
    mbc_sideband_plot_colors: Dict[str, str] = {
        "Dl": KITColors.kit_blue,
        "DSl": KITColors.kit_cyan,
        "bbBkg": KITColors.kit_red,
        "continuum": KITColors.grey,
    }
    mbc_sideband_plot_labels: Dict[str, str] = {
        "Dl": LatexDecayProcesses.to_d_l_nu,
        "DSl": LatexDecayProcesses.to_d_st_l_nu,
        "bbBkg": LatexDecayProcesses.to_b_bbar_bkg,
        "continuum": LatexDecayProcesses.to_continuum_txt,
    }
    mbc_sideband_alt_temp_colors: Dict[str, str] = {
        KITColors.grey: KITColors.kit_orange,
    }

    mbc_sideband_setup = FitSetupContainer(
        template_setup=TemplateSetup(template_setup=mbc_sideband_template_setup),
        plot_color_dict=mbc_sideband_plot_colors,
        plot_labels_dict=mbc_sideband_plot_labels,
        alternative_template_colors=mbc_sideband_alt_temp_colors,
    )

    #
    # Complex Mbc Sideband Setup
    complex_mbc_sideband_template_setup: Dict[str, Tuple[str, ...]] = {
        "continuum": ("CBKG",),
        "bbBkg": ("GBKG", "HBKG", "PiLNu"),
        "DSSl": ("DSSl",),
        "DSSlbkg": ("DSSlbkg",),
        "DSl": ("DSl",),
        "Dl": ("Dl",),
        "DStau": ("DStau",),
        "Dtau": ("Dtau",),
    }
    complex_mbc_sideband_plot_colors: Dict[str, str] = {
        "Dtau": KITColors.kit_green,
        "DStau": KITColors.kit_maygreen,
        "Dl": KITColors.kit_blue,
        "DSl": KITColors.kit_cyan,
        "DSSl": KITColors.kit_red,
        "DSSlbkg": KITColors.kit_orange,
        "bbBkg": KITColors.grey,
        "continuum": KITColors.light_grey,
    }
    complex_mbc_sideband_plot_labels: Dict[str, str] = {
        "Dtau": LatexDecayProcesses.to_d_tau_nu,
        "DStau": LatexDecayProcesses.to_d_st_tau_nu,
        "Dl": LatexDecayProcesses.to_d_l_nu,
        "DSl": LatexDecayProcesses.to_d_st_l_nu,
        "DSSl": LatexDecayProcesses.to_dstst_l_nu_with_dstst_to_d_wwost_pi_zero,
        "DSSlbkg": LatexDecayProcesses.to_dstst_l_nu_with_dstst_to_d_wwost_pi_pm,
        "bbBkg": LatexDecayProcesses.to_b_bbar_bkg,
        "continuum": LatexDecayProcesses.to_continuum_txt,
    }
    complex_mbc_sideband_alt_temp_colors: Dict[str, str] = {
        KITColors.grey: KITColors.kit_brown,
        KITColors.light_grey: KITColors.kit_purple,
    }

    complex_mbc_sideband_setup = FitSetupContainer(
        template_setup=TemplateSetup(template_setup=complex_mbc_sideband_template_setup),
        plot_color_dict=complex_mbc_sideband_plot_colors,
        plot_labels_dict=complex_mbc_sideband_plot_labels,
        alternative_template_colors=complex_mbc_sideband_alt_temp_colors,
    )
