"""
Definitions for Simple Signal Setup with Decoupled BKGs

Felix Metzner 2021
"""

import copy

from typing import Dict, Tuple

from templatefitter.plotter.plot_style import KITColors

from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses

from rdstar.offline_analysis.reco_modes.reco_mode_collections import BSigRecoModes

from rdstar.offline_analysis.mc_matching import default_sig_id_col_name
from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo
from rdstar.offline_analysis.mc_matching.mc_matching_utilities import MCSigIDNameToQueryMappingType
from rdstar.offline_analysis.mc_matching.sig_id_costomizer_tools import (
    SigIDAndGenIDCombiner,
    SigIDAndGenIDCombinerInputType,
)

from rdstar.offline_analysis.fitting.fit_case import TemplateSetup
from rdstar.offline_analysis.fitting.special_fit_template_setups.special_template_setup_info_base import (
    SpecialTemplateSetupInfo,
)


__all__ = [
    "SimpleSignalSplitBkgTemplateSetupInfo",
]


_base_simple_signal_split_bkg_template_setup: Dict[str, Tuple[str, ...]] = {
    "continuum": ("CBKG",),
    "bbBkg": ("GBKG", "HBKG", "PiLNu"),
    "DSSl": ("DSSlbkg", "DSSl"),
    "DSl": ("DSl",),
    "Dl": ("Dl",),
    "DStau": ("DStau",),
    "Dtau": ("Dtau",),
}

base_bkg_query_infos: Dict[str, Tuple[str, ...]] = {
    k: v for k, v in _base_simple_signal_split_bkg_template_setup.items() if k in ("continuum", "bbBkg")
}

main_reco_mode_keys: Tuple[str, ...] = tuple(v.name for v in BSigRecoModes.main_modes.reco_mode_infos)

simple_signal_split_bkg_custom_sig_id_map: SigIDAndGenIDCombinerInputType = {
    (f"{comp_name}_{dec_mode_name}", (dec_mode_index + 1) * 10000 + comp_index): {
        dec_mode_name: tuple(OldMCMatchingIDsQueryInfo.get_sig_id_mapping()[sub_comp].id for sub_comp in bkg_sub_comps)
    }
    for dec_mode_index, dec_mode_name in enumerate(main_reco_mode_keys)
    for comp_index, (comp_name, bkg_sub_comps) in enumerate(base_bkg_query_infos.items())
}

simple_signal_split_bkg_sig_customizer = SigIDAndGenIDCombiner(
    customizing_dict=simple_signal_split_bkg_custom_sig_id_map,
    sig_id_column_name=default_sig_id_col_name,
    decay_mode_dict=copy.copy(BSigRecoModes.main_modes.main_reco_mode_infos_mapping),
)
simple_signal_split_bkg_custom_sig_queries: MCSigIDNameToQueryMappingType = (
    simple_signal_split_bkg_sig_customizer.get_updated_component_dict(
        initial_sig_query_dict=copy.copy(OldMCMatchingIDsQueryInfo.get_sig_id_mapping())
    )
)

_simple_signal_split_bkg_template_setup: Dict[str, TemplateSetup] = {
    dec_mode_name: TemplateSetup(
        template_setup={
            f"{comp_name}_{dec_mode_name}"
            if comp_name in base_bkg_query_infos.keys()
            else comp_name: (f"{comp_name}_{dec_mode_name}",)
            if comp_name in base_bkg_query_infos.keys()
            else sub_comps
            for comp_name, sub_comps in _base_simple_signal_split_bkg_template_setup.items()
        }
    )
    for dec_mode_name in main_reco_mode_keys
}

_simple_signal_split_bkg_plot_colors: Dict[str, str] = {
    "Dtau": KITColors.kit_green,
    "DStau": KITColors.kit_maygreen,
    "Dl": KITColors.kit_blue,
    "DSl": KITColors.kit_cyan,
    "DSSl": KITColors.kit_red,
    "bbBkg": KITColors.grey,
    "continuum": KITColors.light_grey,
}
_simple_signal_split_bkg_plot_labels: Dict[str, str] = {
    "Dtau": LatexDecayProcesses.to_d_tau_nu,
    "DStau": LatexDecayProcesses.to_d_st_tau_nu,
    "Dl": LatexDecayProcesses.to_d_l_nu,
    "DSl": LatexDecayProcesses.to_d_st_l_nu,
    "DSSl": LatexDecayProcesses.to_dstst_l_nu,
    "bbBkg": LatexDecayProcesses.to_b_bbar_bkg,
    "continuum": LatexDecayProcesses.to_continuum_txt,
}

_simple_signal_split_bkg_plot_colors.update(
    {
        k: _simple_signal_split_bkg_plot_colors[k.split("_", 1)[0]]
        for ts in _simple_signal_split_bkg_template_setup.values()
        for k in ts.template_setup
    }
)

_simple_signal_split_bkg_plot_labels.update(
    {
        k: _simple_signal_split_bkg_plot_labels[k.split("_", 1)[0]]
        for ts in _simple_signal_split_bkg_template_setup.values()
        for k in ts.template_setup
    }
)

SimpleSignalSplitBkgTemplateSetupInfo = SpecialTemplateSetupInfo(
    template_setup=_simple_signal_split_bkg_template_setup,
    plot_color_dict=_simple_signal_split_bkg_plot_colors,
    plot_labels_dict=_simple_signal_split_bkg_plot_labels,
    custom_sig_queries=simple_signal_split_bkg_custom_sig_queries,
    sig_id_customizer=simple_signal_split_bkg_sig_customizer,
)
