"""
Extended mc matching sig id observables.

Felix Metzner 2023
"""

import numpy as np
import pandas as pd

from abc import ABC
from typing import Union, Tuple, List, Dict, Callable, Type, Optional

from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.observables.offline_observable_definition_base import OfflineObservableBase

from rdstar.offline_analysis.observables.common_observables import Cols

from rdstar.offline_analysis.mc_matching.mc_matching_utilities import MatchQueryInfo
from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo
from rdstar.offline_analysis.mc_matching.mc_matching_info import ExtendedMCMatchingIDsQueryInfo
from rdstar.offline_analysis.mc_matching.abstract_mc_matching_info import AbstractBaseMCMatchingIDsQueryInfo
from rdstar.offline_analysis.mc_matching.mc_matching_id_collections import MCMatchingCollection, MCMatchingCollections


__all__ = [
    "OldMCMatchingIDsInfo",
    "ExtendedSigIDInfo",
    "SigIDObservables",
]


logger = log_to_default(name=__name__)


class SigIDInfoBaseClass(OfflineObservableBase, ABC):
    _mc_matching_ids_query_info_class: Optional[Type[AbstractBaseMCMatchingIDsQueryInfo]] = None

    ignore_if_present: bool = False
    _other_collections: Tuple[MCMatchingCollection, ...] = ()

    @classmethod
    def get_mc_matching_ids_query_info_class(cls) -> Type[AbstractBaseMCMatchingIDsQueryInfo]:
        assert cls._mc_matching_ids_query_info_class is not None
        return cls._mc_matching_ids_query_info_class

    @classmethod
    def get_new_column_name(cls) -> str:
        return cls.get_mc_matching_ids_query_info_class().get_sig_col_name()

    @classmethod
    def get_required_variables(cls) -> Tuple[str, ...]:
        return cls.get_mc_matching_ids_query_info_class().get_required_columns()

    @classmethod
    def get_mc_matching_sig_id_mapping(cls) -> Dict[str, MatchQueryInfo]:
        return cls.get_mc_matching_ids_query_info_class().get_sig_id_mapping()

    @classmethod
    def get_other_collections(cls) -> Tuple[MCMatchingCollection, ...]:
        if hasattr(cls, "_other_collections"):
            other_collections = cls._other_collections
            assert isinstance(other_collections, tuple), type(other_collections)
            assert all(isinstance(c, MCMatchingCollection) for c in other_collections), (
                [type(c) for c in other_collections if not isinstance(c, MCMatchingCollection)],
            )
            return other_collections
        else:
            raise AttributeError(
                "Classes inheriting from SigIDInfoBaseClass must have the attribute '_other_collections', "
                "at least as empty tuple!"
            )

    @classmethod
    def calculate_new_column_values(
        cls,
        df: pd.DataFrame,
        check_range: bool = True,
    ) -> Union[pd.Series, np.array]:

        new_sig_id_values: np.array = np.full(shape=len(df), fill_value=np.nan)

        n_rows: int = len(df.index)
        n_matched_total: int = 0
        logger.info(f"Number of candidates: {n_rows}")

        for key, sig_id_query_info in cls.get_mc_matching_sig_id_mapping().items():
            logger.info(f"Processing Extended_SIG_ID {sig_id_query_info.id} query {key}...")

            query_fcn: Callable = sig_id_query_info.query
            sig_id_mask: np.ndarray = query_fcn(df)

            n_matching: int = np.sum(sig_id_mask)
            n_matched_total += n_matching
            n_remaining: int = n_rows - n_matched_total
            logger.info(
                f"Number of matching candidates: {n_matching} "
                f"(covered {n_matched_total} of {n_rows} = {n_matched_total / n_rows:1.2f}), "
                f"{n_remaining} rows remaining."
            )

            if not np.all(np.isnan(new_sig_id_values[sig_id_mask])):
                _duplicate_ids: List[int] = list(
                    np.unique(new_sig_id_values[sig_id_mask & (~np.isnan(new_sig_id_values))])
                )
                raise RuntimeError(
                    f"Overwriting Signal ID values of events with the SigID(s) {_duplicate_ids} "
                    f"with the current SigID {sig_id_query_info.id} for {key} events!"
                )

            new_sig_id_values[sig_id_mask] = sig_id_query_info.id

        check = np.sum(np.isnan(new_sig_id_values))
        if check != 0:
            raise RuntimeError(f"Sig ID queries do not cover all events. Number of Rows w/o ID: {check}")

        _all_sig_ids: Tuple[int, ...] = cls.get_mc_matching_ids_query_info_class().get_all_sig_ids()
        assert all(v in _all_sig_ids for v in np.unique(new_sig_id_values)), (
            [v for v in np.unique(new_sig_id_values) if v not in _all_sig_ids],
        )

        return new_sig_id_values

    @classmethod
    def add_sig_id_to_df_dict(
        cls,
        df_dict: Dict[str, Dict[str, pd.DataFrame]],
        add_other_collections: bool,
    ) -> None:
        for data_type in df_dict.keys():
            logger.info(f"Adding {cls.__name__} signal ids for data type: {data_type}")

            cls.add_sig_id_to_df(df=df_dict[data_type]["Y4S_ntuple"], add_other_collections=add_other_collections)

    @classmethod
    def add_sig_id_to_df(
        cls,
        df: pd.DataFrame,
        add_other_collections: bool,
        alternative_sig_id_column_name: Optional[str] = None,
    ) -> None:
        if alternative_sig_id_column_name is None:
            _sig_id_column_name: str = cls.get_new_column_name()
        else:
            _sig_id_column_name = alternative_sig_id_column_name

        if _sig_id_column_name in df.columns:
            if cls.ignore_if_present:
                logger.warning(
                    f"New signal ID column '{_sig_id_column_name}' for definition class '{cls.__name__}' "
                    f"is already in the data frame and {cls.__name__}.ignore_if_present is 'True'!"
                )
            else:
                raise RuntimeError(
                    f"New signal ID column '{_sig_id_column_name}' for definition class '{cls.__name__}' is "
                    f"already in the data frame and {cls.__name__}.ignore_if_present is 'False'."
                )

        new_values = cls.calculate_new_column_values(df=df)
        assert len(new_values) == len(df), (len(new_values), len(df), cls.__name__, _sig_id_column_name)
        df.loc[:, _sig_id_column_name] = new_values

        assert not pd.isnull(df[_sig_id_column_name]).any(), (cls.__name__, _sig_id_column_name)

        if add_other_collections:
            logger.info(f"Adding additional collections of MC matching IDs for definition {cls.__name__}...")
            cls.add_collections(df=df)

    @classmethod
    def add_collections(cls, df: pd.DataFrame) -> None:
        base_sig_id_col_name: str = cls.get_new_column_name()
        assert base_sig_id_col_name in df.columns, (cls.__name__, base_sig_id_col_name)
        base_sig_id_values: np.ndarray = df[base_sig_id_col_name].values

        for mc_sig_id_collection in cls.get_other_collections():
            new_coll_col: str = mc_sig_id_collection.sig_id_col_name
            assert new_coll_col not in df.columns, (new_coll_col, cls.__name__, base_sig_id_col_name)

            logger.info(f"\t- adding collection {mc_sig_id_collection.collection_name}...")

            new_collection_values: np.array = np.full(shape=len(df), fill_value=np.nan)

            for sig_id_group in mc_sig_id_collection.collection_definition:
                is_in_group: np.ndarray = np.isin(element=base_sig_id_values, test_elements=sig_id_group.sig_ids)
                new_collection_values[is_in_group] = sig_id_group.new_id

            check = np.sum(np.isnan(new_collection_values))
            if check != 0:
                raise RuntimeError(
                    f"Sig ID Collection {mc_sig_id_collection.collection_name} for {cls.__name__} "
                    f"do not cover all events. Number of Rows w/o ID: {check}"
                )
            assert len(new_collection_values) == len(df), (
                len(new_collection_values),
                len(df),
                cls.__name__,
                mc_sig_id_collection.collection_name,
                new_coll_col,
            )

            df.loc[:, new_coll_col] = new_collection_values

            assert not pd.isnull(df[new_coll_col]).any(), (
                new_coll_col,
                mc_sig_id_collection.collection_name,
                cls.__name__,
                base_sig_id_col_name,
            )


class OldMCMatchingIDsInfo(SigIDInfoBaseClass):
    _mc_matching_ids_query_info_class: Type[AbstractBaseMCMatchingIDsQueryInfo] = OldMCMatchingIDsQueryInfo
    ignore_if_present: bool = True
    _other_collections: Tuple[MCMatchingCollection, ...] = ()


class ExtendedSigIDInfo(SigIDInfoBaseClass):
    _mc_matching_ids_query_info_class: Type[AbstractBaseMCMatchingIDsQueryInfo] = ExtendedMCMatchingIDsQueryInfo

    _other_collections: Tuple[MCMatchingCollection, ...] = (
        MCMatchingCollections.extended_sig_id__fine__collection,
        MCMatchingCollections.extended_sig_id__simple__collection,
        MCMatchingCollections.extended_sig_id__basic__collection,
        MCMatchingCollections.extended_sig_id__charge__collection,
    )


class SigIDObservables:

    observable_classes: Tuple[Type[SigIDInfoBaseClass], ...] = (
        # New main Sig ID Definition
        ExtendedSigIDInfo,
        OldMCMatchingIDsInfo,
    )

    definitions: Dict[str, Type[SigIDInfoBaseClass]] = {c.get_new_column_name(): c for c in observable_classes}

    required_variables: Tuple[str, ...] = tuple(set([c for d in definitions.values() for c in d.get_required_columns()]))

    @staticmethod
    def add_sig_id_observables_definitions(df: pd.DataFrame, check_range: bool = True) -> None:
        assert all(c in df.columns for c in SigIDObservables.get_required_columns_list()), [
            c for c in SigIDObservables.get_required_columns_list() if c not in df.columns
        ]

        expected_new_cols: List[str] = [
            c
            for c in SigIDObservables.get_new_columns_list()
            if c not in SigIDObservables.get_ignorable_new_columns_list()
        ]
        assert all(c not in df.columns for c in expected_new_cols), [c for c in expected_new_cols if c in df.columns]

        for new_col, definition_class in SigIDObservables.definitions.items():
            if new_col == OldMCMatchingIDsInfo.get_new_column_name():
                if Cols.reco_lepton_truth_matched_first_b_daughter_pdg_via_tau not in df.columns:
                    assert new_col in df.columns, new_col
                    logger.warning(
                        f"Skipping SigMC matching definition `OldMCMatchingIDsInfo`, as a required column is missing... "
                        f"However, the SigID should already be available in the column {new_col}!"
                    )
                    continue
            definition_class.add_sig_id_to_df(df=df, add_other_collections=True)

    @staticmethod
    def get_definition_function(
        name: str,
        fail_when_exists: bool = True,
    ) -> Callable:
        def definition_function(
            df: pd.DataFrame,
            check_range: bool = True,
            accept_existing_column: bool = False,
        ) -> None:
            definition: Type[SigIDInfoBaseClass] = SigIDObservables.definitions[name]
            new_col_name: str = definition.get_new_column_name()

            if fail_when_exists and not accept_existing_column:
                assert new_col_name not in df.columns, new_col_name
            else:
                if new_col_name in df.columns:
                    return

            assert all(c in df.columns for c in definition.get_required_columns()), [
                c for c in definition.get_required_columns() if c not in df.columns
            ]

            if len(np.unique(df.index)) == 0:
                df = df.assign(**{new_col_name: df[definition.get_required_columns()[0]]})
            else:
                new_values = definition.calculate_new_column_values(df=df)
                assert len(new_values) == len(df), (len(new_values), len(df), new_col_name)
                df.loc[:, new_col_name] = new_values

            assert not pd.isnull(df[new_col_name]).any(), new_col_name

        return definition_function

    @staticmethod
    def get_required_columns_list() -> List[str]:
        return list(SigIDObservables.required_variables)

    @staticmethod
    def get_new_columns_list() -> List[str]:
        return list(SigIDObservables.definitions.keys())

    @staticmethod
    def get_ignorable_new_columns_list() -> List[str]:
        return [k for k, v in SigIDObservables.definitions.items() if v.ignore_if_present]

    @staticmethod
    def get_sig_id_obs_definition(name: str) -> Type[SigIDInfoBaseClass]:
        return SigIDObservables.definitions[name]

    @staticmethod
    def get_definition_names() -> Tuple[str, ...]:
        return tuple(SigIDObservables.definitions.keys())
