"""
Collections of PDG-Codes
"""

from typing import Tuple, Dict

__all__ = [
    "PDGCodeCollection",
]


class PDGCodeCollection:

    nothing_pdg_id: int = -999
    photon_pdg_id: int = 22

    # Leptons
    electron_pdg_id: int = 11
    muon_pdg_id: int = 13
    tau_pdg_id: int = 15

    electron_plus_pdg_id: int = -electron_pdg_id
    muon_plus_pdg_id: int = -muon_pdg_id
    tau_plus_pdg_id: int = -tau_pdg_id

    nu_electron_pdg_id: int = 12
    nu_muon_pdg_id: int = 14
    nu_tau_pdg_id: int = 16

    lepton_to_nu_pdg_mapping: Dict[int, int] = {
        electron_pdg_id: nu_electron_pdg_id,
        muon_pdg_id: nu_muon_pdg_id,
        tau_pdg_id: nu_tau_pdg_id,
    }

    light_lepton_pdg_ids: Tuple[int, ...] = (electron_pdg_id, muon_pdg_id)
    nu_light_lepton_pdg_ids: Tuple[int, ...] = (nu_electron_pdg_id, nu_muon_pdg_id)

    all_lepton_pdg_ids: Tuple[int, ...] = light_lepton_pdg_ids + (tau_pdg_id,)
    all_neutrino_pdg_ids: Tuple[int, ...] = (nu_electron_pdg_id, nu_muon_pdg_id, nu_tau_pdg_id)
    all_lepton_with_nus_pdg_ids: Tuple[int, ...] = all_lepton_pdg_ids + all_neutrino_pdg_ids

    # D Mesons
    d_neutral_pdg_id: int = 421
    d_charged_pdg_id: int = 411
    d_plus_pdg_id: int = d_charged_pdg_id
    d_minus_pdg_id: int = -d_plus_pdg_id

    d_pdg_ids: Tuple[int, ...] = (d_charged_pdg_id, d_neutral_pdg_id)

    # D* Mesons
    dst_neutral_pdg_id: int = 423
    dst_charged_pdg_id: int = 413
    dst_plus_pdg_id: int = dst_charged_pdg_id
    dst_minus_pdg_id: int = -dst_plus_pdg_id

    dst_pdg_ids: Tuple[int, ...] = (dst_charged_pdg_id, dst_neutral_pdg_id)

    # D** Mesons
    # D_1 Mesons (2420) (Narrow D** Meson)
    done_neutral_pdg_id: int = 10423
    done_charged_pdg_id: int = 10413
    done_plus_pdg_id: int = done_charged_pdg_id
    done_minus_pdg_id: int = -done_plus_pdg_id

    done_pdg_ids: Tuple[int, ...] = (done_charged_pdg_id, done_neutral_pdg_id)

    # D*_2 Mesons (2460) (Narrow D** Meson)
    dtwostar_neutral_pdg_id: int = 425
    dtwostar_charged_pdg_id: int = 415
    dtwostar_plus_pdg_id: int = dtwostar_charged_pdg_id
    dtwostar_minus_pdg_id: int = -dtwostar_plus_pdg_id

    dtwostar_pdg_ids: Tuple[int, ...] = (dtwostar_charged_pdg_id, dtwostar_neutral_pdg_id)

    # D*_0 Mesons (2400) (Wide D** Meson)
    dzerostar_neutral_pdg_id: int = 10421
    dzerostar_charged_pdg_id: int = 10411
    dzerostar_plus_pdg_id: int = dzerostar_charged_pdg_id
    dzerostar_minus_pdg_id: int = -dzerostar_plus_pdg_id

    dzerostar_pdg_ids: Tuple[int, ...] = (dzerostar_charged_pdg_id, dzerostar_neutral_pdg_id)

    # D'_1 Mesons (H, 2430) (Wide D** Meson)
    dprimeone_neutral_pdg_id: int = 20423
    dprimeone_charged_pdg_id: int = 20413
    dprimeone_plus_pdg_id: int = dprimeone_charged_pdg_id
    dprimeone_minus_pdg_id: int = -dprimeone_plus_pdg_id

    dprimeone_pdg_ids: Tuple[int, ...] = (dprimeone_charged_pdg_id, dprimeone_neutral_pdg_id)

    # D Mesons Combinations
    d_wwo_st_pdg_ids: Tuple[int, ...] = d_pdg_ids + dst_pdg_ids
    neutral_d_wwo_st_pdg_ids: Tuple[int, ...] = (d_neutral_pdg_id, dst_neutral_pdg_id)
    charged_d_wwo_st_pdg_ids: Tuple[int, ...] = (d_charged_pdg_id, dst_charged_pdg_id)

    dstst_pdg_ids: Tuple[int, ...] = done_pdg_ids + dtwostar_pdg_ids + dzerostar_pdg_ids + dprimeone_pdg_ids
    dstst_narrow_mass_pdg_ids: Tuple[int, ...] = done_pdg_ids + dtwostar_pdg_ids
    dstst_wide_mass_pdg_ids: Tuple[int, ...] = dzerostar_pdg_ids + dprimeone_pdg_ids

    neutral_dstst_narrow_mass_pdg_ids: Tuple[int, ...] = (done_neutral_pdg_id, dtwostar_neutral_pdg_id)
    charged_dstst_narrow_mass_pdg_ids: Tuple[int, ...] = (done_charged_pdg_id, dtwostar_charged_pdg_id)
    neutral_dstst_wide_mass_pdg_ids: Tuple[int, ...] = (dzerostar_neutral_pdg_id, dprimeone_neutral_pdg_id)
    charged_dstst_wide_mass_pdg_ids: Tuple[int, ...] = (dzerostar_charged_pdg_id, dprimeone_charged_pdg_id)

    neutral_dstst_pdg_ids: Tuple[int, ...] = neutral_dstst_narrow_mass_pdg_ids + neutral_dstst_wide_mass_pdg_ids
    charged_dstst_pdg_ids: Tuple[int, ...] = charged_dstst_narrow_mass_pdg_ids + charged_dstst_wide_mass_pdg_ids

    all_d_pdg_ids: Tuple[int, ...] = d_pdg_ids + dst_pdg_ids + dstst_pdg_ids

    all_neutral_d_pdg_ids: Tuple[int, ...] = neutral_d_wwo_st_pdg_ids + neutral_dstst_pdg_ids
    all_charged_d_pdg_ids: Tuple[int, ...] = charged_d_wwo_st_pdg_ids + charged_dstst_pdg_ids

    # Additional Higher Excitation D Mesons

    higher_excitation_d2s_pdg_ids: Tuple[int, ...] = (
        100411,  # D(2S)+
        100421,  # D(2S)0
        100413,  # D*(2S)+
        100423,  # D*(2S)0
    )

    # B Mesons
    b_meson_neutral_pdg_id: int = 511
    b_meson_charged_pdg_id: int = 521
    b_meson_plus_pdg_id: int = b_meson_charged_pdg_id
    b_meson_minus_pdg_id: int = -b_meson_plus_pdg_id

    b_meson_pdg_ids: Tuple[int, ...] = (b_meson_neutral_pdg_id, b_meson_charged_pdg_id)

    # Pions
    pi_zero_pdg_id: int = 111
    anti_pi_zero_pdg_id: int = -pi_zero_pdg_id
    pi_neutral_pdg_id: int = pi_zero_pdg_id

    pi_charged_pdg_id: int = 211
    pi_plus_pdg_id: int = pi_charged_pdg_id
    pi_minus_pdg_id: int = -pi_charged_pdg_id

    pion_pdg_ids: Tuple[int, ...] = (pi_zero_pdg_id, pi_charged_pdg_id)

    # Utility Codes
    photon_or_nothing_pdg_ids: Tuple[int, ...] = (photon_pdg_id, nothing_pdg_id, abs(nothing_pdg_id))
    valid_forth_daughter: Tuple[int, ...] = photon_or_nothing_pdg_ids
