"""
Definition of Signal and Sideband Selection Setups

This file covers the definitions of
    - Selection cuts
    - Best candidate selection
for
    - the main selection
    - the q2 sideband
    - the mbc sideband
    - the selection for the form factor study
    - the dedicated D** selection

Felix Metzner 2019
"""
import copy
from typing import Optional, List, Tuple, Dict, Generator, NamedTuple, Type, ClassVar

from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames

from rdstar.mva_selection import mva_selector_setups
from rdstar.cut_information_objects.cut_info import CutInfo
from rdstar.mva_selection.mva_selector_setups import MVASelectorSetups

from rdstar.offline_analysis.histogram_variables import RDStarHistVars
from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.fitting.mva_cut_info_definitions import MVACutSetups
from rdstar.offline_analysis.reco_modes.reco_mode_collections import AllBMesonRecoModes
from rdstar.offline_analysis.selection.selection_info_containers import (
    SelectionSetupBase,
    BCSVariableInfo,
    MVAInfo,
    SubSelectionInfo,
    SelectionSpecifics,
)


__all__ = [
    "SelectionSetups",
    "LooseMVASelection",
]


# Definition of my first version of the selection:
class FirstSelection(SelectionSetupBase):
    selection_name: str = "first_selection"

    # Default Cuts
    fei_prob_cut = CutInfo(
        cut_str="daughter__bo0__cm__spextraInfo__boSignalProbability__bc__bc > 0.001",
        cut_label=LatexObsNames.fei_prob[:-1] + r" > 0.001$",
        cut_name="fei_prob_cut",
        cut_var="daughter__bo0__cm__spextraInfo__boSignalProbability__bc__bc",
        cut_var_latex_label=LatexObsNames.fei_prob,
        cut_type=">",
        cut_limits=0.001,
        latex_unit=LatexUnits.no_unit,
    )
    q_square_cut = CutInfo(
        cut_str="qSquaredReco > 4.",
        cut_label=LatexObsNames.fitting_q2[:-1] + r" > 4.0\," + LatexUnits.gev_sq[1:],
        cut_name="q_square_cut",
        cut_var="qSquaredReco",
        cut_var_latex_label=LatexObsNames.fitting_q2,
        cut_type=">",
        cut_limits=4.0,
        latex_unit=LatexUnits.gev_sq,
    )
    e_extra_cut = CutInfo(
        cut_str="ROE_eextra__boCleanROE__bc < 3.",
        cut_label=LatexObsNames.roe_e_ecl_extra[:-1] + r" < 3.0\," + LatexUnits.gev[1:],
        cut_name="e_extra_cut",
        cut_var="ROE_eextra__boCleanROE__bc",
        cut_var_latex_label=LatexObsNames.roe_e_ecl_extra,
        cut_type="<",
        cut_limits=3.0,
        latex_unit=LatexUnits.gev,
    )
    d_mass_cut = CutInfo(
        cut_str="abs(daughter__bo1__cm__spdaughter__bo0__cm__spdM__bc__bc) < 0.035",
        cut_label=LatexObsNames.d_w_wo_st_abs_delta_m[:-1] + r" < 35\," + LatexUnits.mev[1:],
        cut_name="d_mass_cut",
        cut_var="daughter__bo1__cm__spdaughter__bo0__cm__spdM__bc__bc",
        cut_var_latex_label=LatexObsNames.d_w_wo_st_abs_delta_m,
        cut_type="<",
        cut_limits=0.035,
        cut_abs=True,
        latex_unit=LatexUnits.gev,
    )
    tag_mbc_cut = CutInfo(
        cut_str="daughter__bo0__cm__spMbc__bc > 5.27",
        cut_label=LatexObsNames.b_tag_mbc[:-1] + r" > 5.27\," + LatexUnits.gev[1:],
        cut_name="tag_mbc_cut",
        cut_var="daughter__bo0__cm__spMbc__bc",
        cut_var_latex_label=LatexObsNames.b_tag_mbc,
        cut_type=">",
        cut_limits=5.27,
        latex_unit=LatexUnits.gev,
    )
    tag_delta_e_cut = CutInfo(
        cut_str="-0.05 < daughter__bo0__cm__spdeltaE__bc < 0.1",
        cut_label=LatexObsNames.b_tag_delta_e[:-1] + r"\;\in\;(-0.05, 0.1)\," + LatexUnits.gev[1:],
        cut_name="tag_delta_e_cut",
        cut_var="daughter__bo0__cm__spdeltaE__bc",
        cut_var_latex_label=LatexObsNames.b_tag_delta_e,
        cut_type=("<", "<"),
        cut_limits=(-0.05, 0.1),
        latex_unit=LatexUnits.gev,
    )
    cs_mva_cut = CutInfo(
        cut_str="cs_mva_output > 0.5",
        cut_label=LatexObsNames.mva_cs_output[:-1] + r" > 0.5$",
        cut_name="cs_mva_cut",
        cut_var="cs_mva_output",
        cut_var_latex_label=LatexObsNames.mva_cs_output,
        cut_type=">",
        cut_limits=0.5,
        latex_unit=LatexUnits.no_unit,
    )
    lepton_p_cut = CutInfo(
        cut_str="daughter__bo1__cm__spdaughter__bo1__cm__spp__bc__bc < 2.",
        cut_label=LatexObsNames.lepton_p[:-1] + r" < 2\," + LatexUnits.gev[1:],
        cut_name="lepton_p_cut",
        cut_var="daughter__bo1__cm__spdaughter__bo1__cm__spp__bc__bc",
        cut_var_latex_label=LatexObsNames.lepton_p,
        cut_type="<",
        cut_limits=2.0,
        latex_unit=LatexUnits.gev,
    )
    b_vertex_fit_cut = CutInfo(
        cut_str="daughter__bo1__cm__spchiProb__bc > 0.0",
        cut_label=LatexObsNames.b_sig_vertex_prob[:-1] + r" > 0.0$",
        cut_name="b_vertex_fit_cut",
        cut_var="daughter__bo1__cm__spchiProb__bc",
        cut_var_latex_label=LatexObsNames.b_sig_vertex_prob,
        cut_type=">",
        cut_limits=0.0,
        latex_unit=LatexUnits.no_unit,
    )

    # Sideband Cuts
    q_square_anti_cut = CutInfo(
        cut_str="qSquaredReco <= 4.",
        cut_label=LatexObsNames.fitting_q2[:-1] + r" \leq 4.0\," + LatexUnits.gev_sq[1:],
        cut_name="q_square_anti_cut",
        cut_var="qSquaredReco",
        cut_var_latex_label=LatexObsNames.fitting_q2,
        cut_type="<=",
        cut_limits=4.0,
        latex_unit=LatexUnits.gev_sq,
    )
    tag_mbc_anti_cut = CutInfo(
        cut_str="daughter__bo0__cm__spMbc__bc <= 5.27",
        cut_label=LatexObsNames.b_tag_mbc[:-1] + r" \leq 5.27\," + LatexUnits.gev[1:],
        cut_name="tag_mbc_anti_cut",
        cut_var="daughter__bo0__cm__spMbc__bc",
        cut_var_latex_label=LatexObsNames.b_tag_mbc,
        cut_type="<=",
        cut_limits=5.27,
        latex_unit=LatexUnits.gev,
    )
    cs_mva_anti_cut = CutInfo(
        cut_str="cs_mva_output < 0.5",
        cut_label=LatexObsNames.mva_cs_output[:-1] + r" < 0.5$",
        cut_name="cs_mva_anti_cut",
        cut_var="cs_mva_output",
        cut_var_latex_label=LatexObsNames.mva_cs_output,
        cut_type="<",
        cut_limits=0.5,
        latex_unit=LatexUnits.no_unit,
    )

    cut_infos_for_validation: List[CutInfo] = [
        fei_prob_cut,
        q_square_cut,
        e_extra_cut,
        d_mass_cut,
        tag_mbc_cut,
        tag_delta_e_cut,
        cs_mva_cut,
        lepton_p_cut,
        b_vertex_fit_cut,
    ]

    all_cut_infos: List[CutInfo] = [
        *cut_infos_for_validation,
        q_square_anti_cut,
        tag_mbc_anti_cut,
        cs_mva_anti_cut,
    ]

    bcs_hist_vars: List[HistVariable] = [
        RDStarHistVars.neextra_fine,
        RDStarHistVars.tag_delta_e,
    ]

    # Definition of Best Candidate Selection
    @staticmethod
    def get_bcs_info() -> Tuple[BCSVariableInfo, BCSVariableInfo]:
        primary_bcs_var = BCSVariableInfo(
            bcs_variable=Cols.event_ne_extra,
            absolute_value=False,
            max=False,
        )
        secondary_bcs_var = BCSVariableInfo(
            bcs_variable=Cols.b_tag_delta_e,
            absolute_value=True,
            max=False,
        )
        assert primary_bcs_var.bcs_variable in [
            hv.df_label for hv in FirstSelection.bcs_hist_vars
        ], primary_bcs_var.bcs_variable
        assert secondary_bcs_var.bcs_variable in [
            hv.df_label for hv in FirstSelection.bcs_hist_vars
        ], secondary_bcs_var.bcs_variable
        return primary_bcs_var, secondary_bcs_var

    # Selection Definitions
    @classmethod
    def get_selection(cls, cs_tag: str) -> Generator[SubSelectionInfo, None, None]:
        # Main Selection
        selection_tag: str = f"{cls.selection_name}_with_CS_{cs_tag}"
        cut_list: Tuple[CutInfo, ...] = (
            FirstSelection.fei_prob_cut,
            FirstSelection.q_square_cut,
            FirstSelection.e_extra_cut,
            FirstSelection.d_mass_cut,
            FirstSelection.tag_mbc_cut,
            FirstSelection.tag_delta_e_cut,
            FirstSelection.cs_mva_cut,
            FirstSelection.lepton_p_cut,
            FirstSelection.b_vertex_fit_cut,
        )
        assert all(cut in cls.all_cut_infos for cut in cut_list), f"{selection_tag} contains ill-defined cuts!"

        yield SubSelectionInfo(
            selection_tag=selection_tag,
            cut_list=cut_list,
            post_bcs_cut_list=tuple(),
            check_cuts=(),
            check_bcs=False,
            run_evaluation=True,
        )

        sb_check_cut_list: Tuple[CutInfo, ...] = (
            FirstSelection.e_extra_cut,
            FirstSelection.tag_mbc_cut,
            FirstSelection.cs_mva_cut,
            FirstSelection.lepton_p_cut,
        )

        # Mbc Sideband Selection
        mbc_sb_tag: str = f"{selection_tag}_mbc_sideband"
        mbc_sb_cut_list = [FirstSelection.tag_mbc_anti_cut if c == FirstSelection.tag_mbc_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in mbc_sb_cut_list), f"{mbc_sb_tag} contains ill-defined cuts!"
        mbc_sb_check_cuts = tuple([c for c in sb_check_cut_list if c != FirstSelection.tag_mbc_cut])
        yield SubSelectionInfo(
            selection_tag=mbc_sb_tag,
            cut_list=tuple(mbc_sb_cut_list),
            post_bcs_cut_list=tuple(),
            check_cuts=mbc_sb_check_cuts,
            check_bcs=True,
            run_evaluation=False,
        )

        # q squared Sideband Selection
        q2_sb_tag: str = f"{selection_tag}_q2_sideband"
        q2_sb_cut_list = [FirstSelection.q_square_anti_cut if c == FirstSelection.q_square_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in q2_sb_cut_list), f"{q2_sb_tag} contains ill-defined cuts!"
        q2_sbc_check_cuts = tuple([c for c in sb_check_cut_list if c != FirstSelection.q_square_cut])
        yield SubSelectionInfo(
            selection_tag=q2_sb_tag,
            cut_list=tuple(q2_sb_cut_list),
            post_bcs_cut_list=tuple(),
            check_cuts=q2_sbc_check_cuts,
            check_bcs=True,
            run_evaluation=False,
        )

        # CS MVA Sideband Selection
        cs_mva_sb_tag: str = f"{selection_tag}_csmva_sideband"
        cs_mva_sb_cut_list = [FirstSelection.cs_mva_anti_cut if c == FirstSelection.cs_mva_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in cs_mva_sb_cut_list), f"{cs_mva_sb_tag} contains ill-defined cuts!"
        yield SubSelectionInfo(
            selection_tag=cs_mva_sb_tag,
            cut_list=tuple(cs_mva_sb_cut_list),
            post_bcs_cut_list=tuple(),
            check_cuts=(),
            check_bcs=False,
            run_evaluation=False,
        )


class MVASelection(SelectionSetupBase):
    _selection_name: str = "mva_selection"
    _mva_setup_name: Optional[str] = mva_selector_setups.SecondSelectionMVASetup.selection_mva_name
    selection_name: str = f"{_selection_name}_{_mva_setup_name}"
    _post_mva_cut: Optional[float] = 0.5

    cut_infos_for_validation: List[CutInfo] = [
        FirstSelection.fei_prob_cut,
        FirstSelection.q_square_cut,
        FirstSelection.e_extra_cut,
        FirstSelection.d_mass_cut,
        FirstSelection.tag_mbc_cut,
        FirstSelection.cs_mva_cut,
        FirstSelection.lepton_p_cut,
        FirstSelection.b_vertex_fit_cut,
    ]

    all_cut_infos: List[CutInfo] = [
        *cut_infos_for_validation,
        FirstSelection.q_square_anti_cut,
        FirstSelection.tag_mbc_anti_cut,
        FirstSelection.cs_mva_anti_cut,
    ]

    bcs_hist_vars: List[HistVariable] = FirstSelection.bcs_hist_vars

    @staticmethod
    def get_bcs_info() -> Tuple[BCSVariableInfo, BCSVariableInfo]:
        return FirstSelection.get_bcs_info()

    @classmethod
    def get_selection(cls, cs_tag: str) -> Generator[SubSelectionInfo, None, None]:
        # Main Selection
        selection_tag: str = f"{cls.selection_name}_with_CS_{cs_tag}"
        cut_list: Tuple[CutInfo, ...] = (
            FirstSelection.fei_prob_cut,
            FirstSelection.q_square_cut,
            FirstSelection.e_extra_cut,
            FirstSelection.d_mass_cut,
            FirstSelection.tag_mbc_cut,
            FirstSelection.cs_mva_cut,
            FirstSelection.lepton_p_cut,
            FirstSelection.b_vertex_fit_cut,
        )
        assert all(cut in cls.all_cut_infos for cut in cut_list), f"{selection_tag} contains ill-defined cuts!"

        yield SubSelectionInfo(
            selection_tag=selection_tag,
            cut_list=cut_list,
            post_bcs_cut_list=tuple(),
            check_cuts=(),
            check_bcs=False,
            run_evaluation=True,
        )

        sb_check_cut_list: Tuple[CutInfo, ...] = (
            FirstSelection.e_extra_cut,
            FirstSelection.tag_mbc_cut,
            FirstSelection.cs_mva_cut,
            FirstSelection.lepton_p_cut,
        )

        # q squared Sideband Selection
        q2_sb_tag: str = f"{selection_tag}_q2_sideband"
        q2_sb_cut_list = [FirstSelection.q_square_anti_cut if c == FirstSelection.q_square_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in q2_sb_cut_list), f"{q2_sb_tag} contains ill-defined cuts!"
        q2_sbc_check_cuts = tuple([c for c in sb_check_cut_list if c != FirstSelection.q_square_cut])
        yield SubSelectionInfo(
            selection_tag=q2_sb_tag,
            cut_list=tuple(q2_sb_cut_list),
            post_bcs_cut_list=tuple(),
            check_cuts=q2_sbc_check_cuts,
            check_bcs=True,
            run_evaluation=False,
        )

        # Mbc Sideband Selection
        mbc_sb_tag: str = f"{selection_tag}_mbc_sideband"
        mbc_sb_cut_list = [FirstSelection.tag_mbc_anti_cut if c == FirstSelection.tag_mbc_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in mbc_sb_cut_list), f"{mbc_sb_tag} contains ill-defined cuts!"
        mbc_sbc_check_cuts = tuple([c for c in sb_check_cut_list if c != FirstSelection.tag_mbc_cut])
        yield SubSelectionInfo(
            selection_tag=mbc_sb_tag,
            cut_list=tuple(mbc_sb_cut_list),
            post_bcs_cut_list=tuple(),
            check_cuts=mbc_sbc_check_cuts,
            check_bcs=True,
            run_evaluation=False,
        )

        # CS MVA Sideband Selection
        cs_mva_sb_tag: str = f"{selection_tag}_csmva_sideband"
        cs_mva_sb_cut_list = [FirstSelection.cs_mva_anti_cut if c == FirstSelection.cs_mva_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in cs_mva_sb_cut_list), f"{cs_mva_sb_tag} contains ill-defined cuts!"
        csmva_sbc_check_cuts = tuple([c for c in sb_check_cut_list if c != FirstSelection.cs_mva_cut])
        yield SubSelectionInfo(
            selection_tag=cs_mva_sb_tag,
            cut_list=tuple(cs_mva_sb_cut_list),
            post_bcs_cut_list=tuple(),
            check_cuts=csmva_sbc_check_cuts,
            check_bcs=False,
            run_evaluation=False,
        )


class LooseMVASelection(MVASelection):
    _selection_name: str = "loose_mva_selection"
    _mva_setup_name: Optional[str] = mva_selector_setups.SecondSelectionMVASetup.selection_mva_name
    selection_name: str = f"{_selection_name}_{_mva_setup_name}"
    _post_mva_cut: Optional[float] = 0.2

    cut_infos_for_validation: List[CutInfo] = [
        FirstSelection.fei_prob_cut,
        FirstSelection.q_square_cut,
        FirstSelection.e_extra_cut,
        FirstSelection.d_mass_cut,
        FirstSelection.tag_mbc_cut,
        FirstSelection.cs_mva_cut,
        FirstSelection.lepton_p_cut,
        FirstSelection.b_vertex_fit_cut,
    ]

    all_cut_infos: List[CutInfo] = [
        *cut_infos_for_validation,
        FirstSelection.q_square_anti_cut,
        FirstSelection.tag_mbc_anti_cut,
        FirstSelection.cs_mva_anti_cut,
    ]

    bcs_hist_vars: List[HistVariable] = FirstSelection.bcs_hist_vars


class OptimizedMVASelection(LooseMVASelection):
    _selection_name: str = "optimized_mva_selection"
    _mva_setup_name: Optional[str] = mva_selector_setups.SecondSelectionMVASetup.selection_mva_name
    assert _mva_setup_name is not None
    selection_name: str = f"{_selection_name}_{_mva_setup_name}"
    _post_mva_cut: Optional[float] = 0.575
    _use_post_cut_if_defined: bool = False

    __mva_setup = MVASelectorSetups.get_selection_mva_setup(setup_name=_mva_setup_name)
    __mva_output_col_name: str = __mva_setup.get_selection_mva_output_column_name()
    __mva_col: str = __mva_output_col_name

    optimized_mva_cut_setup = MVACutSetups.get_best_cuts_info(setup=MVACutSetups.reco_mode_2d_setup)
    assert __mva_col in optimized_mva_cut_setup.cut_variables, (__mva_col, optimized_mva_cut_setup.cut_variables)
    assert optimized_mva_cut_setup.reco_mode_column_name == Cols.decay_mode_column_name, (
        optimized_mva_cut_setup.reco_mode_column_name,
        Cols.decay_mode_column_name,
    )

    __reco_mode_id_map: Dict[str, Tuple[int, ...]] = {
        ri.name: ri.ids
        for ri in AllBMesonRecoModes.b_to_all_d_l_modes_by_b_charge_d_excitation_and_pi_charge.reco_mode_infos
    }

    optimized_mva_cut = CutInfo(
        cut_str=optimized_mva_cut_setup.get_cut_query_str(fit_case_reco_mode_id_map=__reco_mode_id_map),
        cut_label=f"Opt. {LatexObsNames.mva_sel_output_short} Crit.",
        cut_name=f"{__mva_output_col_name}_cut",
        cut_var=__mva_output_col_name,
        cut_var_latex_label=LatexObsNames.mva_sel_output,
        cut_type="complex",
        cut_limits=0.0,
        latex_unit=None,
    )

    cut_infos_for_validation: List[CutInfo] = [
        FirstSelection.fei_prob_cut,
        FirstSelection.q_square_cut,
        FirstSelection.e_extra_cut,
        FirstSelection.d_mass_cut,
        FirstSelection.tag_mbc_cut,
        FirstSelection.cs_mva_cut,
        FirstSelection.lepton_p_cut,
        FirstSelection.b_vertex_fit_cut,
        optimized_mva_cut,
    ]

    all_cut_infos: List[CutInfo] = [
        *cut_infos_for_validation,
        FirstSelection.q_square_anti_cut,
        FirstSelection.tag_mbc_anti_cut,
        FirstSelection.cs_mva_anti_cut,
    ]

    bcs_hist_vars: List[HistVariable] = FirstSelection.bcs_hist_vars

    @classmethod
    def get_selection(cls, cs_tag: str) -> Generator[SubSelectionInfo, None, None]:
        # Main Selection
        selection_tag: str = f"{cls.selection_name}_with_CS_{cs_tag}"
        cut_list: Tuple[CutInfo, ...] = (
            FirstSelection.fei_prob_cut,
            FirstSelection.q_square_cut,
            FirstSelection.e_extra_cut,
            FirstSelection.d_mass_cut,
            FirstSelection.tag_mbc_cut,
            FirstSelection.cs_mva_cut,
            FirstSelection.lepton_p_cut,
            FirstSelection.b_vertex_fit_cut,
            OptimizedMVASelection.optimized_mva_cut,
        )
        assert all(cut in cls.all_cut_infos for cut in cut_list), f"{selection_tag} contains ill-defined cuts!"

        yield SubSelectionInfo(
            selection_tag=selection_tag,
            cut_list=cut_list,
            post_bcs_cut_list=tuple(),
            check_cuts=(),
            check_bcs=False,
            run_evaluation=True,
        )

        sb_check_cut_list: Tuple[CutInfo, ...] = (
            FirstSelection.e_extra_cut,
            FirstSelection.tag_mbc_cut,
            FirstSelection.cs_mva_cut,
            FirstSelection.lepton_p_cut,
        )

        # q squared Sideband Selection
        q2_sb_tag: str = f"{selection_tag}_q2_sideband"
        q2_sb_cut_list = [FirstSelection.q_square_anti_cut if c == FirstSelection.q_square_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in q2_sb_cut_list), f"{q2_sb_tag} contains ill-defined cuts!"
        q2_sbc_check_cuts = tuple([c for c in sb_check_cut_list if c != FirstSelection.q_square_cut])
        yield SubSelectionInfo(
            selection_tag=q2_sb_tag,
            cut_list=tuple(q2_sb_cut_list),
            post_bcs_cut_list=tuple(),
            check_cuts=q2_sbc_check_cuts,
            check_bcs=True,
            run_evaluation=False,
        )

        # Mbc Sideband Selection
        mbc_sb_tag: str = f"{selection_tag}_mbc_sideband"
        mbc_sb_cut_list = [FirstSelection.tag_mbc_anti_cut if c == FirstSelection.tag_mbc_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in mbc_sb_cut_list), f"{mbc_sb_tag} contains ill-defined cuts!"
        mbc_sbc_check_cuts = tuple([c for c in sb_check_cut_list if c != FirstSelection.tag_mbc_cut])
        yield SubSelectionInfo(
            selection_tag=mbc_sb_tag,
            cut_list=tuple(mbc_sb_cut_list),
            post_bcs_cut_list=tuple(),
            check_cuts=mbc_sbc_check_cuts,
            check_bcs=True,
            run_evaluation=False,
        )

        # CS MVA Sideband Selection
        cs_mva_sb_tag: str = f"{selection_tag}_csmva_sideband"
        cs_mva_sb_cut_list = [FirstSelection.cs_mva_anti_cut if c == FirstSelection.cs_mva_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in cs_mva_sb_cut_list), f"{cs_mva_sb_tag} contains ill-defined cuts!"
        csmva_sbc_check_cuts = tuple([c for c in sb_check_cut_list if c != FirstSelection.cs_mva_cut])
        yield SubSelectionInfo(
            selection_tag=cs_mva_sb_tag,
            cut_list=tuple(cs_mva_sb_cut_list),
            post_bcs_cut_list=tuple(),
            check_cuts=csmva_sbc_check_cuts,
            check_bcs=False,
            run_evaluation=False,
        )


class FullQ2Selection(SelectionSetupBase):
    selection_name: str = "full_q2_selection"

    cut_infos_for_validation: List[CutInfo] = [
        FirstSelection.e_extra_cut,
        FirstSelection.d_mass_cut,
        FirstSelection.tag_mbc_cut,
        FirstSelection.cs_mva_cut,
        FirstSelection.b_vertex_fit_cut,
    ]

    all_cut_infos: List[CutInfo] = [
        *cut_infos_for_validation,
        FirstSelection.tag_mbc_anti_cut,
    ]

    bcs_hist_vars: List[HistVariable] = FirstSelection.bcs_hist_vars

    @staticmethod
    def get_bcs_info() -> Tuple[BCSVariableInfo, BCSVariableInfo]:
        return FirstSelection.get_bcs_info()

    @classmethod
    def get_selection(cls, cs_tag: str) -> Generator[SubSelectionInfo, None, None]:
        # Main Selection
        selection_tag: str = f"{cls.selection_name}_with_CS_{cs_tag}"
        cut_list: Tuple[CutInfo, ...] = (
            FirstSelection.e_extra_cut,
            FirstSelection.d_mass_cut,
            FirstSelection.tag_mbc_cut,
            FirstSelection.cs_mva_cut,
            FirstSelection.b_vertex_fit_cut,
        )
        assert all(cut in cls.all_cut_infos for cut in cut_list), f"{selection_tag} contains ill-defined cuts!"

        yield SubSelectionInfo(
            selection_tag=selection_tag,
            cut_list=cut_list,
            post_bcs_cut_list=tuple(),
            check_cuts=(),
            check_bcs=False,
            run_evaluation=True,
        )

        # Mbc Sideband Selection
        mbc_sb_tag = f"{selection_tag}_mbc_sideband"
        mbc_sb_cut_list = [FirstSelection.tag_mbc_anti_cut if c == FirstSelection.tag_mbc_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in mbc_sb_cut_list), f"{mbc_sb_tag} contains ill-defined cuts!"
        yield SubSelectionInfo(
            selection_tag=mbc_sb_tag,
            cut_list=tuple(mbc_sb_cut_list),
            post_bcs_cut_list=tuple(),
            check_cuts=(),
            check_bcs=False,
            run_evaluation=False,
        )


class DSSSelection(SelectionSetupBase):
    _selection_name: str = "dss_selection"
    _mva_setup_name: Optional[str] = mva_selector_setups.SecondSelectionDSSMVASetup.selection_mva_name
    selection_name: str = f"{_selection_name}_{_mva_setup_name}"
    _post_mva_cut: Optional[float] = 0.5

    cut_infos_for_validation: List[CutInfo] = [
        FirstSelection.fei_prob_cut,
        FirstSelection.e_extra_cut,
        FirstSelection.d_mass_cut,
        FirstSelection.tag_mbc_cut,
        FirstSelection.cs_mva_cut,
        FirstSelection.lepton_p_cut,
        FirstSelection.b_vertex_fit_cut,
    ]

    all_cut_infos: List[CutInfo] = [
        *cut_infos_for_validation,
        FirstSelection.tag_mbc_anti_cut,
    ]

    bcs_hist_vars: List[HistVariable] = FirstSelection.bcs_hist_vars

    @staticmethod
    def get_bcs_info() -> Tuple[BCSVariableInfo, BCSVariableInfo]:
        return FirstSelection.get_bcs_info()

    @classmethod
    def get_selection(cls, cs_tag: str) -> Generator[SubSelectionInfo, None, None]:
        # Main Selection
        selection_tag: str = f"{cls.selection_name}_with_CS_{cs_tag}"
        cut_list: Tuple[CutInfo, ...] = (
            FirstSelection.fei_prob_cut,
            FirstSelection.e_extra_cut,
            FirstSelection.d_mass_cut,
            FirstSelection.tag_mbc_cut,
            FirstSelection.cs_mva_cut,
            FirstSelection.lepton_p_cut,
            FirstSelection.b_vertex_fit_cut,
        )
        assert all(cut in cls.all_cut_infos for cut in cut_list), f"{selection_tag} contains ill-defined cuts!"

        yield SubSelectionInfo(
            selection_tag=selection_tag,
            cut_list=cut_list,
            post_bcs_cut_list=tuple(),
            check_cuts=(),
            check_bcs=False,
            run_evaluation=True,
        )

        # Mbc Sideband Selection
        mbc_sb_tag: str = f"{selection_tag}_mbc_sideband"
        mbc_sb_cut_list = [FirstSelection.tag_mbc_anti_cut if c == FirstSelection.tag_mbc_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in mbc_sb_cut_list), f"{mbc_sb_tag} contains ill-defined cuts!"
        yield SubSelectionInfo(
            selection_tag=mbc_sb_tag,
            cut_list=tuple(mbc_sb_cut_list),
            post_bcs_cut_list=tuple(),
            check_cuts=(),
            check_bcs=False,
            run_evaluation=False,
        )


class DSSSelectionLooseMVA(DSSSelection):
    _selection_name: str = "dss_selection_loose_mva"
    _mva_setup_name: Optional[str] = mva_selector_setups.SecondSelectionDSSMVASetup.selection_mva_name
    selection_name: str = f"{_selection_name}_{_mva_setup_name}"
    _post_mva_cut: Optional[float] = 0.2

    cut_infos_for_validation: List[CutInfo] = [
        FirstSelection.fei_prob_cut,
        FirstSelection.e_extra_cut,
        FirstSelection.d_mass_cut,
        FirstSelection.tag_mbc_cut,
        FirstSelection.cs_mva_cut,
        FirstSelection.lepton_p_cut,
        FirstSelection.b_vertex_fit_cut,
    ]

    all_cut_infos: List[CutInfo] = [
        *cut_infos_for_validation,
        FirstSelection.tag_mbc_anti_cut,
    ]

    bcs_hist_vars: List[HistVariable] = FirstSelection.bcs_hist_vars


class DSSSelectionFullPL(SelectionSetupBase):
    _selection_name: str = "dss_selection_full_pl"
    _mva_setup_name: Optional[str] = mva_selector_setups.SecondSelectionDSSMVASetup.selection_mva_name
    selection_name: str = f"{_selection_name}_{_mva_setup_name}"
    _post_mva_cut: Optional[float] = 0.5

    cut_infos_for_validation: List[CutInfo] = [
        FirstSelection.fei_prob_cut,
        FirstSelection.e_extra_cut,
        FirstSelection.d_mass_cut,
        FirstSelection.tag_mbc_cut,
        FirstSelection.cs_mva_cut,
        FirstSelection.b_vertex_fit_cut,
    ]

    all_cut_infos: List[CutInfo] = [
        *cut_infos_for_validation,
        FirstSelection.tag_mbc_anti_cut,
    ]

    bcs_hist_vars: List[HistVariable] = FirstSelection.bcs_hist_vars

    @staticmethod
    def get_bcs_info() -> Tuple[BCSVariableInfo, BCSVariableInfo]:
        return FirstSelection.get_bcs_info()

    @classmethod
    def get_selection(cls, cs_tag: str) -> Generator[SubSelectionInfo, None, None]:
        # Main Selection
        selection_tag: str = f"{cls.selection_name}_with_CS_{cs_tag}"
        cut_list: Tuple[CutInfo, ...] = (
            FirstSelection.fei_prob_cut,
            FirstSelection.e_extra_cut,
            FirstSelection.d_mass_cut,
            FirstSelection.tag_mbc_cut,
            FirstSelection.cs_mva_cut,
            FirstSelection.b_vertex_fit_cut,
        )
        assert all(cut in cls.all_cut_infos for cut in cut_list), f"{selection_tag} contains ill-defined cuts!"

        yield SubSelectionInfo(
            selection_tag=selection_tag,
            cut_list=cut_list,
            post_bcs_cut_list=tuple(),
            check_cuts=(),
            check_bcs=False,
            run_evaluation=True,
        )

        # Mbc Sideband Selection
        mbc_sb_tag: str = f"{selection_tag}_mbc_sideband"
        mbc_sb_cut_list = [FirstSelection.tag_mbc_anti_cut if c == FirstSelection.tag_mbc_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in mbc_sb_cut_list), f"{mbc_sb_tag} contains ill-defined cuts!"
        yield SubSelectionInfo(
            selection_tag=mbc_sb_tag,
            cut_list=tuple(mbc_sb_cut_list),
            post_bcs_cut_list=tuple(),
            check_cuts=(),
            check_bcs=False,
            run_evaluation=False,
        )


class DSSSelectionFullPLLooseMVA(DSSSelectionFullPL):
    _selection_name: str = "dss_selection_full_pl_loose_mva"
    _mva_setup_name: Optional[str] = mva_selector_setups.SecondSelectionDSSMVASetup.selection_mva_name
    selection_name: str = f"{_selection_name}_{_mva_setup_name}"
    _post_mva_cut: Optional[float] = 0.2

    cut_infos_for_validation: List[CutInfo] = [
        FirstSelection.fei_prob_cut,
        FirstSelection.e_extra_cut,
        FirstSelection.d_mass_cut,
        FirstSelection.tag_mbc_cut,
        FirstSelection.cs_mva_cut,
        FirstSelection.b_vertex_fit_cut,
    ]

    all_cut_infos: List[CutInfo] = [
        *cut_infos_for_validation,
        FirstSelection.tag_mbc_anti_cut,
    ]

    bcs_hist_vars: List[HistVariable] = FirstSelection.bcs_hist_vars


############################################
# Definition of main event selection setup #
############################################


class MainSelection(SelectionSetupBase):
    selection_name: str = "main_selection"

    # Default Cuts
    fei_prob_cut = CutInfo(
        cut_str="daughter__bo0__cm__spextraInfo__boSignalProbability__bc__bc > 0.001",
        cut_label=LatexObsNames.fei_prob[:-1] + r" > 0.001$",
        cut_name="fei_prob_cut",
        cut_var="daughter__bo0__cm__spextraInfo__boSignalProbability__bc__bc",
        cut_var_latex_label=LatexObsNames.fei_prob,
        cut_type=">",
        cut_limits=0.001,
        latex_unit=LatexUnits.no_unit,
    )
    tag_mbc_cut = CutInfo(
        cut_str="daughter__bo0__cm__spMbc__bc > 5.27",
        cut_label=LatexObsNames.b_tag_mbc[:-1] + r" > 5.27\," + LatexUnits.gev[1:],
        cut_name="tag_mbc_cut",
        cut_var="daughter__bo0__cm__spMbc__bc",
        cut_var_latex_label=LatexObsNames.b_tag_mbc,
        cut_type=">",
        cut_limits=5.27,
        latex_unit=LatexUnits.gev,
    )
    e_extra_cut = CutInfo(
        cut_str="ROE_eextra__boCleanROE__bc < 3.",
        cut_label=LatexObsNames.roe_e_ecl_extra[:-1] + r" < 3.0\," + LatexUnits.gev[1:],
        cut_name="e_extra_cut",
        cut_var="ROE_eextra__boCleanROE__bc",
        cut_var_latex_label=LatexObsNames.roe_e_ecl_extra,
        cut_type="<",
        cut_limits=3.0,
        latex_unit=LatexUnits.gev,
    )
    d_mass_cut = CutInfo(
        cut_str="abs(daughter__bo1__cm__spdaughter__bo0__cm__spdM__bc__bc) < 0.035",
        cut_label=LatexObsNames.d_w_wo_st_abs_delta_m[:-1] + r" < 35\," + LatexUnits.mev[1:],
        cut_name="d_mass_cut",
        cut_var="daughter__bo1__cm__spdaughter__bo0__cm__spdM__bc__bc",
        cut_var_latex_label=LatexObsNames.d_w_wo_st_abs_delta_m,
        cut_type="<",
        cut_limits=0.035,
        cut_abs=True,
        latex_unit=LatexUnits.gev,
    )
    tag_delta_e_cut = CutInfo(
        cut_str="-0.05 < daughter__bo0__cm__spdeltaE__bc < 0.1",
        cut_label=LatexObsNames.b_tag_delta_e[:-1] + r"\;\in\;(-0.05, 0.1)\," + LatexUnits.gev[1:],
        cut_name="tag_delta_e_cut",
        cut_var="daughter__bo0__cm__spdeltaE__bc",
        cut_var_latex_label=LatexObsNames.b_tag_delta_e,
        cut_type=("<", "<"),
        cut_limits=(-0.05, 0.1),
        latex_unit=LatexUnits.gev,
    )
    cs_mva_cut = CutInfo(
        cut_str="cs_mva_output > 0.5",
        cut_label=LatexObsNames.mva_cs_output[:-1] + r" > 0.5$",
        cut_name="cs_mva_cut",
        cut_var="cs_mva_output",
        cut_var_latex_label=LatexObsNames.mva_cs_output,
        cut_type=">",
        cut_limits=0.5,
        latex_unit=LatexUnits.no_unit,
    )
    lepton_p_cut = CutInfo(
        cut_str="daughter__bo1__cm__spdaughter__bo1__cm__spp__bc__bc < 2.",
        cut_label=LatexObsNames.lepton_p[:-1] + r" < 2\," + LatexUnits.gev[1:],
        cut_name="lepton_p_cut",
        cut_var="daughter__bo1__cm__spdaughter__bo1__cm__spp__bc__bc",
        cut_var_latex_label=LatexObsNames.lepton_p,
        cut_type="<",
        cut_limits=2.0,
        latex_unit=LatexUnits.gev,
    )
    b_vertex_fit_cut = CutInfo(
        cut_str="daughter__bo1__cm__spchiProb__bc > 0.0",
        cut_label=LatexObsNames.b_sig_vertex_prob[:-1] + r" > 0.0$",
        cut_name="b_vertex_fit_cut",
        cut_var="daughter__bo1__cm__spchiProb__bc",
        cut_var_latex_label=LatexObsNames.b_sig_vertex_prob,
        cut_type=">",
        cut_limits=0.0,
        latex_unit=LatexUnits.no_unit,
    )

    # Post BCS Cuts
    q_square_cut = CutInfo(
        cut_str="qSquaredReco > 4.",
        cut_label=LatexObsNames.fitting_q2[:-1] + r" > 4.0\," + LatexUnits.gev_sq[1:],
        cut_name="q_square_cut",
        cut_var="qSquaredReco",
        cut_var_latex_label=LatexObsNames.fitting_q2,
        cut_type=">",
        cut_limits=4.0,
        latex_unit=LatexUnits.gev_sq,
    )

    # Sideband Cuts
    q_square_anti_cut = CutInfo(
        cut_str="qSquaredReco <= 4.",
        cut_label=LatexObsNames.fitting_q2[:-1] + r" \leq 4.0\," + LatexUnits.gev_sq[1:],
        cut_name="q_square_anti_cut",
        cut_var="qSquaredReco",
        cut_var_latex_label=LatexObsNames.fitting_q2,
        cut_type="<=",
        cut_limits=4.0,
        latex_unit=LatexUnits.gev_sq,
    )
    tag_mbc_anti_cut = CutInfo(
        cut_str="daughter__bo0__cm__spMbc__bc <= 5.27",
        cut_label=LatexObsNames.b_tag_mbc[:-1] + r" \leq 5.27\," + LatexUnits.gev[1:],
        cut_name="tag_mbc_anti_cut",
        cut_var="daughter__bo0__cm__spMbc__bc",
        cut_var_latex_label=LatexObsNames.b_tag_mbc,
        cut_type="<=",
        cut_limits=5.27,
        latex_unit=LatexUnits.gev,
    )
    cs_mva_anti_cut = CutInfo(
        cut_str="cs_mva_output < 0.5",
        cut_label=LatexObsNames.mva_cs_output[:-1] + r" < 0.5$",
        cut_name="cs_mva_anti_cut",
        cut_var="cs_mva_output",
        cut_var_latex_label=LatexObsNames.mva_cs_output,
        cut_type="<",
        cut_limits=0.5,
        latex_unit=LatexUnits.no_unit,
    )
    cs_mva_loose_anti_cut = CutInfo(
        cut_str="cs_mva_output < 0.85",
        cut_label=LatexObsNames.mva_cs_output[:-1] + r" < 0.85$",
        cut_name="cs_mva_loose_anti_cut",
        cut_var="cs_mva_output",
        cut_var_latex_label=LatexObsNames.mva_cs_output,
        cut_type="<",
        cut_limits=0.85,
        latex_unit=LatexUnits.no_unit,
    )

    cut_infos_for_validation: List[CutInfo] = [
        fei_prob_cut,
        tag_mbc_cut,
        e_extra_cut,
        d_mass_cut,
        tag_delta_e_cut,
        cs_mva_cut,
        lepton_p_cut,
        b_vertex_fit_cut,
        q_square_cut,  # post BCS cut
    ]

    all_cut_infos: List[CutInfo] = [
        *cut_infos_for_validation,
        q_square_anti_cut,
        tag_mbc_anti_cut,
        cs_mva_anti_cut,
        cs_mva_loose_anti_cut,
    ]

    bcs_hist_vars: List[HistVariable] = [
        RDStarHistVars.tag_fei_prob,
        # RDStarHistVars.neextra_fine,  # Not using any signal side info! See also definition in get_bcs_info()!
    ]

    # Definition of Best Candidate Selection
    @staticmethod
    def get_bcs_info() -> Tuple[BCSVariableInfo, BCSVariableInfo]:
        primary_bcs_var = BCSVariableInfo(
            bcs_variable=Cols.fei_probability,
            absolute_value=False,
            max=False,
        )
        assert primary_bcs_var.bcs_variable in [
            hv.df_label for hv in MainSelection.bcs_hist_vars
        ], primary_bcs_var.bcs_variable

        secondary_bcs_var = BCSVariableInfo(  # Not using any signal side info! See below for alternative with sig info!
            bcs_variable=None,
            absolute_value=False,
            max=False,
        )
        # secondary_bcs_var = BCSVariableInfo(
        #     bcs_variable=Cols.event_ne_extra,
        #     absolute_value=False,
        #     max=False,
        # )
        # assert secondary_bcs_var.bcs_variable in [
        #     hv.df_label for hv in MainSelection.bcs_hist_vars
        # ], secondary_bcs_var.bcs_variable
        return primary_bcs_var, secondary_bcs_var

    # Selection Definitions
    @classmethod
    def get_selection(cls, cs_tag: str) -> Generator[SubSelectionInfo, None, None]:
        # Main Selection
        selection_tag: str = f"{cls.selection_name}_with_CS_{cs_tag}"
        cut_list: Tuple[CutInfo, ...] = (
            MainSelection.fei_prob_cut,
            MainSelection.tag_mbc_cut,
            MainSelection.e_extra_cut,
            MainSelection.d_mass_cut,
            # MainSelection.tag_delta_e_cut,
            MainSelection.cs_mva_cut,
            MainSelection.lepton_p_cut,
            MainSelection.b_vertex_fit_cut,
        )

        assert all(cut in cls.all_cut_infos for cut in cut_list), f"{selection_tag} contains ill-defined cuts!"

        post_bcs_cut_list: Tuple[CutInfo, ...] = (MainSelection.q_square_cut,)
        assert all(cut in cls.all_cut_infos for cut in post_bcs_cut_list), f"{selection_tag} contains ill-defined cuts!"

        yield SubSelectionInfo(
            selection_tag=selection_tag,
            cut_list=cut_list,
            post_bcs_cut_list=post_bcs_cut_list,
            check_cuts=(),
            check_bcs=False,
            run_evaluation=True,
        )

        base_sb_check_cut_list: Tuple[CutInfo, ...] = (
            MainSelection.e_extra_cut,
            MainSelection.tag_mbc_cut,
            MainSelection.cs_mva_cut,
            MainSelection.lepton_p_cut,
        )

        main_sb_check_cut_list: Tuple[CutInfo, ...] = base_sb_check_cut_list + (MainSelection.q_square_cut,)
        q2_sb_check_cut_list: Tuple[CutInfo, ...] = base_sb_check_cut_list

        assert all(c in cls.all_cut_infos for c in main_sb_check_cut_list), (
            c.cut_str for c in main_sb_check_cut_list if c not in cls.all_cut_infos
        )

        # Mbc Sideband Selection
        mbc_sb_tag: str = f"{selection_tag}_mbc_sideband"
        mbc_sb_cut_list = [MainSelection.tag_mbc_anti_cut if c == MainSelection.tag_mbc_cut else c for c in cut_list]
        assert all(cut in cls.all_cut_infos for cut in mbc_sb_cut_list), f"{mbc_sb_tag} contains ill-defined cuts!"
        mbc_sb_check_cuts = tuple([c for c in main_sb_check_cut_list if c != MainSelection.tag_mbc_cut])
        yield SubSelectionInfo(
            selection_tag=mbc_sb_tag,
            cut_list=tuple(mbc_sb_cut_list),
            post_bcs_cut_list=tuple(),
            check_cuts=mbc_sb_check_cuts,
            check_bcs=True,
            run_evaluation=False,
        )

        # q squared Sideband Selection
        q2_sb_tag: str = f"{selection_tag}_q2_sideband"
        q2_sb_post_bcs_cut_list = (MainSelection.q_square_anti_cut,)
        assert all(c in cls.all_cut_infos for c in q2_sb_post_bcs_cut_list), f"{q2_sb_tag} contains ill-defined cuts!"
        yield SubSelectionInfo(
            selection_tag=q2_sb_tag,
            cut_list=tuple(cut_list),
            post_bcs_cut_list=q2_sb_post_bcs_cut_list,
            check_cuts=q2_sb_check_cut_list,
            check_bcs=True,
            run_evaluation=False,
        )

        # Full q squared Selection
        full_q2_sb_tag: str = f"{selection_tag}_full_q2_sideband"
        yield SubSelectionInfo(
            selection_tag=full_q2_sb_tag,
            cut_list=tuple(cut_list),
            post_bcs_cut_list=tuple(),
            check_cuts=q2_sb_check_cut_list,
            check_bcs=True,
            run_evaluation=False,
        )

        # D** Evaluation Selection
        dss_eval_cut_list: Tuple[CutInfo, ...] = (
            MainSelection.fei_prob_cut,
            MainSelection.e_extra_cut,
            MainSelection.d_mass_cut,
            # MainSelection.tag_delta_e_cut,
            MainSelection.tag_mbc_cut,
            MainSelection.cs_mva_cut,
            MainSelection.b_vertex_fit_cut,
        )
        assert all(c in cut_list for c in dss_eval_cut_list), (c.cut_str for c in dss_eval_cut_list if c not in cut_list)

        dss_eval_sb_tag: str = f"{selection_tag}_dss_eval_sideband"
        dss_eval_sb_check_cuts = tuple([c for c in main_sb_check_cut_list if c in dss_eval_cut_list])
        yield SubSelectionInfo(
            selection_tag=dss_eval_sb_tag,
            cut_list=dss_eval_cut_list,
            post_bcs_cut_list=tuple(),
            check_cuts=dss_eval_sb_check_cuts,
            check_bcs=True,
            run_evaluation=False,
        )

        # CS MVA Sideband Selection
        cs_mva_sb_tag: str = f"{selection_tag}_csmva_sideband"

        cs_mva_sb_cut_list: Tuple[CutInfo, ...] = (
            MainSelection.cs_mva_loose_anti_cut,
            MainSelection.fei_prob_cut,
            MainSelection.e_extra_cut,
            MainSelection.d_mass_cut,
            MainSelection.lepton_p_cut,
            MainSelection.b_vertex_fit_cut,
        )

        assert all(c in cls.all_cut_infos for c in cs_mva_sb_cut_list), f"{cs_mva_sb_tag} contains ill-defined cuts!"

        cs_mva_sb_check_cut_list: Tuple[CutInfo, ...] = (
            MainSelection.e_extra_cut,
            MainSelection.lepton_p_cut,
        )
        assert all(
            c in cls.all_cut_infos for c in cs_mva_sb_check_cut_list
        ), f"{cs_mva_sb_tag} contains ill-defined cuts!"

        yield SubSelectionInfo(
            selection_tag=cs_mva_sb_tag,
            cut_list=cs_mva_sb_cut_list,
            post_bcs_cut_list=tuple(),
            check_cuts=cs_mva_sb_check_cut_list,
            check_bcs=False,
            run_evaluation=False,
        )


class MainWithMVASelection(MainSelection):
    _selection_name: str = "main_with_mva_selection"

    _mva_setup_name: Optional[str] = mva_selector_setups.SecondSelectionMVASetup.selection_mva_name
    selection_name: str = f"{_selection_name}_{_mva_setup_name}"

    _post_mva_cut: Optional[float] = 0.0

    cut_infos_for_validation: List[CutInfo] = copy.copy(MainSelection.cut_infos_for_validation)

    all_cut_infos: List[CutInfo] = copy.copy(MainSelection.all_cut_infos)

    bcs_hist_vars: List[HistVariable] = copy.copy(MainSelection.bcs_hist_vars)


# D mass Sideband Selection Setup
class DMassSidebandForMainWithMVASelection(MainSelection):
    _selection_name: str = "d_mass_sb_sel_for_main_with_mva_selection"

    _mva_setup_name: Optional[str] = mva_selector_setups.SecondSelectionMVASetup.selection_mva_name
    selection_name: str = f"{_selection_name}_{_mva_setup_name}"

    _post_mva_cut: Optional[float] = 0.0

    cut_infos_for_validation: List[CutInfo] = copy.copy(MainSelection.cut_infos_for_validation)

    all_cut_infos: List[CutInfo] = copy.copy(MainSelection.all_cut_infos)

    bcs_hist_vars: List[HistVariable] = copy.copy(MainSelection.bcs_hist_vars)

    # Selection Definitions
    @classmethod
    def get_selection(cls, cs_tag: str) -> Generator[SubSelectionInfo, None, None]:
        selection_tag: str = f"{cls.selection_name}_with_CS_{cs_tag}"
        cut_list: Tuple[CutInfo, ...] = (
            MainSelection.fei_prob_cut,
            MainSelection.tag_mbc_cut,
            MainSelection.e_extra_cut,
            MainSelection.d_mass_cut,
            # MainSelection.tag_delta_e_cut,
            MainSelection.cs_mva_cut,
            MainSelection.lepton_p_cut,
            MainSelection.b_vertex_fit_cut,
        )
        assert all(c in cls.all_cut_infos for c in cut_list), (c.cut_str for c in cut_list if c not in cls.all_cut_infos)

        check_cut_list: Tuple[CutInfo, ...] = (
            MainSelection.e_extra_cut,
            MainSelection.tag_mbc_cut,
            MainSelection.cs_mva_cut,
            MainSelection.lepton_p_cut,
        )

        yield SubSelectionInfo(
            selection_tag=selection_tag,
            cut_list=cut_list,
            post_bcs_cut_list=tuple(),  # <- Has no q2 cut!
            check_cuts=check_cut_list,
            check_bcs=True,
            run_evaluation=False,
        )


class SelectionCompareInfo(NamedTuple):
    name: str
    selection_names: Tuple[str, ...]


def check_setups_and_provide_correction_mapping(
    selections_dict: Dict[str, Type[SelectionSetupBase]],
    mapping_dict: Dict[str, str],
    used_selection_name: str,
    used_selections_tuple: Tuple[str, ...],
    to_be_evaluated: Tuple[str, ...],
    compare_infos: Tuple[SelectionCompareInfo, ...],
) -> Dict[str, str]:
    assert used_selection_name in selections_dict.keys(), (used_selection_name, selections_dict.keys())
    assert all(n in selections_dict.keys() for n in used_selections_tuple), (
        n for n in used_selections_tuple if n not in selections_dict.keys()
    )
    assert all(n in selections_dict for n in to_be_evaluated), (n for n in to_be_evaluated if n not in selections_dict)

    assert len(set([c.name for c in compare_infos])) == len(compare_infos), [c.name for c in compare_infos]
    assert all(k in selections_dict for kk in compare_infos for k in kk.selection_names), [
        (k.name, k.selection_names) for k in compare_infos if not all(n in selections_dict for n in k.selection_names)
    ]

    assert all(k in selections_dict for k in mapping_dict.keys()), [
        k for k in mapping_dict.keys() if k not in selections_dict
    ]
    assert all(k in selections_dict for k in mapping_dict.values()), [
        k for k in mapping_dict.values() if k not in selections_dict
    ]

    full_resolution_correction_mapping: Dict[str, str] = {
        k: k if k not in mapping_dict else mapping_dict[k] for k in selections_dict.keys()
    }

    assert set(selections_dict.keys()) == set(full_resolution_correction_mapping.keys()), (
        selections_dict.keys(),
        full_resolution_correction_mapping.keys(),
    )

    return full_resolution_correction_mapping


# This interface class should be used to access all defined selections...
class SelectionSetups:
    available_selections: Dict[str, Type[SelectionSetupBase]] = {
        MainSelection.selection_name: MainSelection,
        MainWithMVASelection.selection_name: MainWithMVASelection,
        DMassSidebandForMainWithMVASelection.selection_name: DMassSidebandForMainWithMVASelection,
        FirstSelection.selection_name: FirstSelection,
        MVASelection.selection_name: MVASelection,
        LooseMVASelection.selection_name: LooseMVASelection,
        OptimizedMVASelection.selection_name: OptimizedMVASelection,
        FullQ2Selection.selection_name: FullQ2Selection,
        DSSSelection.selection_name: DSSSelection,
        DSSSelectionLooseMVA.selection_name: DSSSelectionLooseMVA,
        DSSSelectionFullPL.selection_name: DSSSelectionFullPL,
        DSSSelectionFullPLLooseMVA.selection_name: DSSSelectionFullPLLooseMVA,
    }

    __alternative_res_corr_mapping: Dict[str, str] = {
        DMassSidebandForMainWithMVASelection.selection_name: MainWithMVASelection.selection_name,
        DSSSelection.selection_name: LooseMVASelection.selection_name,
        DSSSelectionLooseMVA.selection_name: LooseMVASelection.selection_name,
        DSSSelectionFullPL.selection_name: LooseMVASelection.selection_name,
        DSSSelectionFullPLLooseMVA.selection_name: LooseMVASelection.selection_name,
    }

    __default_res_corr_source_sub_sample_key: str = "q2_sideband"
    __alternative_res_corr_source_sub_sample_key: Dict[str, str] = {
        FullQ2Selection.selection_name: "",
    }

    __evaluate_now: ClassVar[Tuple[str, ...]] = (
        MainSelection.selection_name,
        MainWithMVASelection.selection_name,
        DMassSidebandForMainWithMVASelection.selection_name,
        # FirstSelection.selection_name,
        # MVASelection.selection_name,
        # LooseMVASelection.selection_name,
        # OptimizedMVASelection.selection_name,
        # # FullQ2Selection.selection_name,
        # # DSSSelection.selection_name,
        # # DSSSelectionLooseMVA.selection_name,
        # # DSSSelectionFullPL.selection_name,
        # # DSSSelectionFullPLLooseMVA.selection_name,
    )

    __compare_infos: ClassVar[Tuple[SelectionCompareInfo, ...]] = (
        SelectionCompareInfo(
            name="RDStar",
            selection_names=(
                MainSelection.selection_name,
                FirstSelection.selection_name,
                # MVASelection.selection_name,
                # LooseMVASelection.selection_name,
                # OptimizedMVASelection.selection_name,
            ),
        ),
        SelectionCompareInfo(
            name="Dss",
            selection_names=(
                DSSSelectionFullPL.selection_name,
                DSSSelectionFullPLLooseMVA.selection_name,
            ),
        ),
    )

    my_selections: Tuple[Type[SelectionSetupBase], ...] = (
        MainSelection,
        MainWithMVASelection,
        DMassSidebandForMainWithMVASelection,
        FirstSelection,
        MVASelection,
        LooseMVASelection,
        OptimizedMVASelection,
        DSSSelection,
        DSSSelectionLooseMVA,
        DSSSelectionFullPL,
        DSSSelectionFullPLLooseMVA,
    )

    rdstar_selections: Tuple[str, ...] = (
        MainSelection.selection_name,
        MainWithMVASelection.selection_name,
        DMassSidebandForMainWithMVASelection.selection_name,
        FirstSelection.selection_name,
        MVASelection.selection_name,
        LooseMVASelection.selection_name,
        OptimizedMVASelection.selection_name,
    )

    used_selection_name: str = MainWithMVASelection.selection_name
    used_selections_tuple: Tuple[str, ...] = (
        MainSelection.selection_name,
        MainWithMVASelection.selection_name,
        DMassSidebandForMainWithMVASelection.selection_name,
        LooseMVASelection.selection_name,
        OptimizedMVASelection.selection_name,
    )

    resolution_correction_mapping: Dict[str, str] = check_setups_and_provide_correction_mapping(
        selections_dict=available_selections,
        mapping_dict=__alternative_res_corr_mapping,
        used_selection_name=used_selection_name,
        used_selections_tuple=used_selections_tuple,
        to_be_evaluated=__evaluate_now,
        compare_infos=__compare_infos,
    )

    @staticmethod
    def _get_selection_class(selection_name: str):
        if selection_name in SelectionSetups.available_selections:
            return SelectionSetups.available_selections[selection_name]
        else:
            raise ValueError(f"Selection with name {selection_name} is not available!")

    @staticmethod
    def selection_is_evaluated(selection_name: str) -> bool:
        return selection_name in SelectionSetups.__evaluate_now

    @staticmethod
    def compare_selections_set_infos() -> Tuple[SelectionCompareInfo, ...]:
        return SelectionSetups.__compare_infos

    @staticmethod
    def get_available_selection_names() -> Tuple[str, ...]:
        return tuple([sel_name for sel_name in SelectionSetups.available_selections.keys()])

    @staticmethod
    def get_selection_infos(selection_name: str, cs_tag: str) -> Dict[str, SelectionSpecifics]:
        return SelectionSetups._get_selection_class(selection_name=selection_name).get_selection_specifics(cs_tag=cs_tag)

    @staticmethod
    def get_selection_cuts(
        selection_name: str,
        selection_tag: str,
        cs_tag: str,
    ) -> Tuple[Tuple[str, ...], Tuple[str, ...], Tuple[str, ...]]:
        specs = SelectionSetups.get_selection_infos(selection_name=selection_name, cs_tag=cs_tag)[selection_tag]

        cuts = tuple([c.cut_str for c in specs.cuts])
        check_cuts = tuple([c.cut_str for c in specs.check_cuts])
        base_cuts = tuple([c for c in cuts if c not in check_cuts])

        assert len(base_cuts) + len(check_cuts) == len(cuts), f"{len(base_cuts)} + {len(check_cuts)} == {len(cuts)}"
        return cuts, base_cuts, check_cuts

    @staticmethod
    def get_selection_bcs(
        selection_name: str,
        selection_tag: str,
        cs_tag: str,
    ) -> Tuple[BCSVariableInfo, BCSVariableInfo, bool]:
        specs = SelectionSetups.get_selection_infos(selection_name=selection_name, cs_tag=cs_tag)[selection_tag]
        return specs.primary_bcs_var, specs.secondary_bcs_var, specs.check_bcs

    @staticmethod
    def get_selection_mva_info(selection_name: str) -> Optional[MVAInfo]:
        return SelectionSetups._get_selection_class(selection_name=selection_name).get_mva_info()

    @staticmethod
    def get_selection_mva_output_col(selection_name: str) -> Optional[str]:
        mva_info: Optional[MVAInfo] = SelectionSetups.get_selection_mva_info(selection_name=selection_name)
        if mva_info is not None:
            return mva_info.mva_setup.get_selection_mva_output_column_name()
        else:
            return None

    @staticmethod
    def get_selection_mva_output_col_with_check(selection_name: str) -> str:
        mva_info: Optional[MVAInfo] = SelectionSetups.get_selection_mva_info(selection_name=selection_name)
        if mva_info is not None:
            return mva_info.mva_setup.get_selection_mva_output_column_name()
        else:
            raise TypeError(
                f"You are trying to access the mva output column name for a "
                f"selection {selection_name} which does not have an MVA!"
            )

    @staticmethod
    def get_res_corr_source_sub_sample_key(selection_name: str) -> str:
        assert selection_name in SelectionSetups.resolution_correction_mapping.keys(), selection_name
        assert SelectionSetups.resolution_correction_mapping[selection_name] == selection_name, (
            selection_name,
            SelectionSetups.resolution_correction_mapping[selection_name],
        )

        return SelectionSetups.__alternative_res_corr_source_sub_sample_key.get(
            selection_name,
            SelectionSetups.__default_res_corr_source_sub_sample_key,
        )

    @staticmethod
    def get_selection_tags(selection_name: str, cs_tag: str) -> Tuple[str, ...]:
        selection_class = SelectionSetups._get_selection_class(selection_name=selection_name)
        return selection_class.get_selection_tags(cs_tag=cs_tag)

    @staticmethod
    def get_selection_tags_with_eval_info(selection_name: str, cs_tag: str) -> Tuple[str, ...]:
        selection_class = SelectionSetups._get_selection_class(selection_name=selection_name)
        return selection_class.get_selection_tags_with_eval_info(cs_tag=cs_tag)

    @staticmethod
    def get_main_selection_tag(selection_name: str, cs_tag: str) -> str:
        selection_class = SelectionSetups._get_selection_class(selection_name=selection_name)
        return selection_class.get_main_selection_tag(cs_tag=cs_tag)

    @staticmethod
    def get_selection_cut_label_dict(selection_name: str) -> Dict[str, str]:
        selection_class = SelectionSetups._get_selection_class(selection_name=selection_name)
        return selection_class.get_cut_label_dict()


if __name__ == "__main__":
    for selection_setup in SelectionSetups.available_selections.values():
        selection_setup.print_selection_info(cs_tag="test")
