import os
import b2luigi as luigi

from typing import Dict, Generator

from rdstar.utilities import PathType
from rdstar.offline_analysis.tasks import utilities

from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses
from rdstar.latex_labels.latex_output_utilities import create_latex_command_file, get_latex_cmd_to_attributes_mapping

from rdstar.branching_fractions.b_meson_bf_calculation import write_bf_values_to_latex, creat_latex_variable_table


__all__ = [
    "LatexTasks",
]


class LatexTasks(luigi.WrapperTask):
    offline_config_path = luigi.Parameter(positional=False)  # type: PathType

    latex_dir_path = "LatexOutputs"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        offline_config = utilities.load_yaml_config(yaml_config_path=self.offline_config_path)
        self.working_dir = offline_config["working_dir"]

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield LatexProcessLabelsTask(working_dir=self.working_dir)
        yield LatexBFValueTask(working_dir=self.working_dir)
        yield LatexBFValueTableTask(working_dir=self.working_dir)


class LatexProcessLabelsTask(utilities.BaseOfflineTask):
    working_dir = luigi.Parameter(positional=False)  # type: PathType

    output_key = "LatexProcessLabels"  # type: str
    output_dir_name = "LatexProcessLabels"  # type: str
    output_file_name = "rdstar_latex_process_labels.tex"

    @property
    def priority(self) -> int:
        return 10

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.target_dir_path = os.path.join(
            self.working_dir,
            LatexTasks.latex_dir_path,
            self.output_dir_name,
        )  # type: PathType

        self.output_path = os.path.join(self.target_dir_path, self.output_file_name)  # type: PathType

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.working_dir, exist_ok=True)

        latex_process_label_attrs = LatexDecayProcesses.get_attributes_mapping()
        create_latex_command_file(
            output_path=self.output_path,
            command_mapping_dict=get_latex_cmd_to_attributes_mapping(attributes_mapping=latex_process_label_attrs),
            command_prefix=None,
            overwrite=True,
        )

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = {
            self.output_key: luigi.LocalTarget(self.output_path),
        }  # type: Dict[str, luigi.LocalTarget]
        return output_dict


class LatexBFValueTask(utilities.BaseOfflineTask):
    working_dir = luigi.Parameter(positional=False)  # type: PathType

    output_key = "LatexBFValues"  # type: str
    output_dir_name = "LatexBFValues"  # type: str
    output_file_name = "rdstar_latex_bf_values.tex"

    @property
    def priority(self) -> int:
        return 10

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.target_dir_path = os.path.join(
            self.working_dir,
            LatexTasks.latex_dir_path,
            self.output_dir_name,
        )  # type: PathType

        self.output_path = os.path.join(self.target_dir_path, self.output_file_name)  # type: PathType

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.working_dir, exist_ok=True)

        write_bf_values_to_latex(
            output_path=self.output_path,
            in_percent=True,
            overwrite=True,
        )

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = {
            self.output_key: luigi.LocalTarget(self.output_path),
        }  # type: Dict[str, luigi.LocalTarget]
        return output_dict


class LatexBFValueTableTask(utilities.BaseOfflineTask):
    working_dir = luigi.Parameter(positional=False)  # type: PathType

    output_key = "LatexBFTable"  # type: str
    output_dir_name = "LatexBFTable"  # type: str
    output_file_name = "rdstar_latex_bf_values_table.tex"

    @property
    def priority(self) -> int:
        return 10

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.target_dir_path = os.path.join(
            self.working_dir,
            LatexTasks.latex_dir_path,
            self.output_dir_name,
        )  # type: PathType

        self.output_path = os.path.join(self.target_dir_path, self.output_file_name)  # type: PathType

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.working_dir, exist_ok=True)

        creat_latex_variable_table(
            output_path=self.output_path,
            overwrite=True,
        )

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = {
            self.output_key: luigi.LocalTarget(self.output_path),
        }  # type: Dict[str, luigi.LocalTarget]
        return output_dict
