"""
Evaluation of efficiency dependencies on MC and Data

Felix Metzner 2021
"""

import os
import numpy as np
import pandas as pd

from collections import Counter
from dataclasses import dataclass
from matplotlib import pyplot as plt
from uncertainties import unumpy as unp
from matplotlib.ticker import MaxNLocator, FixedLocator
from typing import Optional, Tuple, List, Sequence, Dict, Callable, ClassVar

from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.plot_utilities import export, AxesType, FigureType
from templatefitter.plotter.plot_style import set_matplotlibrc_params, xlabel_pos, ylabel_pos, KITColors

from analysistools.statistics.limits import binomial_limit_approximation

from rdstar.utilities import PathType

from rdstar.offline_analysis.reco_modes.reco_mode_collections import BSigRecoModes
from rdstar.offline_analysis.reco_modes.reco_mode_utilities import RecoModeCollection, RecoModeGroupNameToTitleDictType

from rdstar.offline_analysis.histogram_variables import RDStarHistVars
from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.data_handling.selected_data_handling import DataSampleLoader
from rdstar.offline_analysis.data_handling.sample_setup_definitions import SampleSetupHandler
from rdstar.offline_analysis.corrections.efficiency_descrepancy_corrector import RecoEffCorrInfoContainer


__all__ = [
    "EffDepEvalInfo",
    "EfficiencyDependencyEvaluator",
]


set_matplotlibrc_params()


@dataclass(frozen=True)
class EffValCutInfos:
    hist_var: HistVariable
    cut_values: Tuple[float, ...]

    def __post_init__(self) -> None:
        assert all(cv >= self.hist_var.scope[0] for cv in self.cut_values), (self.hist_var.scope, self.cut_values)
        assert all(cv <= self.hist_var.scope[1] for cv in self.cut_values), (self.hist_var.scope, self.cut_values)

    @property
    def cut_val_x_axis_label(self) -> str:
        unit_str: str = f" in {self.hist_var.unit}" if self.hist_var.unit else self.hist_var.unit
        return f"Lower Bound on {self.hist_var.variable_name}{unit_str}"

    def get_cut_condition(
        self,
        df: pd.DataFrame,
        cut_value: float,
    ) -> pd.Series:
        assert self.hist_var.df_label in df.columns, self.hist_var.df_label
        assert cut_value in self.cut_values, (cut_value, self.cut_values)

        return df[self.hist_var.df_label] >= cut_value


@dataclass(frozen=True)
class EffDepEvalInfo:
    name: str
    evaluation_hist_vars: Sequence[HistVariable]
    cut_information: EffValCutInfos
    reco_mode_info_collection: RecoModeCollection
    weight_column_name: str
    sample_setup_keys_to_sideband_keys_map: Dict[Tuple[str, ...], Tuple[str, ...]]

    @property
    def is_single_working_point_evaluation(self) -> bool:
        return len(self.cut_information.cut_values) == 1


__sel_mva_cut_values: Tuple[float, ...] = (
    0.5,
    0.525,
    0.55,
    0.575,
    0.6,
    0.625,
    0.65,
    0.675,
    0.7,
    0.725,
    0.75,
    0.775,
    0.808,
    0.825,
    0.85,
    0.88,
    0.9,
    0.91,
    0.92,
    0.93,
    0.94,
    0.95,
    0.96,
    0.97,
)

__cs_mva_cut_values: Tuple[float, ...] = tuple(np.around(np.arange(0.5, 1.0 - 0.03 + 0.01, 0.01), decimals=2))


regular_sample_setup_dict: Dict[Tuple[str, ...], Tuple[str, ...]] = {
    SampleSetupHandler.on_resonance_sample_setup_names: (
        DataSampleLoader.base_identifier,
        "q2_sideband",
        DataSampleLoader.get_sideband_key_for(sideband_keys=(DataSampleLoader.base_identifier, "q2_sideband")),
    ),
    (SampleSetupHandler.off_resonance_sample_name,): (
        "mbc_sideband_without_cut_1",
        DataSampleLoader.get_sideband_key_for(sideband_keys=("mbc_sideband", "q2_sideband_without_cut_1")),
    ),
}


eff_reco_dependency_correction_setup_name: str = "evaluation_at_working_point"

eff_eval_setups: Tuple[EffDepEvalInfo, ...] = (
    EffDepEvalInfo(
        name=eff_reco_dependency_correction_setup_name,
        evaluation_hist_vars=(RDStarHistVars.missing_mass_res_corr_for_fit, RDStarHistVars.pl_star),
        cut_information=EffValCutInfos(
            hist_var=RDStarHistVars.mva_output_loose_for_sbv_var,
            cut_values=(RDStarHistVars.mva_output_loose_for_sbv_var.scope[0],),
        ),
        reco_mode_info_collection=BSigRecoModes.main_modes,
        weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
        sample_setup_keys_to_sideband_keys_map=regular_sample_setup_dict,
    ),
    EffDepEvalInfo(
        name="selection_mva_cut_evaluation",
        evaluation_hist_vars=(RDStarHistVars.missing_mass_res_corr_for_fit, RDStarHistVars.pl_star),
        cut_information=EffValCutInfos(
            hist_var=RDStarHistVars.mva_output_loose_for_sbv_var,
            cut_values=__sel_mva_cut_values,
        ),
        reco_mode_info_collection=BSigRecoModes.main_modes,
        weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
        sample_setup_keys_to_sideband_keys_map=regular_sample_setup_dict,
    ),
    EffDepEvalInfo(
        name="cs_mva_cut_evaluation",
        evaluation_hist_vars=(RDStarHistVars.missing_mass_res_corr_for_fit, RDStarHistVars.pl_star),
        cut_information=EffValCutInfos(
            hist_var=RDStarHistVars.cs_mva_output,
            cut_values=__cs_mva_cut_values,
        ),
        reco_mode_info_collection=BSigRecoModes.main_modes,
        weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
        sample_setup_keys_to_sideband_keys_map=regular_sample_setup_dict,
    ),
)

__names_count = Counter([v.name for v in eff_eval_setups])
assert len(set(v.name for v in eff_eval_setups)) == len(eff_eval_setups), [k for k, v in __names_count.items() if v > 1]


@dataclass(frozen=True)
class OutputFileInfo:
    key: str
    file_path: PathType

    @property
    def target_dir(self) -> PathType:
        return os.path.dirname(self.file_path)

    @property
    def tmp_file_path(self) -> PathType:
        return os.path.join(self.target_dir, f"tmp_{os.path.basename(self.file_path)}")

    @property
    def png_file_path(self) -> PathType:
        return os.path.join(self.target_dir, f"{self.pure_file_name}.png")

    @property
    def tmp_png_file_path(self) -> PathType:
        return os.path.join(self.target_dir, f"tmp_{os.path.basename(self.png_file_path)}")

    @property
    def pure_file_name(self) -> str:
        return os.path.splitext(os.path.basename(self.file_path))[0]

    @property
    def pure_tmp_file_name(self) -> str:
        return f"tmp_{os.path.splitext(os.path.basename(self.file_path))[0]}"


@dataclass(frozen=True)
class CutEvalEffInfo:
    cut_value: float
    reco_mode_info_collection: RecoModeCollection
    mc_sow: Dict[str, float]
    data_count: Dict[str, float]

    def __post_init__(self) -> None:
        all_reco_mode_keys: Tuple[str, ...] = tuple(
            k for k, _ in self.reco_mode_info_collection.reco_mode_info_mapping_as_tuple
        )
        assert set(self.mc_sow.keys()) == set(all_reco_mode_keys), (self.mc_sow.keys(), all_reco_mode_keys)
        assert set(self.mc_sow.keys()) == set(self.data_count.keys()), (self.mc_sow.keys(), self.data_count.keys())

    @staticmethod
    def create(
        cut_value: float,
        reco_mode_info_collection: RecoModeCollection,
        mc_df: pd.DataFrame,
        data_df: pd.DataFrame,
        reco_mode_col_name: str,
        weight_col_name: str,
    ) -> "CutEvalEffInfo":

        mc_sow_dict: Dict[str, float] = dict()
        data_count_dict: Dict[str, float] = dict()

        for reco_mode_key, reco_mode_ids in reco_mode_info_collection.reco_mode_info_mapping_as_tuple:
            mc_reco_mode_cond: pd.Series = mc_df[reco_mode_col_name].isin(reco_mode_ids)
            data_reco_mode_cond: pd.Series = data_df[reco_mode_col_name].isin(reco_mode_ids)

            mc_sum_of_weights: float = mc_df[mc_reco_mode_cond][weight_col_name].sum()
            data_event_count: int = len(data_df[data_reco_mode_cond].index)

            assert reco_mode_key not in mc_sow_dict, reco_mode_key
            assert reco_mode_key not in data_count_dict, reco_mode_key

            mc_sow_dict.update({reco_mode_key: mc_sum_of_weights})
            data_count_dict.update({reco_mode_key: data_event_count})

        return CutEvalEffInfo(
            cut_value=cut_value,
            reco_mode_info_collection=reco_mode_info_collection,
            mc_sow=mc_sow_dict,
            data_count=data_count_dict,
        )


@dataclass(frozen=True)
class EfficiencyPlotValues:
    reco_mode_key: str
    cut_values: np.ndarray
    mc_eff: np.ndarray
    mc_err: np.ndarray
    data_eff: np.ndarray
    data_err: np.ndarray


class EfficiencyDependencyEvaluator:
    base_dir_name: ClassVar[str] = "EfficiencyDependencyEvaluation"
    base_file_name: ClassVar[str] = "efficiency_evaluation"

    correction_providing_setup_name: ClassVar[str] = eff_reco_dependency_correction_setup_name
    correction_method_name: ClassVar[str] = "reco_mode_dependent_eff_correction"

    reco_mode_id_col: ClassVar[str] = Cols.decay_mode_column_name

    relative_eff_y_axis_label: ClassVar[str] = "relative Efficiency in %"
    data_vs_mc_eff_y_axis_label: ClassVar[str] = "Data vs. MC Event Count"

    reconstruction_mode_x_axis_label: ClassVar[str] = "Reconstruction Mode"

    default_fig_size: ClassVar[Tuple[float, float]] = (6.5, 5.0)

    setup_dict: ClassVar[Dict[str, EffDepEvalInfo]] = {v.name: v for v in eff_eval_setups}

    def __init__(
        self,
        evaluation_hist_vars: Sequence[HistVariable],
        cut_information: EffValCutInfos,
        reco_mode_info_collection: RecoModeCollection,
        weight_column_name: str,
        result_dir_path: PathType,
    ) -> None:

        self.evaluation_hist_vars: Tuple[HistVariable, ...] = tuple(evaluation_hist_vars)
        self._cut_information: EffValCutInfos = cut_information
        self.reco_mode_info_collection: RecoModeCollection = reco_mode_info_collection

        self.weight_col_name: str = weight_column_name

        self.base_dir_path: PathType = os.path.join(result_dir_path, self.cut_tag)

    @property
    def cut_infos(self) -> EffValCutInfos:
        return self._cut_information

    @property
    def cut_col_name(self) -> str:
        return self.cut_infos.hist_var.df_label

    @property
    def is_single_working_point_evaluation(self) -> bool:
        return len(self.cut_infos.cut_values) == 1

    @property
    def cut_tag(self) -> str:
        if self.is_single_working_point_evaluation:
            return f"{self.cut_col_name}_working_point"
        else:
            return self.cut_col_name

    def perform_efficiency_evaluation(
        self,
        mc_sample: Optional[pd.DataFrame],
        data_sample: Optional[pd.DataFrame],
        dry_run: bool = False,
    ) -> Dict[str, PathType]:
        output_dict: Dict[str, PathType] = dict()

        for eval_hist_var in self.evaluation_hist_vars:

            if dry_run:
                self.create_efficiency_plots(
                    evaluation_hist_var=eval_hist_var,
                    evaluation_infos=list(),
                    output_dict=output_dict,
                    dry_run=dry_run,
                )

            else:
                os.makedirs(self.base_dir_path, exist_ok=True)

                assert isinstance(mc_sample, pd.DataFrame), type(mc_sample).__name__
                assert isinstance(data_sample, pd.DataFrame), type(data_sample).__name__

                assert self.weight_col_name in mc_sample, self.weight_col_name
                assert self.reco_mode_id_col in mc_sample, self.reco_mode_id_col
                assert self.reco_mode_id_col in data_sample, self.reco_mode_id_col

                data_range_cond: pd.Series = self.get_eval_range_condition(df=data_sample, hist_var=eval_hist_var)
                mc_range_cond: pd.Series = self.get_eval_range_condition(df=mc_sample, hist_var=eval_hist_var)

                eff_eval_infos: List[CutEvalEffInfo] = list()

                for cut_val in self.cut_infos.cut_values:
                    mc_cut_cond: pd.Series = self.cut_infos.get_cut_condition(df=mc_sample, cut_value=cut_val)
                    data_cut_cond: pd.Series = self.cut_infos.get_cut_condition(df=data_sample, cut_value=cut_val)

                    eff_eval_info: CutEvalEffInfo = CutEvalEffInfo.create(
                        cut_value=cut_val,
                        reco_mode_info_collection=self.reco_mode_info_collection,
                        mc_df=mc_sample.loc[mc_range_cond & mc_cut_cond],
                        data_df=data_sample.loc[data_range_cond & data_cut_cond],
                        reco_mode_col_name=self.reco_mode_id_col,
                        weight_col_name=self.weight_col_name,
                    )
                    eff_eval_infos.append(eff_eval_info)

                self.create_efficiency_plots(
                    evaluation_hist_var=eval_hist_var,
                    evaluation_infos=eff_eval_infos,
                    output_dict=output_dict,
                    dry_run=dry_run,
                )

        self.prepare_reco_mode_dependent_eff_correction(
            mc_sample=mc_sample,
            data_sample=data_sample,
            output_dict=output_dict,
            dry_run=dry_run,
        )

        return output_dict

    @staticmethod
    def get_eff_reco_correction_output_key(input_dict_keys: Sequence[str]) -> str:
        relevant_keys: List[str] = [
            k
            for k in input_dict_keys
            if k.startswith(EfficiencyDependencyEvaluator.base_file_name)
            and k.endswith(EfficiencyDependencyEvaluator.correction_method_name)
        ]
        assert len(relevant_keys) == 1, relevant_keys
        return relevant_keys[0]

    def prepare_reco_mode_dependent_eff_correction(
        self,
        mc_sample: Optional[pd.DataFrame],
        data_sample: Optional[pd.DataFrame],
        output_dict: Dict[str, PathType],
        dry_run: bool,
    ) -> None:
        if not self.is_single_working_point_evaluation:
            return

        def get_hv_str(hist_var: HistVariable) -> str:
            scope_lower_str: str = str(hist_var.scope[0]).replace(".", "p")
            scope_upper_str: str = str(hist_var.scope[1]).replace(".", "p")
            return f"{hist_var.df_label}_{scope_lower_str}_{scope_upper_str}"

        var_range_list = "_".join(get_hv_str(hist_var=hv) for hv in self.evaluation_hist_vars)

        base_name: str = f"{self.base_file_name}_{var_range_list}_{self.correction_method_name}"
        file_name: str = f"{base_name}.json"
        file_path: PathType = os.path.join(self.base_dir_path, file_name)

        assert base_name not in output_dict, base_name
        assert file_path not in output_dict.values(), file_path
        output_dict.update({base_name: file_path})

        if not dry_run:
            assert isinstance(mc_sample, pd.DataFrame), type(mc_sample).__name__
            assert isinstance(data_sample, pd.DataFrame), type(data_sample).__name__

            data_range_cond: pd.Series = pd.Series(True, index=data_sample.index)
            mc_range_cond: pd.Series = pd.Series(True, index=mc_sample.index)
            for eval_hist_var in self.evaluation_hist_vars:
                _data_cond: pd.Series = self.get_eval_range_condition(df=data_sample, hist_var=eval_hist_var)
                _mc_cond: pd.Series = self.get_eval_range_condition(df=mc_sample, hist_var=eval_hist_var)

                assert len(_data_cond) == len(data_range_cond), (len(_data_cond), len(data_range_cond))
                assert len(_mc_cond) == len(mc_range_cond), (len(_mc_cond), len(mc_range_cond))

                data_range_cond &= _data_cond
                mc_range_cond &= _mc_cond

            eff_eval_info: CutEvalEffInfo = CutEvalEffInfo.create(
                cut_value=self.cut_infos.cut_values[0],  # No specific cut value, as cut is not relevant here!
                reco_mode_info_collection=self.reco_mode_info_collection,
                mc_df=mc_sample.loc[mc_range_cond],
                data_df=data_sample.loc[data_range_cond],
                reco_mode_col_name=self.reco_mode_id_col,
                weight_col_name=self.weight_col_name,
            )

            reco_mode_keys: Tuple[str, ...] = tuple(i.name for i in self.reco_mode_info_collection.reco_mode_infos)
            reco_mode_ids: Tuple[Tuple[int, ...], ...] = tuple(
                self.reco_mode_info_collection[k].ids for k in reco_mode_keys
            )

            _correction_factors: List[float] = list()
            _correction_factor_errors: List[float] = list()
            for reco_mode_key in reco_mode_keys:
                eff_info = self.get_data_vs_mc_eff_info_for(
                    reco_mode_key=reco_mode_key,
                    efficiency_evaluation_infos=(eff_eval_info,),
                )
                assert len(eff_info.mc_eff) == len(eff_info.mc_err) == 1, (len(eff_info.mc_eff), len(eff_info.mc_err))
                _correction_factors.append(eff_info.mc_eff[0])
                _correction_factor_errors.append(eff_info.mc_err[0])

            correction_factors: Tuple[float, ...] = tuple(_correction_factors)
            correction_factor_errors: Tuple[float, ...] = tuple(_correction_factor_errors)

            correction_info_container: RecoEffCorrInfoContainer = RecoEffCorrInfoContainer(
                reco_mode_id_column_name=self.reco_mode_id_col,
                reco_mode_keys=reco_mode_keys,
                reco_mode_ids=reco_mode_ids,
                base_weight_column_name=self.weight_col_name,
                eval_hist_var_column_names=tuple(hv.df_label for hv in self.evaluation_hist_vars),
                eval_hist_var_scopes=tuple(hv.scope for hv in self.evaluation_hist_vars),
                correction_factors=correction_factors,
                correction_factor_errors=correction_factor_errors,
            )

            correction_info_container.write_to_file(target_file_path=file_path)

    def create_efficiency_plots(
        self,
        evaluation_hist_var: HistVariable,
        evaluation_infos: List[CutEvalEffInfo],
        output_dict: Dict[str, PathType],
        dry_run: bool,
    ) -> None:
        plot_method_mapping: Dict[str, Tuple[Callable, bool]] = {
            "efficiency_relative_to_no_cut": (self.plot_efficiency_relative_to_no_cut, True),
            "efficiency_data_vs_mc": (self.plot_efficiency_data_vs_mc, False),
        }

        for plot_name, (plot_method, requires_cuts) in plot_method_mapping.items():
            if requires_cuts and self.is_single_working_point_evaluation:
                continue

            output_file_info: OutputFileInfo = self.append_output_info(
                eval_hist_var=evaluation_hist_var,
                method_name=plot_name,
                extension="pdf",
                output_dict=output_dict,
            )

            if not dry_run:
                assert len(evaluation_infos) > 0
                plot_method(
                    efficiency_evaluation_infos=tuple(evaluation_infos),
                    output_file_info=output_file_info,
                )
                os.rename(src=output_file_info.tmp_file_path, dst=output_file_info.file_path)
                os.rename(src=output_file_info.tmp_png_file_path, dst=output_file_info.png_file_path)

    def plot_efficiency_relative_to_no_cut(
        self,
        efficiency_evaluation_infos: Tuple[CutEvalEffInfo, ...],
        output_file_info: OutputFileInfo,
    ) -> None:
        set_matplotlibrc_params()

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=self.default_fig_size, dpi=200)  # type: FigureType, AxesType

        reco_mode_title_mapping: RecoModeGroupNameToTitleDictType = (
            self.reco_mode_info_collection.get_reco_mode_name_to_latex_label_mapping(short=False)
        )

        for reco_index, reco_mode_info in enumerate(self.reco_mode_info_collection.values()):

            reco_mode_label: str = reco_mode_title_mapping[reco_mode_info.name]
            eff_infos: EfficiencyPlotValues = self.get_relative_eff_info_for(
                reco_mode_key=reco_mode_info.name,
                efficiency_evaluation_infos=efficiency_evaluation_infos,
            )

            ax.errorbar(
                x=eff_infos.cut_values,
                y=100 * eff_infos.data_eff,
                xerr=None,
                yerr=100 * eff_infos.data_err,
                ls="",
                marker="x",
                alpha=1.0,
                color=KITColors.default_colors[reco_index],
                label=f"Data: {reco_mode_label}",
            )

            mc_markers, mc_caps, mc_bars = ax.errorbar(
                x=eff_infos.cut_values,
                y=100 * eff_infos.mc_eff,
                xerr=None,
                yerr=100 * eff_infos.mc_err,
                ls="",
                marker="_",
                alpha=1.0,
                elinewidth=3,
                color=KITColors.default_colors[reco_index],
                label=f"MC:  {reco_mode_label}",
            )
            [bar.set_alpha(0.4) for bar in mc_bars]
            [cap.set_alpha(0.4) for cap in mc_caps]

        ax.set_xlabel(self.cut_infos.cut_val_x_axis_label, **xlabel_pos)
        ax.set_ylabel(self.relative_eff_y_axis_label, **ylabel_pos)

        ax.legend(frameon=False, loc=0, ncol=1)

        export(
            fig=fig,
            filename=output_file_info.pure_tmp_file_name,
            target_dir=output_file_info.target_dir,
            close_figure=True,
        )

    def get_relative_eff_info_for(
        self,
        reco_mode_key: str,
        efficiency_evaluation_infos: Tuple[CutEvalEffInfo, ...],
    ) -> EfficiencyPlotValues:
        base_cut_val: float = min(self.cut_infos.cut_values)
        _base: List[CutEvalEffInfo] = [v for v in efficiency_evaluation_infos if v.cut_value == base_cut_val]
        assert len(_base) == 1, len(_base)
        base_eval_info: CutEvalEffInfo = _base[0]

        cut_vals: List[float] = list()
        mc_eff_vals: List[float] = list()
        mc_eff_errs: List[float] = list()
        data_eff_vals: List[float] = list()
        data_eff_errs: List[float] = list()

        for eff_info in efficiency_evaluation_infos:
            cut_vals.append(eff_info.cut_value)

            mc_eff_val: float = eff_info.mc_sow[reco_mode_key] / base_eval_info.mc_sow[reco_mode_key]
            mc_eff_vals.append(mc_eff_val)

            if eff_info.mc_sow[reco_mode_key] == base_eval_info.mc_sow[reco_mode_key]:
                mc_eff_errs.append(0.0)
            else:
                mc_eff_err_up, mc_eff_err_down = binomial_limit_approximation(
                    n=base_eval_info.mc_sow[reco_mode_key],
                    k=eff_info.mc_sow[reco_mode_key],
                )  # type: float, float
                assert np.isclose(mc_eff_err_up - mc_eff_val, mc_eff_val - mc_eff_err_down), (
                    mc_eff_val,
                    mc_eff_err_up,
                    mc_eff_err_down,
                    mc_eff_err_up - mc_eff_val,
                    mc_eff_val - mc_eff_err_down,
                )
                mc_eff_errs.append(mc_eff_err_up - mc_eff_val)

            data_eff_val: float = eff_info.data_count[reco_mode_key] / base_eval_info.data_count[reco_mode_key]
            data_eff_vals.append(data_eff_val)

            if eff_info.data_count[reco_mode_key] == base_eval_info.data_count[reco_mode_key]:
                data_eff_errs.append(0.0)
            else:
                data_eff_err_up, data_eff_err_down = binomial_limit_approximation(
                    n=base_eval_info.data_count[reco_mode_key],
                    k=eff_info.data_count[reco_mode_key],
                )  # type: float, float
                assert np.isclose(data_eff_err_up - data_eff_val, data_eff_val - data_eff_err_down)
                data_eff_errs.append(data_eff_err_up - data_eff_val)

            assert len(cut_vals) == len(mc_eff_vals) == len(mc_eff_errs), (cut_vals, mc_eff_vals, mc_eff_errs)
            assert len(cut_vals) == len(data_eff_vals) == len(data_eff_errs), (cut_vals, data_eff_vals, data_eff_errs)

        return EfficiencyPlotValues(
            reco_mode_key=reco_mode_key,
            cut_values=np.array(cut_vals),
            mc_eff=np.array(mc_eff_vals),
            mc_err=np.array(mc_eff_errs),
            data_eff=np.array(data_eff_vals),
            data_err=np.array(data_eff_errs),
        )

    def plot_efficiency_data_vs_mc(
        self,
        efficiency_evaluation_infos: Tuple[CutEvalEffInfo, ...],
        output_file_info: OutputFileInfo,
    ) -> None:
        set_matplotlibrc_params()

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=self.default_fig_size, dpi=200)  # type: FigureType, AxesType

        reco_mode_title_mapping: RecoModeGroupNameToTitleDictType = (
            self.reco_mode_info_collection.get_reco_mode_name_to_latex_label_mapping(short=False)
        )

        if self.is_single_working_point_evaluation:

            tick_label_mapping: Dict[int, str] = dict()

            for reco_index, reco_mode_info in enumerate(self.reco_mode_info_collection.values()):
                reco_mode_label: str = reco_mode_title_mapping[reco_mode_info.name]
                tick_label_mapping[reco_index] = reco_mode_label
                eff_infos = self.get_data_vs_mc_eff_info_for(
                    reco_mode_key=reco_mode_info.name,
                    efficiency_evaluation_infos=efficiency_evaluation_infos,
                )

                assert len(eff_infos.mc_eff) == len(eff_infos.mc_err) == 1, (eff_infos.mc_eff, eff_infos.mc_err)

                ax.errorbar(
                    x=np.array([reco_index]),
                    y=eff_infos.mc_eff,
                    xerr=None,
                    yerr=eff_infos.mc_err,
                    ls="",
                    marker=".",
                    alpha=1.0,
                    color=KITColors.default_colors[reco_index],
                    capsize=4,
                    label=f"{reco_mode_label}",
                )

                eff_value: float = eff_infos.mc_eff[0]
                assert isinstance(eff_value, float), (eff_value, type(eff_value))
                # ax.axhline(
                #     eff_value,
                #     color=KITColors.default_colors[reco_index],
                #     linestyle="--",
                #     linewidth=0.8,
                #     label="",
                # )

            ax.axhline(1.0, color=KITColors.kit_black, linestyle="--", linewidth=1.0, label="")

            ax.xaxis.set_major_locator(MaxNLocator(integer=True))
            x_ticks_loc: List[float] = ax.get_xticks().tolist()
            assert len(ax.get_xticklabels()) == len(x_ticks_loc), (len(ax.get_xticklabels()), x_ticks_loc)
            ax.xaxis.set_major_locator(FixedLocator(x_ticks_loc))

            x_tick_labels: List[str] = [tick_label_mapping.get(int(v), "") for v in x_ticks_loc if float(int(v)) == v]
            assert len(x_ticks_loc) == len(x_tick_labels), (x_ticks_loc, x_tick_labels)
            ax.set_xticklabels(
                x_tick_labels,
                rotation=35,
                ha="right",
                rotation_mode="anchor",
            )

            # ax.set_xlabel(self.reconstruction_mode_x_axis_label, **xlabel_pos)
            ax.set_ylabel(self.data_vs_mc_eff_y_axis_label, **ylabel_pos)

        else:
            for reco_index, reco_mode_info in enumerate(self.reco_mode_info_collection.values()):

                reco_mode_label = reco_mode_title_mapping[reco_mode_info.name]
                eff_infos = self.get_data_vs_mc_eff_info_for(
                    reco_mode_key=reco_mode_info.name,
                    efficiency_evaluation_infos=efficiency_evaluation_infos,
                )

                ax.errorbar(
                    x=eff_infos.cut_values,
                    y=eff_infos.mc_eff,
                    xerr=None,
                    yerr=eff_infos.mc_err,
                    ls="",
                    marker=".",
                    alpha=1.0,
                    color=KITColors.default_colors[reco_index],
                    capsize=4,
                    label=f"{reco_mode_label}",
                )

            ax.set_xlabel(self.cut_infos.cut_val_x_axis_label, **xlabel_pos)
            ax.set_ylabel(self.data_vs_mc_eff_y_axis_label, **ylabel_pos)

            ax.legend(frameon=False, loc=0, ncol=1)

        export(
            fig=fig,
            filename=output_file_info.pure_tmp_file_name,
            target_dir=output_file_info.target_dir,
            close_figure=True,
        )

    def get_data_vs_mc_eff_info_for(
        self,
        reco_mode_key: str,
        efficiency_evaluation_infos: Tuple[CutEvalEffInfo, ...],
    ) -> EfficiencyPlotValues:

        cut_vals: List[float] = list()
        mc_sows: List[float] = list()
        data_event_counts: List[float] = list()

        for eff_info in efficiency_evaluation_infos:
            assert eff_info.cut_value in self.cut_infos.cut_values, (eff_info.cut_value, self.cut_infos.cut_values)

            mc_sow: float = eff_info.mc_sow[reco_mode_key]
            data_event_count: float = eff_info.data_count[reco_mode_key]

            if mc_sow != 0.0:
                cut_vals.append(eff_info.cut_value)
                mc_sows.append(mc_sow)
                data_event_counts.append(data_event_count)

        assert len(cut_vals) == len(mc_sows) == len(data_event_counts), (cut_vals, mc_sows, data_event_counts)

        if self.is_single_working_point_evaluation:
            assert len(cut_vals) == 1, cut_vals

        mc_sow_array = unp.uarray(np.array(mc_sows), np.sqrt(np.array(mc_sows)))
        data_count_array = unp.uarray(np.array(data_event_counts), np.sqrt(np.array(data_event_counts)))

        eff_values_u_array: np.ndarray = data_count_array / mc_sow_array

        return EfficiencyPlotValues(
            reco_mode_key=reco_mode_key,
            cut_values=np.array(cut_vals),
            mc_eff=unp.nominal_values(eff_values_u_array),
            mc_err=unp.std_devs(eff_values_u_array),
            data_eff=np.array([]),
            data_err=np.array([]),
        )

    def append_output_info(
        self,
        eval_hist_var: HistVariable,
        method_name: str,
        extension: str,
        output_dict: Dict[str, PathType],
    ) -> OutputFileInfo:
        base_name: str = f"{self.base_file_name}_{method_name}_for_{self.cut_tag}_over_{eval_hist_var.df_label}"
        file_name: str = f"{base_name}.{extension}"
        file_path = os.path.join(self.base_dir_path, file_name)

        assert base_name not in output_dict, base_name
        assert file_path not in output_dict.values(), file_path
        output_dict.update({base_name: file_path})

        return OutputFileInfo(key=file_name, file_path=file_path)

    @staticmethod
    def get_eval_range_condition(df: pd.DataFrame, hist_var: HistVariable) -> pd.Series:
        var_col: str = hist_var.df_label
        assert var_col in df.columns, var_col
        return (df[var_col] >= hist_var.scope[0]) & (df[var_col] <= hist_var.scope[1])
