"""
Steering file for the n-tuple production for the D/D* low level cut study.

Patrick Ecker 2020
"""

import basf2

from modularAnalysis import setAnalysisConfigParams, statistics

from analysistools.processing import job_interface
from analysistools.processing import path_utilities
from analysistools.processing import skim_cuts_utils

from rdstar.online_analysis import b2bii_utils
from rdstar.online_analysis import decay_utilities, skim_cuts
from rdstar.online_analysis.decay_utilities import Reconstructor

from rdstar.studies.low_level_cut_study.utilities import decay_channels_cut_study
from rdstar.studies.low_level_cut_study.reconstruction.low_level_study_observables import LowLevelStudyObservables


cmd_args = job_interface.get_commandline_args()  # type: job_interface.BatchJobCMDLineArgs

path_utilities.setup_environment_vars()
path_utilities.setup_databases(config=cmd_args.config)
path_utilities.fei_backward_compatibility()

main_path = basf2.create_path()  # type: basf2.Path

b2bii_utils.b2bii_read_and_convert(
    path=main_path,
    input_files=cmd_args.input_files,
    input_sequences=cmd_args.input_sequences,
)

setAnalysisConfigParams(configParametersAndValues={"mcMatchingVersion": "Belle"}, path=main_path)

skim_cuts_utils.apply_sampling_skim(
    path=main_path,
    sampling_rate=cmd_args.sampling_rate,
)

# Get Final State Particles
main_path.add_module(module="CurlerCloneMarker")
decay_utilities.fsp.fsp_collection.get_final_state_particles(path=main_path, use_mc_info=cmd_args.config["use_mc_info"])
skim_cuts_utils.apply_skim_from_dict(main_path, skim_cuts.pre_skim_cuts)

observables = LowLevelStudyObservables(use_mc_info=cmd_args.config["use_mc_info"])
reconstructor = Reconstructor(path=main_path, config=cmd_args.config, decay_channels_module=decay_channels_cut_study)

# Build intermediate D mesons
reconstructor.reconstruct_D()
reconstructor.reconstruct_Dstar()

###########
# D meson #
###########

main_path.add_module(
    "VariablesToNtuple",
    fileName=cmd_args.output_dict["dzero_ntuple"],
    particleList="D0:all",
    variables=observables.get_d_meson_observables(),
)

main_path.add_module(
    "VariablesToNtuple",
    fileName=cmd_args.output_dict["dcharged_ntuple"],
    particleList="D+:all",
    variables=observables.get_d_meson_observables(),
)

############
# D* meson #
############

main_path.add_module(
    "VariablesToNtuple",
    fileName=cmd_args.output_dict["dstzero_ntuple"],
    particleList="D*0:all",
    variables=observables.get_d_star_meson_observables(),
)

main_path.add_module(
    module="VariablesToNtuple",
    fileName=cmd_args.output_dict["dstcharged_ntuple"],
    particleList="D*+:all",
    variables=observables.get_d_star_meson_observables(),
)

basf2.print_path(path=main_path)
basf2.process(path=main_path)
print(statistics)
